/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _ACTION_HPP
#define _ACTION_HPP

#include <boost/mpl/bool.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/pair.hpp>
#include <boost/mpl/assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include <queue>
#include <cassert>
#include "semigroup.hpp"
#include "root.hpp"
#include "map.hpp"
#include "set.hpp"

//a mixed binary operation S x X -> X on two structures S and X, together with an operation S x S -> S.
//The first operation is called 'MixedBinaryOperation',
//the second operation is called 'ScalarOperation'.
//S is called ScalarSet
//X is called BaseSet
//
//Examples are:
//left- and right-scalar multiplication on a module (or more general, on a semigroup)
//group operations
//left- and right-multiplication of a vector or a matrix by a square matrix
//
//An action f : S x X -> X can fulfill
//f(a*b,x) = f(a,f(b,x)) (left-associative; left-action) or
//f(a*b,x) = f(b,f(a,x)) (right-associative; right-action).
//If there is a choice, there will be a bias to left-associative actions.

template<typename SSOp, typename SXOp> struct Action_Atom;
//traits:
template<typename Action_base, typename Action = Action_base> struct IsLeftAssociative;
template<typename Action_base, typename Action = Action_base> struct IsRightAssociative;
//metafunctions;
template<typename SSOp, typename SXOp> struct Action;
//function objects:
template<typename Action_base, typename Action = Action_base, typename TC = TypeConversionByImplementation> struct ActionInPlace; 
template<typename Ac, typename id> struct SetOfOrbits;
template<typename Ac, typename id> struct Orbit;
template<typename Action_base, typename Action = Action_base> struct InnerIterator;

template<typename SSOp, typename SXOp>
struct Action_Atom {
    BOOST_MPL_ASSERT_RELATION((SSOp::NumberOfArguments::value), ==, 2);
    BOOST_MPL_ASSERT((boost::is_same<typename SSOp::ArgumentType1, typename SSOp::ResultType>));
    BOOST_MPL_ASSERT((boost::is_same<typename SSOp::ArgumentType2, typename SSOp::ResultType>));
    BOOST_MPL_ASSERT_RELATION((SXOp::NumberOfArguments::value), ==, 2);
    //Because of the bias to left-actions, we assert the syntax of SXOp to be operator()(a,x) and not operator()(x,a).
    BOOST_MPL_ASSERT((boost::is_same<typename SXOp::ArgumentType2, typename SXOp::ResultType>));
    BOOST_MPL_ASSERT((boost::is_same<typename SSOp::ResultType, typename SXOp::ArgumentType1>));
    typedef SSOp ScalarOperator;
    typedef SXOp MixedOperator;
    typedef typename SXOp::ArgumentType1 ScalarSet;
    typedef typename SXOp::ArgumentType2 BaseSet;
    //typedef Semigroup<SSOp> ScalarSemigroup;
};

template<typename Action>
struct IsLeftAssociative<Root, Action> : public boost::mpl::bool_<false> {};
//false means: undefinied
//true means: yes. This is f(g*h,x) == f(g,f(h,x)) for all g,h in ScalarSet, x in BaseSet.

template<typename Action>
struct IsRightAssociative<Root, Action> : public boost::mpl::bool_<false> {};
//false means: undefinied
//true means: yes. This is f(g*h,x) == f(h,f(g,x)) for all g,h in ScalarSet, x in BaseSet.

template<typename Action>
class ActionInPlace<Root, Action, TypeConversionByImplementation> {
    public: 
	typedef Action BaseAction;
    private:
	typedef typename Action::type::BaseSet X_;
	typedef typename Action::type::ScalarSet S_;
    public:
	template <typename Args> struct sig {
	    typedef Element<X_> & type;
	};
	typedef Element<X_> & result_type;
	//Again, syntax is consistent with left-actions, because the substitution is done in the right place.
	Element<X_> & operator()(Element<S_> const & a, Element<X_> & x) const {
	    return (x = typename Action::type::MixedOperator()(a,x));
	}
};

/*
template<typename Action>
class Orbits<Root, Action, TypeConversionByImplementation> {
    public:
	typedef Action BaseAction;
};
*/

template<typename SSOp, typename SXOp> struct Action {
    typedef Action_Atom<SSOp, SXOp> type;
};

template<typename Ac_base, typename Ac> struct IsLeftAssociative :
    public IsLeftAssociative<
	typename Base<Ac_base, IsLeftAssociative<Root> >::type,
	Ac
    >
{};

template<typename Ac_base, typename Ac> struct IsRightAssociative :
    public IsRightAssociative<
	typename Base<Ac_base, IsRightAssociative<Root> >::type,
	Ac
    >
{};

template<typename Ac_base, typename Ac, typename TC>
struct ActionInPlace :
    public ActionInPlace<
	typename Base<Ac_base, ActionInPlace<Root> >::type,
	Ac,
	typename TypeConversion<Ac_base, ActionInPlace<Root> >::type
    >
{};

template<typename Ac_base, typename Ac> struct InnerIterator :
    public InnerIterator<
	typename Base<Ac_base, InnerIterator<Root> >::type,
	Ac
    >
{};

template<typename Ac, typename id>
struct SetOfOrbits {
    typedef Ac BaseAction;
    typedef Orbit<Ac,id> ElementType;
};

template<typename Ac, typename id>
struct Base<SetOfOrbits<Ac, id> > {
    typedef typename Ac::type::BaseSet type;
};

template<typename Ac, typename id>
class Element<SetOfOrbits<Ac, id> > : public Element<typename Ac::type::BaseSet> {
    typedef SetOfOrbits<Ac, id> Parent;
};

template<typename Ac, typename id, typename Derived>
struct NameStream<SetOfOrbits<Ac, id>, Derived> : public NameStream<Root, Derived> {
    typedef Derived T_;
    typedef typename T_::BaseAction Ac_; 
    typedef typename Ac_::type::BaseSet BS_;
    typedef typename Ac_::type::ScalarSet SS_;
    std::ostream & operator()(std::ostream & o) const {
	return NameStream<BS_>()(NameStream<SS_>()(o) << " \\\\ ");
    }
};

template<typename Ac, typename id, typename Derived>
class ElementName<SetOfOrbits<Ac,id>, Derived> : public ElementName<Root, Derived> {
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_; 
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
    public:
	std::ostream& operator()(std::ostream& o, Element<T_> const & a) const {
	    return o << "O(" << ((Element<BS_>)a) << ")";
	}
};

template<typename Ac, typename id, typename Derived>
class Random<SetOfOrbits<Ac,id>, Derived> : public Random<Root, Derived> {
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_; 
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
    public:
	Element<T_> operator()() const {
	    return Element<T_>()(Random<BS_>()());
	}
};

template<typename Ac, typename id, typename Derived>
class Representative<SetOfOrbits<Ac,id>, Derived> : public Representative<Root, Derived> {
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_; 
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
    public:
	Element<T_> operator()() const {
	    return Element<T_>()(Representative<BS_>()());
	}
};

template<typename Ac, typename id>
struct Base<SetOfOrbits<Ac,id>, RankElement<Root> > {
    typedef Root type;
};

template<typename Ac, typename id>
struct Base<SetOfOrbits<Ac,id>, UnrankElement<Root> > {
    typedef Root type;
};

template<typename Ac, typename id>
struct Base<SetOfOrbits<Ac,id>, Cardinality<Root> > {
    typedef Root type;
};

template<typename Ac, typename id, typename Derived>
class Iterator<SetOfOrbits<Ac, id>, Derived> :
    public boost::iterator_facade<
        Iterator<SetOfOrbits<Ac, id>, Derived>,
        Element<Derived> const,
        boost::single_pass_traversal_tag,
        Element<Derived> const &,
        signed int
    >
{
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
	typedef typename T_::ElementType ET_;
    friend class boost::iterator_core_access;
    friend class Begin<SetOfOrbits<Ac, id>, Derived>;
    friend class IsAtEnd<SetOfOrbits<Ac, id>, Derived>;
    friend class InnerIterator<SetOfOrbits<Ac, id>, Derived>;
    public:
        typedef Orbit<Ac_, Iterator<Derived> > InnerOrbitType;
	typedef Iterator<InnerOrbitType> InnerIteratorType;
        Iterator() : inner_it(visited) {}

    private:
	//at the end all elements of BS_ will be contained in the hash_set.
	//so a bitvector would be more efficient.
	//on the other hand, an inner iterator over a single orbits using a bitvector would be a waste in many situations.
	HASH_SET(Element<BS_>) visited;
	Iterator<BS_> outer_it;
	InnerIteratorType inner_it;

	/////////

    public:
	void increment() {
	    while(!IsAtEnd<Orbit<Ac_, Iterator<Derived> > >()(inner_it)) {
		++inner_it;
	    }
	    while(visited.count(*outer_it) != 0) {
		++outer_it;
		if(IsAtEnd<BS_>()(outer_it)) return;
	    }
	    ET_::init(*outer_it);
	    assert(inner_it.newElements.empty());
	    inner_it.newElements.push(ET_::representative);
	    visited.insert(ET_::representative);
	}

	bool equal(Iterator<SetOfOrbits<Ac,id> > const & other) const {
	    return (outer_it == other.outer_it);
	}

        Element<T_> const & dereference() const {
            return (Element<T_> &)(*outer_it);
        }
};

template<typename Ac, typename id, typename Derived>
struct Begin<SetOfOrbits<Ac, id>, Derived> : public Begin<Root, Derived> {
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
	typedef typename T_::ElementType ET_;
    public:
	Iterator<T_> operator()() const {
	    Iterator<T_> it;
	    it.outer_it = Begin<BS_>()();
	    ET_::init(*(it.outer_it));
	    it.inner_it.newElements.push(ET_::representative);
	    it.visited.insert(ET_::representative);
	    return it;
	}
};

template<typename Ac, typename id, typename Derived>
struct IsAtEnd<SetOfOrbits<Ac, id>, Derived> : public IsAtEnd<Root, Derived> {
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
	typedef typename T_::ElementType ET_;
    public:
	bool operator()(Iterator<T_> const & it) const {
	    return IsAtEnd<BS_>()(it.outer_it);
	}
};

template<typename Ac, typename id, typename Derived>
class InnerIterator<SetOfOrbits<Ac, id>, Derived> {
    private:
	typedef Derived T_;
	typedef typename T_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
	typedef typename T_::ElementType ET_;
	typedef Iterator<Orbit<Ac_, Iterator<Derived> > > InnerIteratorType;
    public:
	InnerIteratorType operator()(Iterator<T_> const & it) {
	    return it.inner_it;
	}
};

template<typename Ac, typename id>
struct Orbit : public Ac::type::BaseSet {
    typedef Ac BaseAction;
    static Element<typename Ac::type::BaseSet> representative;
    static void init(Element<typename Ac::type::BaseSet> const & repIn) {
	representative = repIn;
    }
};

template<typename Ac, typename id>
struct Base<Orbit<Ac, id> > {
    typedef typename Ac::type::BaseSet type;
};

template<typename Ac, typename id>
Element<typename Ac::type::BaseSet> Orbit<Ac, id>::representative;

template<typename Ac, typename id>
struct Element<Orbit<Ac, id> > : public Element<typename Ac::type::BaseSet> {
    Element(Element<typename Ac::type::BaseSet> const & repIn) : Element<typename Ac::type::BaseSet>(repIn) {}
};

template<typename Ac, typename id, typename Derived>
struct NameStream<Orbit<Ac, id>, Derived> : public NameStream<Root, Derived> {
    typedef Derived O_;
    typedef typename O_::BaseAction Ac_; 
    typedef typename Ac_::type::BaseSet BS_;
    typedef typename Ac_::type::ScalarSet SS_;
    std::ostream & operator()(std::ostream & o) const {
	return o << ((Element<SetOfOrbits<Ac_, id> >&)Representative<O_, id>()());
    }
};

template<typename Ac, typename id, typename Derived>
class ElementName<Orbit<Ac,id>, Derived> : public ElementName<Root, Derived> {
    private:
	typedef Derived O_;
	typedef typename O_::BaseAction Ac_; 
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
    public:
	std::ostream& operator()(std::ostream& o, Element<O_> const & a) const {
	    return ElementName<BS_>()(o,(Element<BS_>&)a);
	}
};

template<typename Ac, typename id>
struct Base<Orbit<Ac,id>, Random<Root> > {
    typedef Root type;
};

template<typename Ac, typename id, typename Derived>
class Representative<Orbit<Ac,id>, Derived> : public Representative<Root, Derived> {
    private:
	typedef Derived O_;
	typedef typename O_::BaseAction Ac_; 
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
    public:
	Element<O_> operator()() const {
	    return O_::representative;
	}
};

template<typename Ac, typename id>
struct Base<Orbit<Ac,id>, RankElement<Root> > {
    typedef Root type;
};

template<typename Ac, typename id>
struct Base<Orbit<Ac,id>, UnrankElement<Root> > {
    typedef Root type;
};

template<typename Ac, typename id>
struct Base<Orbit<Ac,id>, Cardinality<Root> > {
    typedef Root type;
};

template<typename Ac, typename id, typename Derived>
class Iterator<Orbit<Ac, id>, Derived> :
    public boost::iterator_facade<
        Iterator<Orbit<Ac, id>, Derived>,
        Element<Derived> const,
        boost::single_pass_traversal_tag,
        Element<Derived> const &,
        signed int
    >
{
    private:
	typedef Derived O_;
	typedef typename O_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
    friend class boost::iterator_core_access;
    friend class Begin<Orbit<Ac, id>, Derived>;
    friend class IsAtEnd<Orbit<Ac, id>, Derived>;
    friend class Begin<SetOfOrbits<Ac, id>, Derived>;
    public:
        Iterator() : visited(internalHash) {}

        Iterator(HASH_SET(Element<BS_>) & hash) : visited(hash) {
	    //newElements.push(O_::representative);
	    //visited.insert(O_::representative);
	};

    private:
	HASH_SET(Element<BS_>) & visited;
	HASH_SET(Element<BS_>) internalHash;

	/////////

    public:
	std::queue<Element<BS_> > newElements;

	void increment() {
	    Element<BS_> a = newElements.front();
	    newElements.pop();
	    typedef std::vector<Element<SS_> > SS_vec;
	    typedef typename SS_vec::const_iterator Iter;
	    SS_vec gens = Generators<Semigroup<SOp_> >()();
	    for(Iter i = gens.begin(); i != gens.end(); ++i) {
		Element<BS_> b = MOp_()(*i,a);
		if (visited.count(b) == 0) {
		    newElements.push(b);
		    visited.insert(b);
		}
	    }
	}

	bool equal(Iterator<Orbit<Ac,id> > const & other) const {
	    return (newElements.empty() && other.newElements.empty()) || (newElements.front() == other.newElements.front());
	}

        Element<O_> const & dereference() const {
            //return Element<O_>(newElements.front());
            return (Element<O_> &) newElements.front();
        }
};

template<typename Ac, typename id, typename Derived>
struct Begin<Orbit<Ac, id>, Derived> : public Begin<Root, Derived> {
    private:
	typedef Derived O_;
	typedef typename O_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
    public:
	Iterator<O_> operator()() const {
	    Iterator<O_> it;
	    it.newElements.push(O_::representative);
	    it.visited.insert(O_::representative);
	    return it;
	}
};

template<typename Ac, typename id, typename Derived>
struct IsAtEnd<Orbit<Ac, id>, Derived> : public IsAtEnd<Root, Derived> {
    private:
	typedef Derived O_;
	typedef typename O_::BaseAction Ac_;
	typedef typename Ac_::type::BaseSet BS_;
	typedef typename Ac_::type::ScalarSet SS_;
	typedef typename Ac_::type::ScalarOperator SOp_;
	typedef typename Ac_::type::MixedOperator MOp_;
    public:
	bool operator()(Iterator<O_> const & it) const {
	    return (it.newElements.empty());
	}
};

#endif
