/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _CARTESIAN_POWER_SET_HPP
#define _CARTESIAN_POWER_SET_HPP

#include <iostream>
#include <vector>
#include <algorithm>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include "set.hpp"
#include "mathHelper.hpp"
#include "root.hpp"

//cartesianPowerSet.hpp
template<typename S, typename VWExp> class CartesianPowerSet;
template<typename CP_base, typename CP = CP_base, typename TC = TypeConversionByImplementation> struct Exponent;
template<typename CP_base, typename CP = CP_base, typename TC = TypeConversionByImplementation> struct Coordinate;

template<typename S, typename VWExp> class Element<CartesianPowerSet<S, VWExp> >;

template<typename S, typename VWExp>
class CartesianPowerSet {
    public:
	typedef S BaseSet;
	typedef std::vector<Element<BaseSet> > DataV_;
    public:
	static unsigned int getExponent() {
	    return VWExp::value();
	}

	static void init(unsigned int const & exponentIn) {
	    VWExp::init(exponentIn);
	}
};

template<typename S, typename VWExp>
class Element<CartesianPowerSet<S, VWExp> > : public Element_root<CartesianPowerSet<S, VWExp> > {
    public:
	typedef CartesianPowerSet<S, VWExp> CP_;
	typedef typename CP_::BaseSet CPBaseSet_;
	typedef typename CP_::DataV_ DataV_;

	DataV_ data;

	Element() : data(std::vector<Element<CPBaseSet_> >(Exponent<CP_>()())) {}
	Element(std::vector<Element<CPBaseSet_> > const & dataIn) : data(dataIn) {}
	Element(Element<CPBaseSet_> const & dataIn) : data(Exponent<CP_>()(),dataIn) {}

	Element<CPBaseSet_> const & getCoordinate(unsigned int const i) const {
	    return data[i];
	}

	Element<CPBaseSet_> & getCoordinate(unsigned int const i) {
	    return data[i];
	}
};

template<typename CP, typename TC>
struct Exponent<Root, CP, TC> {
    template <typename Args> struct sig {
        typedef unsigned int type;
    };
    unsigned int operator()() {
	return CP::getExponent();
    }
};

template<typename CP, typename TC>
struct Coordinate<Root, CP, TC> {
    template <typename Args> struct sig {
        typedef Element<typename CP::BaseSet> & type;
    };

    Element<typename CP::BaseSet> & operator()(Element<CP> & a, unsigned int const i) const {
	return a.getCoordinate(i);
    }

    Element<typename CP::BaseSet> const & operator()(Element<CP> const & a, unsigned int const i) const {
	return a.getCoordinate(i);
    }
};

template<typename S_base, typename S, typename TC> struct Exponent :
    public Exponent<
        typename Base<S_base, Exponent<Root> >::type,
        S,
        typename TypeConversion<S_base, Exponent<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Coordinate :
    public Coordinate<
        typename Base<S_base, Coordinate<Root> >::type,
        S,
        typename TypeConversion<S_base, Coordinate<Root> >::type
    >
{};

// set.hpp
template<typename S, typename VWExp, typename Derived>
struct Equals<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public Equals<Root, Derived> {
    typedef Derived CP_;
    bool operator()(Element<CP_> const & v, Element<CP_> const & w) {
	return (v.data == w.data);
    }
};

template<typename S, typename VWExp, typename Derived>
struct NameStream<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public NameStream<Root, Derived> {
    std::ostream & operator()(std::ostream & o) {
	return (NameStream<S>()(o << "[") << "]^" << Exponent<Derived>()());
    }
};

template<typename S, typename VWExp, typename Derived>
class ElementName<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public ElementName<Root, Derived> {
    private:
	typedef Derived CP_;
    public:
    std::ostream & operator()(std::ostream & o, Element<CP_> const & v) const {
	o << "[ ";
	for(size_t i = 0; i < Exponent<CP_>()(); i++) {
	    if (i != 0) {
		o << ", ";
	    }
	    o << v.getCoordinate(i);
	}
	o << " ]";
	return o;
    }
};

template<typename S, typename VWExp, typename Derived>
class HashValue<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public HashValue<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;
    public:
	size_t operator()(Element<CP_> const & v) {
	    size_t res = 0;
	    unsigned int i = Exponent<CP_>()() - 1;
	    while(true) {
		res += HashValue<CPBaseSet_>()(v.getCoordinate(i));
		if (i == 0) {
		    return res;
		}
		--i;
		res *= Cardinality<CPBaseSet_>()();
	    }
	}
};

template<typename S, typename VWExp, typename Derived>
class RankElement<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public RankElement<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;
    public:
	unsigned int operator()(Element<CP_> const & v) {
	    unsigned int res = 0;
	    unsigned int i = Exponent<CP_>()() - 1;
	    while(true) {
		res += RankElement<CPBaseSet_>()(v.getCoordinate(i));
		if (i == 0) {
		    return res;
		}
		--i;
		res *= Cardinality<CPBaseSet_>()();
	    }
	}
};

template<typename S, typename VWExp, typename Derived>
class UnrankElement<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public UnrankElement<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;
    public:
	Element<CP_> operator()(unsigned int const & n) {
	    Element<CP_> res;
	    unsigned int i = 0;
	    unsigned int ub = Exponent<CP_>()();
	    unsigned int nn = n;
	    while(true) {
		//res.data[i] = UnrankElement<CPBaseSet_>()(nn % Cardinality<CPBaseSet_>()());
		Coordinate<CP_>()(res,i) = UnrankElement<CPBaseSet_>()(nn % Cardinality<CPBaseSet_>()());
		//SetCoordinate<CP_>()(res,i,UnrankElement<CPBaseSet_>()(nn % Cardinality<CPBaseSet_>()()));
		++i;
		if (i == ub) {
		    return res;
		}
		nn /= Cardinality<CPBaseSet_>()();
	    }
	}
};

template<typename S, typename VWExp, typename Derived>
class Compare<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public Compare<Root, Derived> {
    //lexicographic order
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;
    public:
	bool operator()(Element<CP_> const & v, Element<CP_> const & w) {
	    unsigned int const exp = Exponent<CP_>()();
	    for(unsigned int i = 0; i < exp; ++i) {
		if(v.getCoordinate(i) != w.getCoordinate(i)) {
		    return Compare<CPBaseSet_>()(v.getCoordinate(i), w.getCoordinate(i));
		}
	    }
	    return false;
	}
};

template<typename S, typename VWExp, typename Derived>
class Cardinality<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public Cardinality<Root, Derived> {
    //only works if it fits into a unsigned int
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;

    public:
	unsigned int operator()() {
	    return mathHelper::pow(Cardinality<CPBaseSet_>()(),Exponent<CP_>()());
	}
};

template<typename S, typename VWExp, typename Derived>
struct Random<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public Random<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;

    public:
	Element<CP_> operator()() const {
	    using boost::lambda::_1;
	    Element<CP_> res = Element<CP_>(std::vector<Element<CPBaseSet_> >(Exponent<CP_>()()));
	    //why doesn't this work instead?:
	    //Element<CP_>(std::vector<Element<CPBaseSet_> >(Exponent<CP_>()()));
	    for_each(res.data.begin(),res.data.end(),_1 = boost::lambda::bind(Random<CPBaseSet_>()));
	    return res;
	}
};

template<typename S, typename VWExp, typename Derived>
struct Representative<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public Representative<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;

    public:
	Element<CP_> operator()() const {
	    using boost::lambda::_1;
	    Element<CP_> res = Element<CP_>(std::vector<Element<CPBaseSet_> >(Exponent<CP_>()()));
	    //why doesn't this work instead?:
	    //Element<CP_>(std::vector<Element<CPBaseSet_> >(Exponent<CP_>()()));
	    for_each(res.data.begin(),res.data.end(),_1 = boost::lambda::bind(Representative<CPBaseSet_>()));
	    return res;
	}
};

template<typename S, typename VWExp, typename Derived>
struct End<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation>;

template<typename S, typename VWExp, typename Derived>
struct IsAtBegin<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation>;

template<typename S, typename VWExp, typename Derived>
struct IsAtEnd<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation>;

template<typename S, typename VWExp, typename Derived>
struct Iterator<CartesianPowerSet<S, VWExp>, Derived> : 
public boost::iterator_facade<
    Iterator<CartesianPowerSet<S, VWExp>, Derived>,
    Element<Derived> const,
    boost::incrementable_traversal_tag
>
{
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;
	typedef std::vector<Iterator<S> > ItV_;
	typedef typename ItV_::iterator ItIt_;
	typedef typename CP_::DataV_ DataV_;
	typedef typename DataV_::iterator _DataIt;

    friend class boost::iterator_core_access;
    friend struct End<CartesianPowerSet<S, VWExp>, Derived>;
    friend struct IsAtBegin<CartesianPowerSet<S, VWExp>, Derived>;
    friend struct IsAtEnd<CartesianPowerSet<S, VWExp>, Derived>;

    public:
	Iterator() :
	//in this implementation the standard constructor sets the iterator to the begin.
	    coordinateIterators(Exponent<CP_>()(),Begin<S>()()),
	    value(DataV_(Exponent<CP_>()(), (*(coordinateIterators[0])))),
	    atEnd(false) {}

    private:
	ItV_ coordinateIterators;
	Element<CP_> value;
	bool atEnd;

	void increment() {
	    ItIt_ iIt = coordinateIterators.begin();
	    _DataIt iData = value.data.begin();
	    _DataIt iDataEnd = value.data.end();
	    while(iData != iDataEnd) {
		++(*iIt);
		if (!IsAtEnd<S>()(*iIt)) {
		    (*iData) = (*(*iIt));
		    return;
		} else {
		    (*iIt) = Begin<S>()();
		    (*iData) = (*(*iIt));
		    ++iData;
		    ++iIt;
		}
	    }
	    atEnd = true;
	}

	bool equal(Iterator<CartesianPowerSet<S, VWExp>, Derived> const & other) const {
	    return (value == other.value && atEnd == other.atEnd);
	}

	Element<CP_> const & dereference() const {
	    return value;
	}
};

template<typename S, typename VWExp, typename Derived>
class Begin<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public Begin<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;

    public:
	Iterator<Derived> operator()() const {
	    return Iterator<Derived>();
	}
};

template<typename S, typename VWExp, typename Derived>
class End<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public End<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;

    public:
	Iterator<Derived> operator()() const {
	    Iterator<Derived> result;
	    result.atEnd = true;
	    return result;
	}
};

template<typename S, typename VWExp, typename Derived>
class IsAtBegin<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public IsAtBegin<Root, Derived> {
    private: 
	typedef Derived CP_;
	typedef typename CP_::BaseSet _S;
	typedef std::vector<Iterator<_S> > ItV_;
	typedef typename ItV_::const_iterator ItIt_;

    public:
	bool operator()(Iterator<CP_> const & it) const {
	    if (it.atEnd) {
		return false;
	    }
	    ItIt_ iItEnd = it.coordinateIterators.end();
	    for(ItIt_ iIt = it.coordinateIterators.begin(); iIt != iItEnd; ++iIt) {
		if (!IsAtBegin<_S>()(*iIt)) {
		    return false;
		}
	    }
	    return true;
	}
};

template<typename S, typename VWExp, typename Derived>
class IsAtEnd<CartesianPowerSet<S, VWExp>, Derived, TypeConversionByImplementation> : public IsAtEnd<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::BaseSet CPBaseSet_;

    public:
	bool operator()(Iterator<CP_> const & it) const {
	    return it.atEnd;
	}
};
#endif
