/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _CHAIN_RING_HPP
#define _CHAIN_RING_HPP

#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/bind.hpp>
#include <boost/mpl/bool.hpp>
#include <set>
#include <vector>
#include "ring.hpp"
#include "root.hpp"
#include "matrixMul.hpp"
#include "semidirectProduct.hpp"
#include "automorphismGroup.hpp"

template<typename R_base, typename R = R_base> struct IsChainRing;

template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct BasePrime;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Rank;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct ChainLength;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct ResidueFieldOrder;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Theta;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Alpha;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Period;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Height;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct EisensteinDegree;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct PartialRightThetaAdicExpansion;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct RightThetaAdicExpansion;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct CompressedRightThetaAdicExpansion;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct SetRightThetaAdicExpansion;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct RightMultiplicationByThetaPower;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct RightDivisionByThetaPower;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct GeneralizedRightInverse;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct GeneralizedLeftInverse;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct RepresentativeModThetaPower;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct InPlaceRightMultiplicationByThetaPower;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct InPlaceRightDivisionByThetaPower;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct InPlaceRepresentativeModThetaPower;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Sigma;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct SigmaExp;
//Sigma is a mapping R->R such that a * Theta = Theta * sigma(a) for all ring elements in R.
//CAUTION: The direction is left to right!

template<typename R>
struct IsChainRing<Root, R> : public boost::mpl::bool_<false> {};
//false means: undefinied
//true means: yes

template<typename R_base, typename R> struct IsChainRing :
    public IsChainRing<
        typename Base<R_base, IsCommutative<Root> >::type,
        R
    >
{};

template<typename R>
struct BasePrime<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const;
};

template<typename R>
struct Rank<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const;
};

template<typename R>
struct ChainLength<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const;
};

template<typename R>
struct ResidueFieldOrder<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const;
};

template<typename R>
struct Theta<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    Element<typename R::type::BaseSet> operator()() const;
};

template<typename R>
struct Alpha<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    Element<typename R::type::BaseSet> operator()() const;
};

template<typename R> 
struct Period<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(Element<typename R::type::BaseSet> const & a) const {
	return repeatedApplicationIndex(
	    a,
	    boost::bind(typename R::type::Mul(), _1, Theta<R>()()),
	    EqualsNeutralElement<Semigroup<typename R::type::Add> >(),
	    boost::mpl::true_(),
	    boost::mpl::true_()
	) - 1;
    }
};

template<typename R> 
struct Height<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(Element<typename R::type::BaseSet> const & a) const {
	BOOST_MPL_ASSERT((IsChainRing<R>));
	return ChainLength<R>()() - Period<R>()(a);
    }
};

template<typename R> 
struct EisensteinDegree<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const {
	BOOST_MPL_ASSERT((IsChainRing<R>));
	Element<typename R::type::BaseSet> x=NeutralElement<Semigroup<typename R::type::Mul> >()();
	unsigned int p=BasePrime<R>()();
	x=Power<Semigroup<typename R::type::Add> >()(x,p);
	return Height<R>()(x);
    }
};

template<typename R> 
struct PartialRightThetaAdicExpansion<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef std::vector<unsigned int> type;
    };
    std::vector<unsigned int> & operator()(Element<typename R::type::BaseSet> const & a, unsigned int const len) const;
};

//{a_0, ..., a_(q-1)} set of Teichmueller-elements, t:=theta
//for x=a_(p_0)*t^0+a_(p_1)*t^1+...+a_(p_(m-1))*t^(m-1) return the vector of unsigned integers (p_0,p_1,...,p_(m-1))
template<typename R> 
struct RightThetaAdicExpansion<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef std::vector<unsigned int> type;
    };
    std::vector<unsigned int> & operator()(Element<typename R::type::BaseSet> const & a) const {
	return PartialRightThetaAdicExpansion<R>()(a,ChainLength<R>()());
    }
};

template<typename R> 
struct CompressedRightThetaAdicExpansion<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(Element<typename R::type::BaseSet> const & a) const {
	unsigned int res;
	std::vector<unsigned int> rtae = RightThetaAdicExpansion<R>()(a);
	typedef std::vector<unsigned int>::const_reverse_iterator It;
	It it = rtae.rbegin();
	res = (*it);
	It ub = rtae.rend();
	unsigned int q = ResidueFieldOrder<R>()();
	while((++it) != ub ) {
	    res *= q;
	    res += (*it);
	}
	return res;
    }
};

//{a_0, ..., a_(q-1)} set of Teichmueller-elements, t:=theta, v = (v_s,...,v_(m-1)) vector of unsigned integers, then
//a_(p_0)*t^0+...+a_(p_(m-1))*t^(m-1) |-> a_(p_0)*t^0+...+a_(p_(s-1))*t^(s-1)+a_(v_s)*t^(p_s)+...+a_(v_(m-1))*t^(m-1)
//the following standard implementation only works for "full" input of size ChainLenth<R>
template<typename R> 
struct SetRightThetaAdicExpansion<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template <typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> & type;
    };
    Element<typename R::type::BaseSet> & operator()(Element<typename R::type::BaseSet> & a, std::vector<unsigned int> const & v) const {
	a = NeutralElement<Semigroup<typename R::type::Add> >()();
	unsigned int m = ChainLength<R>()();
	for(unsigned int i = 0; i < m; ++i) {
	    unsigned int coef = v[i];
	    if (coef != 0) {
		InRightPlace<Semigroup<typename R::type::Add> >()(RightMultiplicationByThetaPower<R>()(Power<Semigroup<typename R::type::Mul> >()(Alpha<R>()(),coef-1),i),a);
	    }
	}
	return a;
    }
};

template<typename R> 
struct RightMultiplicationByThetaPower<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    Element<typename R::type::BaseSet> operator()(Element<typename R::type::BaseSet> const & a, unsigned int const p) const {
	return a * Power<Semigroup<typename R::type::Mul> >()(Theta<R>()(),p);
    }
};

//{a_0, ..., a_(q-1)} set of Teichmueller-elements, t:=theta, r:= power of theta
//for x=a_(p_0)*t^0+a_(p_1)*t^1+...+a_(p_(m-1))*t^(m-1) return a_(p_r)*t^0+a_(p_(r+1))*t^1+...+a_(p_(m-1))*t^(m-1-r)
template<typename R> 
struct RightDivisionByThetaPower<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    Element<typename R::type::BaseSet> operator()(Element<typename R::type::BaseSet> const & a, unsigned int const p) const {
	std::vector<unsigned int> v = RightThetaAdicExpansion<R>()(a);
	for (unsigned int i=0; i < p; i++){
	    v.erase(v.begin());
	    v.push_back(0);
	}
	Element<typename R::type::BaseSet> b;
	SetRightThetaAdicExpansion<R>()(b,v);
	return b;
    }
};

template<typename R> 
struct GeneralizedRightInverse<Root, R, TypeConversionByImplementation> {
//given a chain ring element x of height i, compute a unit y with xy = theta^i
//For i > 0 the element y is not unique.
//In contrast to the ordinary inverse, there is not necessarily an element y satisfying xy = yx = theta^i.
//For that reason, the function is called 'GeneralizedRightInverse' and not 'Generalized Inverse'
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    typedef typename R::type::BaseSet R_BS_;
    Element<R_BS_> operator()(Element<R_BS_> const & a) const {
	Element<R_BS_> thetaPow = Power<Semigroup<typename R::type::Mul> >()(Theta<R>()(),Height<R>()(a));
	Iterator<R_BS_> it = Begin<R_BS_>()();
	for( ; !IsAtEnd<R_BS_>()(it); ++it) {
	    if(typename R::type::Mul()(a,*it) == thetaPow) break;
	}
	return (*it);
    }
};

template<typename R> 
struct GeneralizedLeftInverse<Root, R, TypeConversionByImplementation> {
//given a chain ring element x of height i, compute a unit y with xy = theta^i
//For i > 0 the element y is not unique.
//In contrast to the ordinary inverse, there is not necessarily an element y satisfying xy = yx = theta^i.
//For that reason, the function is called 'GeneralizedLeftInverse' and not 'Generalized Inverse'
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    typedef typename R::type::BaseSet R_BS_;
    Element<R_BS_> operator()(Element<R_BS_> const & a) const {
	Element<R_BS_> thetaPow = Power<Semigroup<typename R::type::Mul> >()(Theta<R>()(),Height<R>()(a));
	Iterator<R_BS_> it = Begin<R_BS_>()();
	for( ; !IsAtEnd<R_BS_>()(it); ++it) {
	    if(typename R::type::Mul()(*it,a) == thetaPow) break;
	}
	return (*it);
    }
};

//{a_0, ..., a_(q-1)} set of Teichmueller-elements, t:=theta, r:= power of theta
//for x=a_(p_0)*t^0+a_(p_1)*t^1+...+a_(p_(m-1))*t^(m-1) return a_(p_0)*t^0+a_(p_1)*t^1+...+a_(p_(r-1))*t^(r-1)
template<typename R> 
struct RepresentativeModThetaPower<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    Element<typename R::type::BaseSet> operator()(Element<typename R::type::BaseSet> & a, unsigned int const p) const;
};

//as above, but overwrite the given element
template<typename R> 
struct InPlaceRightMultiplicationByThetaPower<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> & type;
    };
    Element<typename R::type::BaseSet> & operator()(Element<typename R::type::BaseSet> & a, unsigned int const p) const {
	a = a * Power<Semigroup<typename R::type::Mul> >()(Theta<R>()(),p);
	return a;
    }
};

//as above, but overwrite the given element
template<typename R> 
struct InPlaceRightDivisionByThetaPower<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> & type;
    };
    Element<typename R::type::BaseSet> & operator()(Element<typename R::type::BaseSet> & a, unsigned int const p) const {
	std::vector<unsigned int> v=RightThetaAdicExpansion<R>()(a);
	for (unsigned int i=0; i<p; i++){
	    v.erase(v.begin());
	    v.push_back(0);
	}
	SetRightThetaAdicExpansion<R>()(a,v);
	return a;
    }
};

//as above, but overwrite the given element
template<typename R> 
struct InPlaceRepresentativeModThetaPower<Root, R, TypeConversionByImplementation> {
    BOOST_MPL_ASSERT((IsChainRing<R>));
    template<typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> & type;
    };
    Element<typename R::type::BaseSet> & operator()(Element<typename R::type::BaseSet> & a, unsigned int const p) const;
};

template<typename R> 
struct Sigma<Root, R, TypeConversionByImplementation> {
//Sigma is a mapping R->R such that a * Theta = Theta * sigma(a) for all ring elements in R.
//CAUTION: The direction is left to right!
    BOOST_MPL_ASSERT((IsChainRing<R>));
    typedef typename R::type::BaseSet R_BS_;
    template<typename Args> struct sig {
	typedef Element<R_BS_> type;
    };
    Element<R_BS_> operator()(Element<R_BS_> const & a) const {
	Element<R_BS_> res = NeutralElement<Semigroup<typename R::type::Mul> >()();
	Element<R_BS_> alphatheta = typename R::type::Mul()(Alpha<R>()(), Theta<R>()());
	Element<R_BS_> sigmaalpha = Alpha<R>()();
	unsigned int pPower = 1;
	unsigned int qminus1 = ResidueFieldOrder<R>()() - 1;
	while(typename R::type::Mul()(Theta<R>()(), sigmaalpha) != alphatheta) {
	    sigmaalpha = Power<Semigroup<typename R::type::Mul> >()(sigmaalpha,BasePrime<R>()());
	    pPower *= BasePrime<R>()();
	    pPower %= qminus1;
	}
	std::vector<unsigned int> v = RightThetaAdicExpansion<R>()(a);
	typedef typename std::vector<unsigned int>::iterator It;
	It ub = v.end();
	for(It it = v.begin(); it != v.end(); ++it) {
	    if ((*it) != 0) {
		--(*it);
		(*it) *= pPower;
		(*it) %= qminus1;
		++(*it);
	    }
	}
	SetRightThetaAdicExpansion<R>()(res,v);
	return res;
    }
};

template<typename R> 
struct SigmaExp<Root, R, TypeConversionByImplementation> {
//SigmaExp is a mapping R->R such that a * Theta = Theta * sigma(a) for all ring elements in R.
//CAUTION: The direction is left to right!
    BOOST_MPL_ASSERT((IsChainRing<R>));
    typedef typename R::type::BaseSet R_BS_;
    template<typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const {
	unsigned int res = 0;
	Element<R_BS_> alphatheta = typename R::type::Mul()(Alpha<R>()(), Theta<R>()());
	Element<R_BS_> sigmaalpha = Alpha<R>()();
	while(typename R::type::Mul()(Theta<R>()(), sigmaalpha) != alphatheta) {
	    sigmaalpha = Power<Semigroup<typename R::type::Mul> >()(sigmaalpha,BasePrime<R>()());
	    ++res;
	}
	assert(res < ResidueFieldOrder<R>()());
	return res;
    }
};

template<typename CR1, typename CR2> 
class ApplyAlphaThetaImagesToElement {
    private: 
	typedef typename CR1::type::BaseSet CR1_BS_;
	typedef typename CR2::type::BaseSet CR2_BS_;

    public:
	Element<CR2_BS_> operator()(Element<CR1_BS_> const & x, Element<CR2_BS_> const & alphaIm, Element<CR2_BS_> const & thetaIm) const {
	    Element<CR2_BS_> res = Zero<CR2_BS_>()();
	    std::vector<unsigned int> tae = RightThetaAdicExpansion<CR1>()(x);
	    Element<CR2_BS_> thetaPower = One<CR2_BS_>()();
	    unsigned int i = 0;
	    while (true) {
		if (tae[i] != 0) {
		    res += Power<Semigroup<typename CR2::type::Mul> >()(alphaIm,tae[i] - 1) * thetaPower;
		}
		++i;
		if (i == ChainLength<CR2>()()) {
		    return res;
		}
		thetaPower *= thetaIm;
	    }
	}
};

template<typename R_base, typename R, typename TC> struct BasePrime :
    public BasePrime<
        typename Base<R_base, BasePrime<Root> >::type,
	R,
	typename TypeConversion<R_base, BasePrime<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Rank :
    public Rank<
        typename Base<R_base, Rank<Root> >::type,
	R,
	typename TypeConversion<R_base, Rank<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct ChainLength :
    public ChainLength<
        typename Base<R_base, ChainLength<Root> >::type,
	R,
	typename TypeConversion<R_base, ChainLength<Root> >::type
    >
{};

//////////


template<typename R_base, typename R, typename TC> struct ResidueFieldOrder :
    public ResidueFieldOrder<
        typename Base<R_base, ResidueFieldOrder<Root> >::type,
	R,
	typename TypeConversion<R_base, ResidueFieldOrder<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Theta :
    public Theta<
        typename Base<R_base, Theta<Root> >::type,
	R,
	typename TypeConversion<R_base, Theta<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Alpha :
    public Alpha<
        typename Base<R_base, Alpha<Root> >::type,
	R,
	typename TypeConversion<R_base, Alpha<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Period :
    public Period<
        typename Base<R_base, Period<Root> >::type,
	R,
	typename TypeConversion<R_base, Period<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Height :
    public Height<
        typename Base<R_base, Height<Root> >::type,
	R,
	typename TypeConversion<R_base, Height<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct EisensteinDegree :
    public EisensteinDegree<
        typename Base<R_base, EisensteinDegree<Root> >::type,
	R,
	typename TypeConversion<R_base, EisensteinDegree<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct PartialRightThetaAdicExpansion :
    public PartialRightThetaAdicExpansion<
        typename Base<R_base, PartialRightThetaAdicExpansion<Root> >::type,
	R,
	typename TypeConversion<R_base, PartialRightThetaAdicExpansion<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct RightThetaAdicExpansion :
    public RightThetaAdicExpansion<
        typename Base<R_base, RightThetaAdicExpansion<Root> >::type,
	R,
	typename TypeConversion<R_base, RightThetaAdicExpansion<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct CompressedRightThetaAdicExpansion :
    public CompressedRightThetaAdicExpansion<
        typename Base<R_base, CompressedRightThetaAdicExpansion<Root> >::type,
	R,
	typename TypeConversion<R_base, CompressedRightThetaAdicExpansion<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct SetRightThetaAdicExpansion :
    public SetRightThetaAdicExpansion<
        typename Base<R_base, SetRightThetaAdicExpansion<Root> >::type,
	R,
	typename TypeConversion<R_base, SetRightThetaAdicExpansion<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct RightMultiplicationByThetaPower :
    public RightMultiplicationByThetaPower<
        typename Base<R_base, RightMultiplicationByThetaPower<Root> >::type,
	R,
	typename TypeConversion<R_base, RightMultiplicationByThetaPower<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct RightDivisionByThetaPower :
    public RightDivisionByThetaPower<
        typename Base<R_base, RightDivisionByThetaPower<Root> >::type,
	R,
	typename TypeConversion<R_base, RightDivisionByThetaPower<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct GeneralizedRightInverse :
    public GeneralizedRightInverse<
        typename Base<R_base, GeneralizedRightInverse<Root> >::type,
	R,
	typename TypeConversion<R_base, GeneralizedRightInverse<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct GeneralizedLeftInverse :
    public GeneralizedLeftInverse<
        typename Base<R_base, GeneralizedLeftInverse<Root> >::type,
	R,
	typename TypeConversion<R_base, GeneralizedLeftInverse<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct RepresentativeModThetaPower :
    public RepresentativeModThetaPower<
        typename Base<R_base, RepresentativeModThetaPower<Root> >::type,
	R,
	typename TypeConversion<R_base, RepresentativeModThetaPower<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct InPlaceRightMultiplicationByThetaPower :
    public InPlaceRightMultiplicationByThetaPower<
        typename Base<R_base, InPlaceRightMultiplicationByThetaPower<Root> >::type,
	R,
	typename TypeConversion<R_base, InPlaceRightMultiplicationByThetaPower<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct InPlaceRightDivisionByThetaPower :
    public InPlaceRightDivisionByThetaPower<
        typename Base<R_base, InPlaceRightDivisionByThetaPower<Root> >::type,
	R,
	typename TypeConversion<R_base, InPlaceRightDivisionByThetaPower<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct InPlaceRepresentativeModThetaPower :
    public InPlaceRepresentativeModThetaPower<
        typename Base<R_base, InPlaceRepresentativeModThetaPower<Root> >::type,
	R,
	typename TypeConversion<R_base, InPlaceRepresentativeModThetaPower<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Sigma :
    public Sigma<
        typename Base<R_base, Sigma<Root> >::type,
	R,
	typename TypeConversion<R_base, Sigma<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct SigmaExp :
    public SigmaExp<
        typename Base<R_base, SigmaExp<Root> >::type,
	R,
	typename TypeConversion<R_base, SigmaExp<Root> >::type
    >
{};

/*
template<typename R> 
//R should be a TableChainRing;
class AutomorphismGroup_ChainRing_impl : public AutomorphismGroup_root<R> {};

template<typename R>
class Element_AutomorphismGroup_ChainRing_impl : public Element_root<R> {
    private:
	Element<typename R::type::BaseSet> imageTheta; //image of Theta<R>()()
	Element<typename R::type::BaseSet> imageAlpha; //image of Alpha<R>()()

    public:
	typedef AutomorphismGroup<R> BaseSet;
    
	Element_AutomorphismGroup_ChainRing_impl(){};
	Element_AutomorphismGroup_ChainRing_impl(Element<typename R::type::BaseSet> const imageThetaIn, Element<typename R::type::BaseSet> const imageAlphaIn):
	    imageTheta(imageThetaIn),
	    imageAlpha(imageAlphaIn)
	{}

	Element<typename R::type::BaseSet> getImageTheta() const {
	    return imageTheta;
	}

	Element<typename R::type::BaseSet> getImageAlpha() const {
	    return imageAlpha;
	}

	void setImageTheta(Element<typename R::type::BaseSet> const imageThetaIn) {
	    imageTheta = imageThetaIn;
	}

	void setImageAlpha(Element<typename R::type::BaseSet> const imageAlphaIn) {
	    imageAlpha = imageAlphaIn;
	}
};

template<typename R>
struct Compare_AutomorphismGroup_ChainRing_impl {
    typedef AutomorphismGroup<R> A;
    template<typename Args> struct sig {
        typedef bool type;
    };
    bool operator()(Element<A> const & a, Element<A> const & b) const {
        if (a.getImageTheta() < b.getImageTheta()) {
            return true;
        } else {
            return (a.getImageAlpha() == b.getImageAlph()) || (a.getImageTheta() < b.getImageTheta());
        }
    }
};

template<typename R>
struct HashValue_AutomorphismGroup_ChainRing_impl {
    typedef AutomorphismGroup<R> A;
    template<typename Args> struct sig {
        typedef size_t type;
    };
    size_t operator()(Element<A> const & a) const {
        return SetOrder<R>()() * (a.getImageAlpha()).getVal() + (a.getImageTheta()).getVal();
    }
};

template<typename R>
struct Equals_AutomorphismGroup_ChainRing_impl {
    typedef AutomorphismGroup<R> A;
    template<typename Args> struct sig {
        typedef bool type;
    };
    bool operator()(Element<A> const & a, Element<A> const & b) const {
        return (a.getImageTheta() == b.getImageTheta()) && (a.getImageAlpha() == b.getImageAlpha());
    }
};

template<typename R>
struct ElementName_AutomorphismGroup_ChainRing_impl {
    typedef AutomorphismGroup<R> A;
    template<typename Args> struct sig {
        typedef std::ostream & type;
    };
    std::ostream & operator()(std::ostream & o, Element<A> const & a) const {
        return o << "[ theta |-> " << a.getImageTheta() << ", alpha |-> " << a.getImageAlpha() << " ]";
    }
};

template<typename R>
struct Mul_AutomorphismGroup_ChainRing_impl : public BinaryOperation<AutomorphismGroup<R> > {
    typedef AutomorphismGroup<R> A;
    Element<A> operator()(Element<A> const & a, Element<A> const & b) const {
        return Element<A>(Application<A>()(a,b.getImageTheta()),Application<A>()(a,b.getImageAlpha()));
    }
};

template<typename R>
struct NeutralElement_Mul_AutomorphismGroup_ChainRing_impl {
    typedef AutomorphismGroup<R> A;
    template<typename Args> struct sig {
        typedef Element<A> type;
    };
    Element<A> operator()() const {
        return Element<A>(Theta<R>()(),Alpha<R>()());
    }
};

template<typename R>
struct EqualsNeutralElement_Mul_AutomorphismGroup_ChainRing_impl {
    typedef AutomorphismGroup<R> A;
    template<typename Args> struct sig {
        typedef bool type;
    };
    bool operator()(Element<A> const &a) const {
        return (a.getImageTheta() == Theta<R>()() && a.getImageAlpha() == Alpha<R>()());
    }
};
*/

template<typename R>
unsigned int basePrime(Element<typename R::type::BaseSet> const & a) {
    return BasePrime<R>()();
}

template<typename R>
unsigned int rank(Element<typename R::type::BaseSet> const & a) {
    return Rank<R>()();
}

template<typename R>
unsigned int chainLength(Element<typename R::type::BaseSet> const & a) {
    return ChainLength<R>()();
}

template<typename R>
unsigned int residueFieldOrder(Element<typename R::type::BaseSet> const & a) {
    return ResidueFieldOrder<R>()();
}

template<typename R>
Element<typename R::type::BaseSet> theta() {
    return Theta<R>()();
}

template<typename R>
Element<typename R::type::BaseSet> alpha() {
    return Alpha<R>()();
}

template<typename R>
unsigned int height(Element<typename R::type::BaseSet> const & a) {
    return Height<R>()(a);
}

template<typename R>
unsigned int period(Element<R> const & a) {
    return Period<typename R::BaseTableChainRing>()(a);
}

template<typename Vres, typename Mleft, typename Vright, typename CR>
struct ProjectiveMatrixVectorMul_Op :
    public Map<
	boost::mpl::vector<
	    Mleft,
	    Vright
	>,
	Vres
    >
{
    private:
	typedef typename CR::type::BaseSet CR_BS_;
    public:
	Element<Vres> operator()(Element<Mleft> const & m, Element<Vright> const & v) {
	    Element<Vres> res = MatrixVectorMul_impl<Vres, Mleft, Vright, CR>()(m, v);
	    typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    It ub = res.data.end();
	    It it;
	    It pivot = res.data.begin();
	    unsigned int period = 0;
	    for(it = res.data.begin(); it != ub; ++it) {
		unsigned int per_new = Period<CR>()(*it);
		if (per_new > period) {
		    period = per_new;
		    pivot = it;
		    if(period == ChainLength<CR>()()) {
			break;
		    }
		}
	    }
	    Element<CR_BS_> inv = GeneralizedRightInverse<CR>()(*pivot);
	    for(it = res.data.begin(); it != ub; ++it) {
		InLeftPlace<Semigroup<typename CR::type::Mul> >()(inv,(*it));
	    }
	    return res;
	}
};

template<typename Vres, typename GammaM, typename Vright, typename CR>
struct ProjectiveGammaMatrixVectorMul_Op :
    public Map<
	boost::mpl::vector<
	    GammaM,
	    Vright
	>,
	Vres
    >
{
    private:
	typedef typename CR::type::BaseSet CR_BS_;
	typedef AutomorphismGroup<CR> AG_;
    public:
	Element<Vres> operator()(Element<GammaM> const & m, Element<Vright> const & v) {
	    Element<Vright> res;
            typedef typename std::vector<Element<CR_BS_> >::const_iterator ItC;
            typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    It ub = res.data.end();
	    ItC vit = v.data.begin();
	    for(It resit = res.data.begin(); resit != ub; ++resit) {
		//*resit = Application<Semigroup<typename GammaM::HOp_> >()(Right<GammaM>()(m),*vit);
		//in the following line, it must be possible to cast Right<GammaM>()(m) to an element of AG.
		*resit = Application<AG_>()(Right<GammaM>()(m),*vit);
		//*resit = typename GammaM::HOp_()(Right<GammaM>()(m),*vit);
		++vit;
	    }
	    res = MatrixVectorMul_impl<Vres, typename GammaM::LeftBaseSet, Vright, CR>()(Left<GammaM>()(m), res);
	    typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    ub = res.data.end();
	    It it;
	    It pivot = res.data.begin();
	    unsigned int period = 0;
	    for(it = res.data.begin(); it != ub; ++it) {
		unsigned int per_new = Period<CR>()(*it);
		if (per_new > period) {
		    period = per_new;
		    pivot = it;
		    if(period == ChainLength<CR>()()) {
			break;
		    }
		}
	    }
	    Element<CR_BS_> inv = GeneralizedRightInverse<CR>()(*pivot);
	    for(it = res.data.begin(); it != ub; ++it) {
		(*it) = typename CR::type::Mul()(*it,inv);
	    }
	    return res;
	}
};

template<typename Vres, typename GammaM, typename Vright, typename CR>
struct ProjectiveReverseGammaMatrixVectorMul_Op :
    public Map<
	boost::mpl::vector<
	    GammaM,
	    Vright
	>,
	Vres
    >
{
    private:
	typedef typename CR::type::BaseSet CR_BS_;
	typedef AutomorphismGroup<CR> AG_;
    public:
	Element<Vres> operator()(Element<GammaM> const & m, Element<Vright> const & v) {
	    Element<Vright> res;
            typedef typename std::vector<Element<CR_BS_> >::const_iterator ItC;
            typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    It ub = res.data.end();
	    ItC vit = v.data.begin();
	    typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    for(It resit = res.data.begin(); resit != ub; ++resit) {
		//*resit = Application<Semigroup<typename GammaM::HOp_> >()(Right<GammaM>()(m),*vit);
		//in the following line, it must be possible to cast Right<GammaM>()(m) to an element of AG.
		*resit = Application<AG_>()(Right<GammaM>()(m),*vit);
		//*resit = typename GammaM::HOp_()(Right<GammaM>()(m),*vit);
		++vit;
	    }
	    typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    res = ReverseMatrixVectorMul_impl<Vres, typename GammaM::LeftBaseSet, Vright, CR>()(Left<GammaM>()(m), res);
	    typedef typename std::vector<Element<CR_BS_> >::iterator It;
	    ub = res.data.end();
	    It it;
	    It pivot = res.data.begin();
	    unsigned int period = 0;
	    for(it = res.data.begin(); it != ub; ++it) {
		unsigned int per_new = Period<CR>()(*it);
		if (per_new > period) {
		    period = per_new;
		    pivot = it;
		    if(period == ChainLength<CR>()()) {
			break;
		    }
		}
	    }
	    Element<CR_BS_> inv = GeneralizedLeftInverse<CR>()(*pivot);
	    for(it = res.data.begin(); it != ub; ++it) {
		InRightPlace<Semigroup<typename CR::type::Mul> >()(inv,(*it));
	    }
	    return res;
	}
};

template<typename R>
struct TestChainRingProperty {
    typedef typename R::type::BaseSet R_BS_;
    bool operator()() {
	typedef typename std::set<Element<R_BS_>, Compare<R_BS_> >::const_iterator It;
	std::set<Element<R_BS_>, Compare<R_BS_> > zeroDivisors;
	for(Iterator<R_BS_> it = Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(it); ++it) {
	    if (IsZeroDivisor<R>()(*it)) {
		zeroDivisors.insert(*it);
	    }
	}
	if(zeroDivisors.size() == 1) return true; //handle the case that R is a finite field.
	std::set<Element<R_BS_>, Compare<R_BS_> > generatorCandidates(zeroDivisors.begin(),zeroDivisors.end());
	It ub = zeroDivisors.end();
	for(It it = zeroDivisors.begin(); it != ub; ++it) {
	    for(It it2 = zeroDivisors.begin(); it2 != ub; ++it2) {
		generatorCandidates.erase(typename R::type::Mul()(*it,*it2));
	    }
	}
	if (generatorCandidates.size() == 0) {
	    std::cout << "no candidates left!" << std::endl;
	    return false;
	}
	Element<R_BS_> theta = (*(generatorCandidates.begin()));
	for(Iterator<R_BS_> it = Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(it); ++it) {
	    zeroDivisors.erase(typename R::type::Mul()(*it,theta));
	}
	return (zeroDivisors.empty());
    }
};

template<typename R>
struct AddTableAsTeX {
    typedef typename R::type::BaseSet R_BS_;
    typedef Iterator<R_BS_> It;
    std::string operator()() {
	unsigned int card=Cardinality<R_BS_>()();
    	std::stringstream ss;
	ss<<"\\begin{center}\nAddition table for the ring "<<Name<R_BS_>()()<<":\n\\end{center}"<<std::endl;
	ss<<"\\setlength{\\arraycolsep}{0.3em}\n";
	ss<<"\\[\n\\begin{array}{c|";
	for (unsigned int i=0; i<card; ++i){
	    ss<<"c";
	}
	ss<<"}\n";
	ss<<"+";
	for (It itcol=Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(itcol); ++itcol){
	    ss<<"&"<<(*itcol);
	}
	ss<<"\\\\\\hline"<<std::endl;
	for (It itrow=Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(itrow); ++itrow){
	    ss<<*itrow;
	    for (It itcol=Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(itcol); ++itcol){
	        ss<<"&"<<(*itrow)+(*itcol);
	    }
	    ss<<"\\\\"<<std::endl;
	}    
	ss<<"\\end{array}\n\\]";
    	return ss.str();
    }
};    

template<typename R>
struct MulTableAsTeX {
    typedef typename R::type::BaseSet R_BS_;
    typedef Iterator<R_BS_> It;
    std::string operator()() {
	unsigned int card=Cardinality<R_BS_>()();
    	std::stringstream ss;
	ss<<"\\begin{center}\nMultiplication table for the ring "<<Name<R_BS_>()()<<":\n\\end{center}"<<std::endl;
	ss<<"\\setlength{\\arraycolsep}{0.3em}\n";
	ss<<"\\[\n\\begin{array}{c|";
	for (unsigned int i=0; i<card; ++i){
	    ss<<"c";
	}
	ss<<"}\n";
	ss<<"\\cdot";
	for (It itcol=Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(itcol); ++itcol){
	    ss<<"&"<<(*itcol);
	}
	ss<<"\\\\\\hline"<<std::endl;
	for (It itrow=Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(itrow); ++itrow){
	    ss<<*itrow;
	    for (It itcol=Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(itcol); ++itcol){
	        ss<<"&"<<(*itrow)*(*itcol);
	    }
	    ss<<"\\\\"<<std::endl;
	}    
	ss<<"\\end{array}\n\\]";
    	return ss.str();
    }
};    
#endif
