/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

// raw matrix data structure
// implementations for full rectangular and square matrices

#ifndef _MATRIX_SET_HPP
#define _MATRIX_SET_HPP

#include "cartesianPowerSet.hpp"
#include "root.hpp"
#include "valueWrapperPolicy.hpp"
#include <boost/mpl/pair.hpp>
#include <boost/mpl/int.hpp>

template<typename R, typename MS_WV> struct MatrixSet; //dummy class
template<typename R, typename id> struct RectangularMatrixSet;
template<typename R, typename K_VW> struct SquareMatrixSet;
template<typename M_base, typename M = M_base> struct Entry;
template<typename M_base, typename M = M_base> struct NumberOfRows;
template<typename M_base, typename M = M_base> struct NumberOfColumns;

template<typename R, typename VW_MS>
struct MatrixSet : public CartesianPowerSet<R, VW_MS> {};

template<typename R, typename id>
class RectangularMatrixSet : public MatrixSet<R, ValueWrapperByInitFunction<unsigned int, id> > {
    private:
	static unsigned int numberOfRows;
	static unsigned int numberOfColumns;

    public:
	static void init(unsigned int const numberOfRowsIn, unsigned int const numberOfColumnsIn) {
	    numberOfRows = numberOfRowsIn;
	    numberOfColumns = numberOfColumnsIn;
	    CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >::init(numberOfRows * numberOfColumns);
	}

	static unsigned int getNumberOfRows() {
	    return numberOfRows;
	}

	static unsigned int getNumberOfColumns() {
	    return numberOfColumns;
	}
};

template<typename R, typename id>
unsigned int RectangularMatrixSet<R, id>::numberOfRows;

template<typename R, typename id>
unsigned int RectangularMatrixSet<R, id>::numberOfColumns;

template<typename R, unsigned int k>
class SquareMatrixSet<R, ValueWrapperByTemplateParameter<unsigned int, k> > : public MatrixSet<R, ValueWrapperByTemplateParameter<unsigned int, k*k> > {
    public:
	typedef ValueWrapperByTemplateParameter<unsigned int, k*k> MS_VW_;
	typedef ValueWrapperByTemplateParameter<unsigned int, k> K_VW_;
	static unsigned int getNumberOfRows() {
	    return k;
	}
};

template<typename R, typename id>
class SquareMatrixSet<R, ValueWrapperByInitFunction<unsigned int, id> > : public MatrixSet<R, ValueWrapperByInitFunction<unsigned int, boost::mpl::pair<id, boost::mpl::int_<37> > > > {
    public:
	typedef ValueWrapperByInitFunction<unsigned int, boost::mpl::pair<id, boost::mpl::int_<37> > > MS_VW_;
	typedef ValueWrapperByInitFunction<unsigned int, id> K_VW_;
	static void init(unsigned int const numberOfRowsIn) {
	    K_VW_::init(numberOfRowsIn);
	    CartesianPowerSet<R, MS_VW_>::init(numberOfRowsIn * numberOfRowsIn);
	}

	static unsigned int getNumberOfRows() {
	    return K_VW_::value();
	}
};

template<typename M>
struct Entry<Root, M> {
    typedef typename M::BaseSet S;
    template <typename Args> struct sig {
        typedef Element<S> & type;
    };
    Element<S> & operator()(Element<M> const & a, unsigned int const row, unsigned int const column) const;
};

template<typename M>
struct NumberOfRows<Root, M> {
    template <typename Args> struct sig {
        typedef unsigned int type;
    };
    unsigned int operator()() const;
};

template<typename M>
struct NumberOfColumns<Root, M> {
    template <typename Args> struct sig {
        typedef unsigned int type;
    };
    unsigned int operator()() const;
};

template<typename M_base, typename M> struct Entry : public Entry<typename Base<M_base, Entry<Root> >::type, M> {};
template<typename M_base, typename M> struct NumberOfRows : public NumberOfRows<typename Base<M_base, NumberOfRows<Root> >::type, M> {};
template<typename M_base, typename M> struct NumberOfColumns : public NumberOfColumns<typename Base<M_base, NumberOfColumns<Root> >::type, M> {};

// * common implementations for rectangular and square matrices *
template<typename R, typename MS_VW, typename Derived>
struct Entry<MatrixSet<R, MS_VW>, Derived> : public Entry<Root, Derived> {
    typedef Derived _M;
    Element<R> & operator()(Element<_M> & a, unsigned int const row, unsigned int const column) const {
	return Coordinate<_M>()(a,row * NumberOfColumns<_M>()() + column);
    }

    Element<R> const & operator()(Element<_M> const & a, unsigned int const row, unsigned int const column) const {
	return Coordinate<_M>()(a,row * NumberOfColumns<_M>()() + column);
    }
};

template<typename R, typename MS_VW, typename Derived>
struct NumberOfRows<MatrixSet<R, MS_VW>, Derived> : public NumberOfRows<Root, Derived> {
    typedef Derived _M;
    unsigned int operator()() const {
	return _M::getNumberOfRows();
    }
};

template<typename R, typename MS_VW, typename Derived>
struct ElementName<MatrixSet<R, MS_VW>, Derived> : public ElementName<Root, Derived> {
    typedef Derived _M;
    std::ostream& operator()(std::ostream& o, Element<_M> const & a) const {
	for(unsigned int r = 0; r < NumberOfRows<_M>()(); r++) {
	    if (r != 0) {
		o << "\n";
	    }
	    o << "[ ";
	    for(unsigned int c = 0; c < NumberOfColumns<_M>()(); c++) {
		if (c != 0) {
		    o << "   ";
		}
		o << Entry<_M>()(a,r,c);
	    }
	    o << " ]";
	}
        return o;
    }
};

// * RectangularMatrixSet *

template<typename R, typename id>
class Element<RectangularMatrixSet<R, id> > : public Element<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> > > {
    public:
        //manually forward the constructors since this is not done automatically.
        Element() {};
        Element(std::vector<Element<R> > const & dataIn) : Element<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> > >(dataIn) {};
};

//matrixSet.hpp

template<typename R, typename id, typename Derived>
struct Entry<RectangularMatrixSet<R, id>, Derived> : public Entry<MatrixSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct NumberOfRows<RectangularMatrixSet<R, id>, Derived> : public NumberOfRows<MatrixSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct NumberOfColumns<RectangularMatrixSet<R, id>, Derived> : public NumberOfColumns<Root, Derived> {
    typedef Derived _M;
    unsigned int operator()() const {
	return _M::getNumberOfColumns();
    }
};

//cartesianPower.hpp
template<typename R, typename id, typename Derived>
struct Exponent<RectangularMatrixSet<R, id>, Derived> : public Exponent<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct Coordinate<RectangularMatrixSet<R, id>, Derived> : public Coordinate<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

//set.hpp

template<typename R, typename id, typename Derived>
struct NameStream<RectangularMatrixSet<R, id>, Derived> : public NameStream<Root, Derived> {
    typedef Derived _M;
    std::ostream& operator()(std::ostream& o) const {
        return NameStream<typename _M::BaseSet>()(o << "Mat(" << NumberOfRows<_M>()() << ", " << NumberOfColumns<_M>()() << ", ") << ")";
    }
};

template<typename R, typename id, typename Derived>
struct Equals<RectangularMatrixSet<R, id>, Derived> : public Equals<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct ElementName<RectangularMatrixSet<R, id>, Derived> : public ElementName<MatrixSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct HashValue<RectangularMatrixSet<R, id>, Derived> : public HashValue<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct Compare<RectangularMatrixSet<R, id>, Derived> : public Compare<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct Cardinality<RectangularMatrixSet<R, id>, Derived> : public Cardinality<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct Random<RectangularMatrixSet<R, id>, Derived> : public Random<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct Iterator<RectangularMatrixSet<R, id>, Derived> : public Iterator<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct Begin<RectangularMatrixSet<R, id>, Derived> : public Begin<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct End<RectangularMatrixSet<R, id>, Derived> : public End<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct IsAtBegin<RectangularMatrixSet<R, id>, Derived> : public IsAtBegin<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

template<typename R, typename id, typename Derived>
struct IsAtEnd<RectangularMatrixSet<R, id>, Derived> : public IsAtEnd<CartesianPowerSet<R, ValueWrapperByInitFunction<unsigned int, id> >, Derived> {};

// * SquareMatrixSet *

template<typename R, typename K_VW>
class Element<SquareMatrixSet<R, K_VW> > : public Element<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_> > {
    public:
        //manually forward the constructors since this is not done automatically.
        Element() {};
        Element(std::vector<Element<R> > const & dataIn) : Element<CartesianPowerSet<R,typename SquareMatrixSet<R, K_VW>::MS_VW_> >(dataIn) {};
};

// matrixSet.hpp

template<typename R, typename K_VW, typename Derived>
struct Entry<SquareMatrixSet<R, K_VW>, Derived> : public Entry<MatrixSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_>, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct NumberOfRows<SquareMatrixSet<R, K_VW>, Derived> : public NumberOfRows<MatrixSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_>, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct NumberOfColumns<SquareMatrixSet<R, K_VW>, Derived> : public NumberOfColumns<Root, Derived> {
    typedef Derived _M;
    unsigned int operator()() const {
	return _M::getNumberOfRows();
    }
};

//cartesianPowerSet.hpp

template<typename R, typename K_VW, typename Derived>
struct Exponent<SquareMatrixSet<R, K_VW>, Derived> : public Exponent<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct Coordinate<SquareMatrixSet<R, K_VW>, Derived> : public Coordinate<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

//set.hpp

template<typename R, typename K_VW, typename Derived>
struct Equals<SquareMatrixSet<R, K_VW>, Derived> : public Equals<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct NameStream<SquareMatrixSet<R, K_VW>, Derived> : public NameStream<Root, Derived> {
    typedef Derived _M;
    std::ostream& operator()(std::ostream& o) const {
        return NameStream<typename _M::BaseSet>()(o << "Mat(" << NumberOfRows<_M>()() << ", ") << ")";
    }
};

template<typename R, typename K_VW, typename Derived>
struct ElementName<SquareMatrixSet<R, K_VW>, Derived> : public ElementName<MatrixSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_>, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct HashValue<SquareMatrixSet<R, K_VW>, Derived> : public HashValue<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct Compare<SquareMatrixSet<R, K_VW>, Derived> : public Compare<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct Cardinality<SquareMatrixSet<R, K_VW>, Derived> : public Cardinality<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct Random<SquareMatrixSet<R, K_VW>, Derived> : public Random<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct Iterator<SquareMatrixSet<R, K_VW>, Derived> : public Iterator<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct Begin<SquareMatrixSet<R, K_VW>, Derived> : public Begin<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct End<SquareMatrixSet<R, K_VW>, Derived> : public End<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct IsAtBegin<SquareMatrixSet<R, K_VW>, Derived> : public IsAtBegin<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};

template<typename R, typename K_VW, typename Derived>
struct IsAtEnd<SquareMatrixSet<R, K_VW>, Derived> : public IsAtEnd<CartesianPowerSet<R, typename SquareMatrixSet<R, K_VW>::MS_VW_ >, Derived> {};
#endif
