/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef NUMBER_RANGE_SET_HPP_
#define NUMBER_RANGE_SET_HPP_

#include <cassert>
#include <ostream>
#include <boost/mpl/assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include "set.hpp"
#include "valueWrapperPolicy.hpp"
#include "root.hpp"

template<typename LB_VW, typename UB_VW>
//gives the set LB_VW::value() ... UP_VW::value() - 1,
//so please note that the lower bound is included in the set,
//while the upper bound is not.
class NumberRangeSet {
    BOOST_MPL_ASSERT((boost::is_same<typename LB_VW::ValueType, typename UB_VW::ValueType>));
    public:
	typedef typename LB_VW::ValueType ValueType;
	typedef LB_VW LowerBound_ValueWrapper;
	typedef UB_VW UpperBound_ValueWrapper;

	static void initLowerBound(ValueType const & lb) {
	    LB_VW::init(lb);
	}

	static void initUpperBound(ValueType const & ub) {
	    UB_VW::init(ub);
	}

	static void init(ValueType const & lb, ValueType const & ub) {
	    assert(lb <= ub);
	    LB_VW::init(lb);
	    UB_VW::init(ub);
	}
};

/*
template<typename LB_VW, typename UB_VW>
struct Base<NumberRangeSet<LB_VW, UB_VW> > {
    typedef Root type;
};
*/

template<typename NRS_base, typename NRS = NRS_base, typename TC = TypeConversionByImplementation>
struct LowerBound;

template<typename NRS_base, typename NRS = NRS_base, typename TC = TypeConversionByImplementation>
struct UpperBound;

template<typename LB_VW, typename UB_VW, typename Derived>
struct LowerBound<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef typename Derived::ValueType type;
    };
    typename Derived::ValueType operator()() const {
	return Derived::LowerBound_ValueWrapper::value();
    }
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct UpperBound<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef typename Derived::ValueType type;
    };
    typename Derived::ValueType operator()() const {
	return Derived::UpperBound_ValueWrapper::value();
    }
};

template<typename NRS_base, typename NRS, typename TC> struct LowerBound :
    public LowerBound<
	typename Base<NRS_base, LowerBound<Root> >::type,
	NRS,
	typename TypeConversion<NRS_base, LowerBound<Root> >::type
    >
{};

template<typename NRS_base,typename  NRS, typename TC> struct UpperBound :
    public UpperBound<
	typename Base<NRS_base, UpperBound<Root> >::type,
	NRS,
	typename TypeConversion<NRS_base, UpperBound<Root> >::type
    >
{};

// set.hpp

template<typename LB_VW, typename UB_VW>
struct Element<NumberRangeSet<LB_VW, UB_VW> > : public Element_root<NumberRangeSet<LB_VW, UB_VW> > {
    private:
	typedef Element_root<NumberRangeSet<LB_VW, UB_VW> > Parent_;
	typedef typename Parent_::BaseSet::ValueType ValueType_;
	ValueType_ val;
    public:
	Element(){};
	Element(ValueType_ valIn) : val(valIn) {};

	ValueType_ & value() {
	    return val;
	}

	ValueType_ const & value() const {
	    return val;
	}

	void setValue(ValueType_ const & valIn) {
	    val = valIn;
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class Equals<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public Equals<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	bool operator()(Element<NRS_> const &a, Element<NRS_> const &b) const {
	    return (a.value() == b.value());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class NameStream<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public NameStream<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	std::ostream& operator()(std::ostream& o) const {
	    return (o << "{" << LowerBound<NRS_>()() << ", ..., " << UpperBound<NRS_>()() - 1 << "}");
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class ElementName<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public ElementName<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	std::ostream& operator()(std::ostream& o, Element<NRS_> const & a) const {
	    return (o << a.value());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class HashValue<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public HashValue<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	size_t operator()(Element<NRS_> const & a) const {
	    return a.value();
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class RankElement<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public RankElement<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	unsigned int operator()(Element<NRS_> const & a) const {
	    return a.value() - LowerBound<NRS_>()();
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class UnrankElement<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public UnrankElement<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	Element<NRS_> operator()(unsigned int const & n) const {
	    return Element<NRS_>(n + LowerBound<NRS_>()());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class Compare<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public HashValue<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	bool operator()(Element<NRS_> const & a, Element<NRS_> const & b) const {
	    return (a.value() < b.value());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
class Iterator<NumberRangeSet<LB_VW, UB_VW>, Derived> :
    public boost::iterator_facade<
        Iterator<NumberRangeSet<LB_VW, UB_VW>, Derived>,
        Element<Derived> const,
        boost::random_access_traversal_tag,
        Element<Derived> const &,
        signed int
    >
{
    friend class boost::iterator_core_access;
    public:
        Iterator() {}
        //undefined starting position

        Iterator(unsigned int const n) : value(Element<Derived>(n)) {};

    private:
        Element<Derived> value;

        void increment() {
            ++value.value();
        }

        void decrement() {
            --value.value();
        }

        bool equal(Iterator<NumberRangeSet<LB_VW, UB_VW>, Derived> const & other) const {
            return Equals<Derived>()(value == other.value);
        }

        void advance(signed int const n) {
            value.value() += n;
        }

        Element<Derived> const & dereference() const {
            return value;
        }
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct Begin<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public Begin<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	Iterator<NRS_> operator()() const {
	    return Iterator<NRS_>(LowerBound<NRS_>()());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct End<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public End<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	Iterator<Derived> operator()() const {
	    return Iterator<NRS_>(UpperBound<NRS_>()());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct IsAtBegin<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public IsAtBegin<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	bool operator()(Iterator<NRS_> const & i) const {
	    return ((*i) == Element<NRS_>(LowerBound<NRS_>()()));
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct IsAtEnd<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public IsAtEnd<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	bool operator()(Iterator<NRS_> const & i) const {
	    return ((*i) == Element<NRS_>(UpperBound<NRS_>()()));
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct Cardinality<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public Cardinality<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	unsigned int operator()() const {
	    return UpperBound<NRS_>()() - LowerBound<NRS_>()();
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct Random<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public Random<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	Element<NRS_> operator()() const {
	    return Element<NRS_>(LowerBound<NRS_>()() + rand() % Cardinality<NRS_>()());
	}
};

template<typename LB_VW, typename UB_VW, typename Derived>
struct Representative<NumberRangeSet<LB_VW, UB_VW>, Derived, TypeConversionByImplementation> : public Representative<Root, Derived> {
    private:
	typedef Derived NRS_;
    public:
	Element<NRS_> operator()() const {
	    return Element<NRS_>(LowerBound<NRS_>()());
	}
};

#endif
