/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _RING_HPP
#define _RING_HPP

#include <boost/bind.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/mpl/assert.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/int.hpp>
#include <boost/mpl/equal_to.hpp>
#include "additiveSemigroup.hpp"
#include "multiplicativeSemigroup.hpp"
#include "root.hpp"
#include "algorithm.hpp"

//the default implementation is for *finite* rings!

template<typename OpAdd, typename OpMul>
struct Ring_Atom {
    //add is an abelian group
    BOOST_MPL_ASSERT_RELATION((OpAdd::NumberOfArguments::value), ==, 2);
    BOOST_MPL_ASSERT((boost::is_same<typename OpAdd::ArgumentType1, typename OpAdd::ResultType>));
    BOOST_MPL_ASSERT((boost::is_same<typename OpAdd::ArgumentType2, typename OpAdd::ResultType>));
    BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<OpAdd> >));
    BOOST_MPL_ASSERT((HasInverseElements<Semigroup<OpAdd> >));
    BOOST_MPL_ASSERT((IsCommutative<Semigroup<OpAdd> >));
    //mul is semigroup
    BOOST_MPL_ASSERT_RELATION((OpMul::NumberOfArguments::value), ==, 2);
    BOOST_MPL_ASSERT((boost::is_same<typename OpMul::ArgumentType1, typename OpMul::ResultType>));
    BOOST_MPL_ASSERT((boost::is_same<typename OpMul::ArgumentType2, typename OpMul::ResultType>));
    //OpAdd and OpMul are defined on the same set.
    BOOST_MPL_ASSERT((boost::is_same<typename OpAdd::ResultType, typename OpMul::ResultType>));

    typedef OpAdd Add;
    typedef OpMul Mul;
    typedef typename OpAdd::ResultType BaseSet;
};

template<typename OpAdd, typename OpMul>
struct Ring {
    typedef Ring_Atom<OpAdd,OpMul> type;
};

template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct IsZeroDivisor;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct NilpotencyIndex;

template<typename R>
struct IsZeroDivisor<Root, R, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(Element<typename R::type::BaseSet> const & a) const {
	return !IsInvertibleElement<Semigroup<typename R::type::Mul> >()(a);
    }
};

template<typename R> 
struct NilpotencyIndex<Root, R, TypeConversionByImplementation> {
//returns 0 if 'a' is not nilpotent, otherwise the smallest positive integer n with a^n == 0
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(Element<typename R::type::BaseSet> const & a) const {
	return repeatedApplicationIndex(
		a,
		boost::bind(typename R::type::Mul(),_1,_1),
		EqualsNeutralElement<Semigroup<typename R::type::Add> >(),
		boost::mpl::bool_<false>(),
		boost::mpl::bool_<false>()
	);
    }
};

template<typename R_base, typename R, typename TC> struct IsZeroDivisor :
    public IsZeroDivisor<
	typename Base<R_base, IsZeroDivisor<Root> >::type,
	R,
	typename TypeConversion<R_base, IsZeroDivisor<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct NilpotencyIndex :
    public NilpotencyIndex<
	typename Base<R_base, NilpotencyIndex<Root> >::type,
	R,
	typename TypeConversion<R_base, NilpotencyIndex<Root> >::type
    >
{};

template<typename OpAdd, typename OpMul>
struct RingWithUnity {
    //multiplication has a neutral element
    
    typedef Ring_Atom<OpAdd,OpMul> type;
};

template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct MinusOne;
template<typename R_base, typename R = R_base, typename TC = TypeConversionByImplementation> struct Characteristic;

template<typename R, typename TC>
struct MinusOne<Root, R, TC> {
    template <typename Args> struct sig {
	typedef Element<typename R::type::BaseSet> type;
    };
    Element<typename R::type::BaseSet> operator()() const {
	return InverseElement<Semigroup<typename R::type::Add> >()(NeutralElement<Semigroup<typename R::type::Mul> >()());
    }
};

template<typename R>
struct Characteristic<Root, R, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const {
	return ElementOrder<Semigroup<typename R::type::Add> >()(NeutralElement<Semigroup<typename R::type::Mul> >()());
    }
};

template<typename R_base, typename R, typename TC> struct MinusOne :
    public MinusOne<
	typename Base<R_base, MinusOne<Root> >::type,
	R,
	typename TypeConversion<R_base, MinusOne<Root> >::type
    >
{};

template<typename R_base, typename R, typename TC> struct Characteristic :
    public Characteristic<
	typename Base<R_base, Characteristic<Root> >::type,
	R,
	typename TypeConversion<R_base, Characteristic<Root> >::type
    >
{};

//functions
template<typename S>
Element<S> minusOne() {
    return MinusOne<RingWithUnity<Add<S>, Mul<S> > >()();
}

template<typename S>
unsigned int characteristic() {
    return Characteristic<RingWithUnity<Add<S>, Mul<S> > >()();
}

template<typename S>
bool isZeroDivisor(Element<S> const & a) {
    return IsZeroDivisor<Ring<Add<S>, Mul<S> > >()(a);
}

template<typename S>
unsigned int nilpotencyIndex(Element<S> const & a) {
    return NilpotencyIndex<Ring<Add<S>, Mul<S> > >()(a);
}

template<typename R>
struct TestDistributivity {
    typedef typename R::type::BaseSet R_BS_;
    typedef typename R::type::Add R_Add_;
    typedef typename R::type::Mul R_Mul_;
    bool operator()() const {
	for(Iterator<R_BS_> it1 = Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(it1); ++it1) {
	    for(Iterator<R_BS_> it2 = Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(it2); ++it2) {
		for(Iterator<R_BS_> it3 = Begin<R_BS_>()(); !IsAtEnd<R_BS_>()(it3); ++it3) {
		    if(R_Mul_()(*it1, R_Add_()(*it2, *it3)) != R_Add_()(R_Mul_()(*it1, *it2), R_Mul_()(*it1,*it3))) {
			std::cout << "Left Distributivity failed for " << (*it1) << ", " << (*it2) << ", " << (*it3) << std::endl;
			return false;
		    }
		    if(R_Mul_()(R_Add_()(*it1, *it2),*it3) != R_Add_()(R_Mul_()(*it1, *it3), R_Mul_()(*it2,*it3))) {
			std::cout << "Right Distributivity failed for " << (*it1) << ", " << (*it2) << ", " << (*it3) << std::endl;
			return false;
		    }
		}
	    }
	}
	return true;
    }
};

template<typename R>
struct TestRingAxioms {
    typedef Semigroup<typename R::type::Add> AddSG;
    typedef Semigroup<typename R::type::Mul> MulSG;
    bool operator()() const {
	if(!TestAssociativity<AddSG>()()) return false;
	if(!TestCommutativity<AddSG>()()) return false;
	if(!TestNeutralElement<AddSG>()()) return false;
	if(!TestInverseElements<AddSG>()()) return false;
	if(!TestAssociativity<MulSG>()()) return false;
	if(!TestNeutralElement<MulSG>()()) return false;
	if(!TestDistributivity<R>()()) return false;
	return true;
    }
};

#endif
