/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _SEMIGROUP_HPP
#define _SEMIGROUP_HPP

//general binary inner operation on a set S:
//consisting of
//base stucture S
//operation BinaryOperation<S>
//Elements Element<S>
//
#include "root.hpp"

template<typename Op> struct Semigroup_Atom;
//traits:
template<typename SG_base, typename SG = SG_base> struct HasNeutralElement;
template<typename SG_base, typename SG = SG_base> struct HasInverseElements;
template<typename SG_base, typename SG = SG_base> struct IsCommutative;
//metafunctions:
template<typename Op> struct Semigroup;
template<typename SG> struct Opposite; 
template<typename SG> struct LeftAction;
template<typename SG> struct RightAction;
//function objects:
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct OppositeOperator;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct NeutralElement;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct EqualsNeutralElement;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct IsInvertibleElement;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct ElementOrder;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct InverseElement;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct Power;
template<typename SG_base, typename SG = SG_base, typename TC = TypeConversionDefault> struct Generators;
//here templated typedefs could be used:
template<typename SG> struct InRightPlace;

#include <iostream>
#include <boost/bind.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/assert.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/type_traits/is_same.hpp>
#include "mathHelper.hpp"
#include "algorithm.hpp"
#include "set.hpp"
#include "map.hpp"
#include "action.hpp"

template<typename SG> struct InLeftPlace;

template<typename Op> struct Semigroup_Atom {
    BOOST_MPL_ASSERT_RELATION((Op::NumberOfArguments::value), ==, 2);
    BOOST_MPL_ASSERT((boost::is_same<typename Op::ArgumentType1, typename Op::ResultType>));
    BOOST_MPL_ASSERT((boost::is_same<typename Op::ArgumentType2, typename Op::ResultType>));
    typedef Op Operator;
    typedef typename Op::ResultType BaseSet;
};

//traits:
template<typename SG>
struct HasNeutralElement<Root, SG> : public boost::mpl::bool_<false> {};
//false means: undefinied
//true means: yes

template<typename SG>
struct HasInverseElements<Root, SG> : public boost::mpl::bool_<false> {};
//false means: undefinied
//true means: yes

template<typename SG>
struct IsCommutative<Root, SG> : public boost::mpl::bool_<false> {};
//false means: undefinied
//true means: yes

template<typename Op>
struct Semigroup {
    typedef Semigroup_Atom<Op> type;
};

template<typename SG_base, typename SG, bool IsComm>
struct OppositeOperator_impl;

template<typename SG_base, typename SG>
struct OppositeOperator_impl<SG_base, SG, true> : public SG::type::Operator {};

template<typename SG_base, typename SG>
struct OppositeOperator_impl<SG_base, SG, false> : public Map<typename SG::type::Operator::ArgumentTypes, typename SG::type::Operator::ResultType> {
    private:
	typedef typename SG::type::BaseSet S_;
    public:
	Element<S_> operator()(Element<S_> const & a, Element<S_> const & b) const {
	    return typename SG::type::Operator()(a,b);
	}
};

template<typename SG_base, typename SG>
class OppositeOperator<SG_base, SG, TypeConversionByImplementation> : public OppositeOperator_impl<SG_base, SG, IsCommutative<SG>::value> {};

template<typename SG>
struct Opposite {
    typedef Semigroup_Atom<OppositeOperator<SG> > type;
};

//Left operation fixes the right parameter and thus gives an action of the scalar structure on itself.
//For example: The left multiplication in a group gives a group operation of the group on itself.
//If the binary operation is associative, the resulting action is left associative.

template<typename SG>
struct LeftAction {
    typedef Action_Atom<typename SG::type::Operator, typename SG::type::Operator> type;
};

//If the binary operation is associative, the resulting 'RightAction' is left associative.
template<typename SG>
struct RightAction {
    typedef Action_Atom<typename SG::type::Operator, typename Opposite<SG>::type::Operator> type;
};

template<typename SG>
class NeutralElement<Root, SG, TypeConversionByImplementation> {
    private:
	typedef typename SG::type::Operator Op_;
	typedef typename SG::type::BaseSet S_;
    public:
	template <typename Args> struct sig {
	    typedef Element<S_> type;
	};
	Element<S_> operator()() const;
};

template<typename SG>
class EqualsNeutralElement<Root, SG, TypeConversionByImplementation> {
    private:
	typedef typename SG::type::Operator Op_;
	typedef typename SG::type::BaseSet S_;
    public:
	typedef bool result_type;
	template <typename Args> struct sig {
	    typedef bool type;
	};
	bool operator()(Element<S_> const & a) const ;
};

template<typename SG>
class IsInvertibleElement<Root, SG, TypeConversionByImplementation> {
    private:
	typedef typename SG::type::Operator Op_;
	typedef typename SG::type::BaseSet S_;
    public:
	template <typename Args> struct sig {
	    typedef bool type;
	};
	bool operator()(Element<S_> const & a) const {
	    return repeatedApplicationReaches(
		a,
		boost::bind(Op_(), _1, a),
		EqualsNeutralElement<SG>(),
		typename HasInverseElements<SG>::type()
	    );
	}
};

template<typename SG>
struct ElementOrder<Root, SG, TypeConversionByImplementation> {
    private:
	typedef typename SG::type::Operator Op_;
	typedef typename SG::type::BaseSet S_;
    public:
	template <typename Args> struct sig {
	    typedef unsigned int type;
	};
	unsigned int operator()(Element<S_> const & a) const { 
	    return repeatedApplicationIndex(
		a,
		boost::bind(Op_(), _1, a),
		EqualsNeutralElement<SG>(),
		typename HasInverseElements<SG>::type(),
		typename HasInverseElements<SG>::type()
	    );
	}
};

template<typename SG>
struct InverseElement<Root, SG, TypeConversionByImplementation> {
//should work for all invertible Elements
    private:
	typedef typename SG::type::Operator Op_;
	typedef typename SG::type::BaseSet S_;
    public:
	template <typename Args> struct sig {
	    typedef Element<S_> type;
	};
	Element<S_> operator()(Element<S_> const & a) const {
	    BOOST_MPL_ASSERT((HasNeutralElement<SG>));
	    Element<S_> b = NeutralElement<SG>()();
	    for(Element<S_> c = a; ! EqualsNeutralElement<SG>()(c); c = Op_()(c,a)) {
		b = Op_()(b,a);
	    }
	    return b;
	}
};

template<typename SG>
struct Power<Root, SG, TypeConversionByImplementation> { //binary exponentation
    typedef typename SG::type::Operator Op_;
    typedef typename SG::type::BaseSet S_;
    Element<S_> operator()(Element<S_> const & a, unsigned int const expIn) const {
	BOOST_MPL_ASSERT((HasNeutralElement<SG>)); //liesse sich auch allgemein programmieren.
	if (expIn == 0) return NeutralElement<SG>()();
	if (expIn == 1) return a;
	Element<S_> result = NeutralElement<SG>()();
	Element<S_> pow = a;
	unsigned int exp = expIn;
Power_loop:
	if (exp & 1) {
	    InRightPlace<SG>()(pow, result);
	}
	exp >>= 1;
	if (exp == 0) {
	    return result;
	} else {
	    InRightPlace<SG>()(pow,pow);
	    goto Power_loop;
	}
    }
};

template<typename SG>
struct Generators<Root, SG, TypeConversionByImplementation> { //binary exponentation
    typedef typename SG::type::Operator Op_;
    typedef typename SG::type::BaseSet S_;
    std::vector<Element<S_> > operator()() const {
    }
};

template<typename SG_base, typename SG> struct HasNeutralElement :
    public HasNeutralElement<
	typename Base<SG_base, HasNeutralElement<Root> >::type,
	SG
    >
{};

template<typename SG_base, typename SG> struct HasInverseElements :
    public HasInverseElements<
	typename Base<SG_base, HasInverseElements<Root> >::type,
	SG
    >
{};

template<typename SG_base, typename SG> struct IsCommutative :
    public IsCommutative<
	typename Base<SG_base, IsCommutative<Root> >::type,
	SG
    >
{};

/*
template<typename Op_base, typename Op, typename TC> struct LeftAction :
    public LeftAction<
	typename Base<Op_base, LeftAction<Root> >::type,
	Op,
	typename TypeConversion<Op, LeftAction<Root> >::type
    >
{};

template<typename Op_base, typename Op, typename TC> struct RightAction :
    public RightAction<
	typename Base<Op_base, RightAction<Op> >::type,
	Op,
	typename TypeConversion<Op, RightAction<Root> >::type
    >
{};
*/

template<typename SG_base, typename SG, typename TC> struct NeutralElement :
    public NeutralElement<
	typename Base<SG_base, NeutralElement<Root> >::type,
	SG,
	typename TypeConversion<SG_base, NeutralElement<Root> >::type
    >
{};

template<typename SG_base, typename SG, typename TC> struct EqualsNeutralElement :
    public EqualsNeutralElement<
	typename Base<SG_base, EqualsNeutralElement<Root> >::type,
	SG,
	typename TypeConversion<SG_base, EqualsNeutralElement<Root> >::type
    >
{};

template<typename SG_base, typename SG, typename TC> struct IsInvertibleElement :
    public IsInvertibleElement<
	typename Base<SG_base, IsInvertibleElement<Root> >::type,
	SG,
	typename TypeConversion<SG_base, IsInvertibleElement<Root> >::type
    >
{};

template<typename SG_base, typename SG, typename TC> struct ElementOrder :
    public ElementOrder<
	typename Base<SG_base, ElementOrder<Root> >::type,
	SG,
	typename TypeConversion<SG_base, ElementOrder<Root> >::type
    >
{};

template<typename SG_base, typename SG, typename TC> struct InverseElement :
    public InverseElement<
	typename Base<SG_base, InverseElement<Root> >::type,
	SG,
	typename TypeConversion<SG_base, InverseElement<Root> >::type
    >
{};

template<typename SG_base, typename SG, typename TC> struct Power :
    public Power<
	typename Base<SG_base, Power<Root> >::type,
	SG,
	typename TypeConversion<SG_base, Power<Root> >::type
    >
{};

template<typename SG_base, typename SG, typename TC> struct Generators :
    public Generators<
	typename Base<SG_base, Generators<Root> >::type,
	SG,
	typename TypeConversion<SG_base, Generators<Root> >::type
    >
{};

//IMPORTANT:
//Instead of writing a specialization to InRightPlace, write a specialization to ActionInPlace<LeftAction>.
//Instead of writing a specialization to InLeftPlace, write a specialization to ActionInPlace<RightAction>.
//For this reason, we do not provide InRightPlace<Root, ...> or InLeftPlace<Root, ...>.
//See coordinateWiseBinaryOperation.hpp for an example.

//If you have the choice, always use 'InRightPlace' and not 'InLeftPlace'.
//The reason is that 'InRightPlace' is the 'ActionInPlace' of 'LeftAction',
//and whenever no symmetric treatment is possible,
//syntax and implementations are biased to and optimized for 'LeftAction'.

template<typename SG> struct InRightPlace : public ActionInPlace<LeftAction<SG> > {};
template<typename SG> struct InLeftPlace : public ActionInPlace<RightAction<SG> > {};

template<typename SG, typename Derived> struct IsLeftAssociative<LeftAction<SG>, Derived> : public boost::mpl::bool_<true> {};
template<typename SG, typename Derived> struct IsRightAssociative<LeftAction<SG>, Derived> : public boost::mpl::bool_<true> {};

template<typename Op>
unsigned int elementOrder(Element<typename Op::BaseSet> const & a) {
    return ElementOrder<Op>()(a);
}

template<typename SG>
struct Conjugation_Operator : public Map<boost::mpl::vector<typename SG::type::BaseSet, typename SG::type::BaseSet>, typename SG::type::BaseSet> {
    Element<typename SG::type::BaseSet> operator()(Element<typename SG::type::BaseSet> const & a, Element<typename SG::type::BaseSet> const & b) {
	return typename SG::type::Operator()(typename SG::type::Operator()(a,b),InverseElement<SG>()(a));
    }
};

template<typename SG>
struct ConjugationAction { 
    typedef Action_Atom<typename SG::type::Operator, Conjugation_Operator<SG> > type;
};

template<typename SG>
struct TestAssociativity {
    typedef typename SG::type::BaseSet SG_BS_;
    typedef typename SG::type::Operator SG_OP_;
    bool operator()() {
	for(Iterator<SG_BS_> it1 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it1); ++it1) {
	    for(Iterator<SG_BS_> it2 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it2); ++it2) {
		for(Iterator<SG_BS_> it3 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it3); ++it3) {
		    if (SG_OP_()(*it1,SG_OP_()(*it2,*it3)) != SG_OP_()(SG_OP_()(*it1, *it2), *it3)) {
			std::cout << "Associativity failed for " << (*it1) << ", " << (*it2) << ", " << (*it3) << std::endl;
			return false;
		    }
		}
	    }
	}
	return true;
    }
};

template<typename SG>
struct TestNeutralElement {
    typedef typename SG::type::BaseSet SG_BS_;
    typedef typename SG::type::Operator SG_OP_;
    bool operator()() {
	Element<SG_BS_> e = NeutralElement<SG>()();
	for(Iterator<SG_BS_> it1 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it1); ++it1) {
	    if ((SG_OP_()(*it1, e) != (*it1)) || SG_OP_()(e,*it1) != (*it1)) {
		std::cout << "Neutral element failed for " << (*it1) << std::endl;
		return false;
	    }
	}
	return true;
    }
};

template<typename SG>
struct TestInverseElements {
    typedef typename SG::type::BaseSet SG_BS_;
    typedef typename SG::type::Operator SG_OP_;
    bool operator()() {
	Element<SG_BS_> e = NeutralElement<SG>()();
	for(Iterator<SG_BS_> it1 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it1); ++it1) {
	    Element<SG_BS_> inv = InverseElement<SG>()(*it1);
	    if ((SG_OP_()(*it1, inv) != e) || (SG_OP_()(inv,*it1) != e)) {
		std::cout << "Inverse element failed for " << (*it1) << std::endl;
		return false;
	    }
	}
	return true;
    }
};

template<typename SG>
struct TestCommutativity {
    typedef typename SG::type::BaseSet SG_BS_;
    typedef typename SG::type::Operator SG_OP_;
    bool operator()() {
	for(Iterator<SG_BS_> it1 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it1); ++it1) {
	    for(Iterator<SG_BS_> it2 = Begin<SG_BS_>()(); !IsAtEnd<SG_BS_>()(it2); ++it2) {
		if (SG_OP_()(*it1,*it2) != SG_OP_()(*it2, *it1)) {
		    std::cout << "Commutativity failed for " << (*it1) << ", " << (*it2) << std::endl;
		    return false;
		}
	    }
	}
	return true;
    }
};

#endif
