/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _SET_HPP
#define _SET_HPP

// implements a static (known at compilation time) mathematical structure S.
// the elements of this structure have type Element<S>.
// Thus, the parent structure of each element is determined by its type,
// and we do not need to store pointers to the parent structure.

#include <iostream>
#include <sstream>
#include <string>
#include <boost/mpl/bool.hpp>
#include <boost/static_assert.hpp>
#include "root.hpp"

template<typename S_base> struct Element;
//'Element<...>' usually has several constructors which must be forwarded manually in derived classes.
//For this reason, we don't provide automatic inheritance here and 'Element' only takes one template parameter.
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Equals;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct NameStream;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Name;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct ElementName;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct HashValue;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct RankElement;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct UnrankElement;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Compare;
template<typename S_base, typename S = S_base> struct Iterator;
//'Iterator' should be implemented using boost::iterator_facade which does not fit into out TypeConversion-concept.
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Begin;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct End;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct IsAtBegin;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct IsAtEnd;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Cardinality;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Representative;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Random;

template<typename S>
struct Element_root {
    typedef S BaseSet;
};

template<typename S>
struct Equals<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(Element<S> const &a, Element<S> const &b) const;
    //{
//	BOOST_STATIC_ASSERT((boost::mpl::bool_<false>::value));
 //   }
};

template<typename S, typename Derived>
struct Equals<S, Derived, TypeConversionByCast> : public Equals<Root, S> {
    bool operator()(Element<Derived> const & a, Element<Derived> const & b) const {
	return Equals<S>()((Element<S>) a, (Element<S>) b);
    }
};

//For Types S it is not possible to write an operator<< for things like 'std::cout << S << std::endl'.
//So we provide NameStream with possibly better performance, but uglier syntax 'NameStream<S>()(std::cout) << std::endl';
//and Name with the nicer syntax 'std::cout << Name<S>()() << std::endl'
//Only NameStream needs to be written for specific sets.
template<typename S>
struct NameStream<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
        typedef std::ostream& type;
    };
    std::ostream& operator()(std::ostream& o) const;
};

template<typename S>
struct Name<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef std::string type;
    };
    std::string operator()() const {
	std::ostringstream oss;
	NameStream<S>()(oss);
	return oss.str();
    }
};

//Here the situation is different. For an object 'a' of type Element<S> the
//syntax 'std::cout << a' is perfectly legal.
template<typename S>
struct ElementName<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef std::ostream& type;
    };
    std::ostream& operator()(std::ostream& o, Element<S> const &a) const;
};

//for hashing:
template<typename S>
struct HashValue<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef size_t type;
    };
    size_t operator()(Element<S> const & a) const {
	return RankElement<S>()(a);
    };
};

//a total ordering (for example for inserting in trees):
template<typename S>
struct Compare<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(Element<S> const & a, Element<S> const & b) const;
};

template<typename S>
struct RankElement<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(Element<S> const & a) const;
};

template<typename S>
struct UnrankElement<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef Element<S> type;
    };
    Element<S> operator()(unsigned int const & n) const;
};

template<typename S>
class Iterator<Root, S> {};
//should be implemented using boost::iterator_facade

template<typename S>
struct Begin<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef Iterator<S> type;
    };
    Iterator<S> operator()() const;
};

template<typename S>
struct End<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef Iterator<S> type;
    };
    Iterator<S> operator()() const;
};

template<typename S>
struct IsAtBegin<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(Iterator<S> const &) const; 
};

template<typename S>
struct IsAtEnd<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(Iterator<S> const &) const; 
};

template<typename S>
struct Cardinality<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()() const {
	//standard implementation: count all the elements by iterating over the structure
	unsigned int result = 0;
	for(Iterator<S> it = Begin<S>()(); !IsAtEnd<S>()(it); ++it) {
	    ++result;
	}
	return result;
    }
};

template<typename S>
struct Random<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef Element<S> type;
    };
    Element<S> operator()() const {
	return UnrankElement<S>()(rand() % Cardinality<S>()());
    }
};

template<typename S>
struct Representative<Root, S, TypeConversionByImplementation> {
    template <typename Args> struct sig {
	typedef Element<S> type;
    };
    Element<S> operator()() const;
};

template<typename S> struct Element : public Element_root<S> {};

template<typename S_base, typename S, typename TC> struct Equals :
    public Equals<
	typename Base<S_base, Equals<Root> >::type,
	S,
	typename TypeConversion<S_base, Equals<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct NameStream :
    public NameStream<
	typename Base<S_base, NameStream<Root> >::type,
	S,
	typename TypeConversion<S_base, NameStream<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Name :
    public Name<
	typename Base<S_base, Name<Root> >::type,
	S,
	typename TypeConversion<S_base, Name<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct ElementName :
    public ElementName<
	typename Base<S_base, ElementName<Root> >::type,
	S,
	typename TypeConversion<S_base, ElementName<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct HashValue :
    public HashValue<
	typename Base<S_base, HashValue<Root> >::type,
	S,
	typename TypeConversion<S_base, HashValue<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct RankElement :
    public RankElement<
	typename Base<S_base, RankElement<Root> >::type,
	S,
	typename TypeConversion<S_base, RankElement<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct UnrankElement :
    public UnrankElement<
	typename Base<S_base, UnrankElement<Root> >::type,
	S,
	typename TypeConversion<S_base, UnrankElement<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Compare :
    public Compare<
	typename Base<S_base, Compare<Root> >::type,
	S,
	typename TypeConversion<S_base, Compare<Root> >::type
    >
{};

template<typename S_base, typename S> struct Iterator :
    public Iterator<
	typename Base<S_base, Iterator<Root> >::type,
	S
    >
{};

template<typename S_base, typename S, typename TC> struct Begin :
    public Begin<
	typename Base<S_base, Begin<Root> >::type,
	S,
	typename TypeConversion<S_base,Begin<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct End :
    public End<
	typename Base<S_base, End<Root> >::type,
	S,
	typename TypeConversion<S_base, End<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct IsAtBegin :
    public IsAtBegin<
	typename Base<S_base, IsAtBegin<Root> >::type,
	S,
	typename TypeConversion<S_base, IsAtBegin<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct IsAtEnd :
    public IsAtEnd<
	typename Base<S_base, IsAtEnd<Root> >::type,
	S,
	typename TypeConversion<S_base, IsAtEnd<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Cardinality :
    public Cardinality<
	typename Base<S_base, Cardinality<Root> >::type,
	S,
	typename TypeConversion<S_base, Cardinality<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Random :
    public Random<
	typename Base<S_base, Random<Root> >::type,
	S,
	typename TypeConversion<S_base, Random<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Representative :
    public Representative<
	typename Base<S_base, Representative<Root> >::type,
	S,
	typename TypeConversion<S_base, Representative<Root> >::type
    >
{};

//functions
template<typename S>
bool operator==(Element<S> const &a, Element<S> const &b) {
    return Equals<S>()(a,b);
}

template<typename S>
bool operator!=(Element<S> const &a, Element<S> const &b) {
    return !Equals<S>()(a,b);
}

template<typename S>
std::ostream& operator<<(std::ostream & o, Element<S> const &a) {
    return ElementName<S>()(o,a);
}

template<typename S>
unsigned int order() {
    return Cardinality<S>()();
}

template<typename S>
std::ostream& name(std::ostream & o) {
    return Name<S>()(o);
}

template<typename S>
Element<S> random() {
    return Random<S>()();
}
#endif
