/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _TABLE_CHAINRING_HPP
#define _TABLE_CHAINRING_HPP

//TableChainRing: chain ring with addition and multiplication read from a file
template<typename id> class TableChainRing_BaseSet;
template<typename id> struct TableChainRing_Add;
template<typename id> struct TableChainRing_Mul;
template<typename id> class TableChainRing;


//template<typename id> class Element<TableChainRing_BaseSet<id> >;
//template<typename id> class AutomorphismGroup<TableChainRing<id> >;
//template<typename id> class Element<AutomorphismGroup<TableChainRing<id> > >;

#include <iostream>
#include <fstream>
#include <cassert>
#include <string>
#include <vector>
#include <map>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include "mathHelper.hpp"
#include "chainRing.hpp"
#include "numberRangeSet.hpp"
#include "automorphismGroup.hpp"
#include "action.hpp"
#include "semigroup.hpp"
#include "set.hpp"
#include "root.hpp"
#include "map.hpp"
#include "permutationGroup.hpp"
#include "semigroupByGenerators.hpp"
#ifdef DB_SUPPORT
    #include "Datenbank/db_tableChainRing.h"
#endif

template<typename id>
class TableChainRing : public RingWithUnity<TableChainRing_Add<id>, TableChainRing_Mul<id> > {
    public:
	typedef TableChainRing_BaseSet<id> R_BS;
	typedef std::vector<Element<R_BS> > ElVec;
	typedef typename ElVec::iterator ElVecIter;
	typedef std::vector<std::string> StrVec;
	typedef PermutationGroup<R_BS> R_PG;
	typedef typename R_PG::type::BaseSet R_PG_BS;
	typedef std::vector<Element<R_PG_BS> > AutVec;

    private:
	static unsigned int basePrime;
	static unsigned int rank;
	static unsigned int chainLength;
	static unsigned int residueFieldOrder;
	//order is stored in TableChainRing_BaseSet
	static std::string name;
	static StrVec elementName;
	static ElVec addTable;
	static ElVec mulTable;
	static ElVec rightInvTable;

	static void init_right_inv_table();

    public:
	template<typename CR> static void initFromChainRing();

	#ifdef DB_SUPPORT
	    static bool initFromDB(unsigned int dbid);
	#endif
	static void init(std::istream & is); 
	static void init(std::string const filename);

	static void reduceToFather();

	static unsigned int getBasePrime();
	static unsigned int getRank(); 
	static unsigned int getChainLength();
	static unsigned int getResidueFieldOrder(); 
	static Element<TableChainRing_BaseSet<id> > getAddTable(unsigned int const a, unsigned int const b); 
	static Element<TableChainRing_BaseSet<id> > getMulTable(unsigned int const a, unsigned int const b); 
	static Element<TableChainRing_BaseSet<id> > getInvTable(unsigned int const a); 
	static std::string getElementName(unsigned int const a); 
	static std::string getName(); 
};

template<typename id> unsigned int TableChainRing<id>::getBasePrime() {
    return basePrime;
}

template<typename id> unsigned int TableChainRing<id>::getRank() {
    return rank;
}

template<typename id> unsigned int TableChainRing<id>::getChainLength() {
    return chainLength;
}

template<typename id> unsigned int TableChainRing<id>::getResidueFieldOrder() {
    return residueFieldOrder;
}

template<typename id> Element<TableChainRing_BaseSet<id> > TableChainRing<id>::getAddTable(unsigned int const a, unsigned int const b) {
    return addTable[a * Cardinality<TableChainRing_BaseSet<id> >()() + b];
}

template<typename id> Element<TableChainRing_BaseSet<id> > TableChainRing<id>::getMulTable(unsigned int const a, unsigned int const b) {
    return mulTable[a * Cardinality<TableChainRing_BaseSet<id> >()() + b];
}

template<typename id> Element<TableChainRing_BaseSet<id> > TableChainRing<id>::getInvTable(unsigned int const a) {
    return rightInvTable[a];
}

template<typename id> std::string TableChainRing<id>::getElementName(unsigned int const a) {
    return elementName[a];
}

template<typename id> std::string TableChainRing<id>::getName() {
    return name;
}

template<typename id>
unsigned int TableChainRing<id>::basePrime;

template<typename id>
unsigned int TableChainRing<id>::rank;

template<typename id>
unsigned int TableChainRing<id>::residueFieldOrder;

template<typename id>
unsigned int TableChainRing<id>::chainLength;

template<typename id>
std::string TableChainRing<id>::name;

template<typename id>
typename TableChainRing<id>::StrVec TableChainRing<id>::elementName;

template<typename id>
typename TableChainRing<id>::ElVec TableChainRing<id>::addTable;

template<typename id>
typename TableChainRing<id>::ElVec TableChainRing<id>::mulTable;

template<typename id>
typename TableChainRing<id>::ElVec TableChainRing<id>::rightInvTable;

template<typename id>
class TableChainRing_BaseSet : public NumberRangeSet<ValueWrapperByTemplateParameter<unsigned int, 0>, ValueWrapperByInitFunction<unsigned int, id> >
{
    public:
	typedef TableChainRing<id> BaseTableChainRing;
};

template<typename id>
struct Base<TableChainRing_BaseSet<id> > {
    typedef NumberRangeSet<ValueWrapperByTemplateParameter<unsigned int, 0>, ValueWrapperByInitFunction<unsigned int, id> > type;
};

template<typename id>
class Element<TableChainRing_BaseSet<id> > : public Element<NumberRangeSet<ValueWrapperByTemplateParameter<unsigned int, 0>, ValueWrapperByInitFunction<unsigned int, id> > > {
    private:
	typedef NumberRangeSet<ValueWrapperByTemplateParameter<unsigned int, 0>, ValueWrapperByInitFunction<unsigned int, id> > Parent_;

    public:
        Element() : Element<Parent_>() {};
        Element(unsigned int valIn) : Element<Parent_>(valIn) {};
};

// set.hpp
template<typename id, typename Derived>
struct ElementName<TableChainRing_BaseSet<id>, Derived> : public ElementName<Root, Derived> {
    typedef Derived R_;
    std::ostream & operator()(std::ostream & o, Element<R_> const & a) const {
	return o << TableChainRing<id>::getElementName(a.value());
    }
};

template<typename id, typename Derived>
struct NameStream<TableChainRing_BaseSet<id>, Derived> : public NameStream<Root, Derived> {
    std::ostream & operator()(std::ostream & o) const {
	return o << TableChainRing<id>::getName();
    }
};

//because of the constructors, Iterator cannot be generated automatically.
template<typename id, typename Derived>
class Iterator<TableChainRing_BaseSet<id>, Derived> : public Iterator<typename Base<TableChainRing_BaseSet<id>, Iterator<Root> >::type, Derived> {
    private:
        typedef Iterator<typename Base<TableChainRing_BaseSet<id>, Iterator<Root> >::type, Derived> Parent_;
    public:
        Iterator() : Parent_() {};
        Iterator(unsigned int const & val) : Parent_(val) {};
};

//additiveSemigroup
template<typename id>
struct TableChainRing_Add : public Map<boost::mpl::vector<TableChainRing_BaseSet<id>, TableChainRing_BaseSet<id> >,TableChainRing_BaseSet<id> > {
    typedef TableChainRing_BaseSet<id> CRS_;
    typedef TableChainRing<id> CR_;
    Element<CRS_> operator()(Element<CRS_> const & a, Element<CRS_> const & b) const {
	return CR_::getAddTable(a.value(),b.value());
    }
};

template<typename id, typename Derived>
struct HasNeutralElement<Semigroup<TableChainRing_Add<id> >, Derived> : public boost::mpl::bool_<true> {};

template<typename id, typename Derived>
struct HasInverseElements<Semigroup<TableChainRing_Add<id> >, Derived> : public boost::mpl::bool_<true> {};

template<typename id, typename Derived>
struct IsCommutative<Semigroup<TableChainRing_Add<id> >, Derived> : public boost::mpl::bool_<true> {};

template<typename id, typename Derived>
struct NeutralElement<Semigroup<TableChainRing_Add<id> >, Derived> : public NeutralElement<Root, Derived> {
    typedef typename Derived::type::BaseSet CRS_;
    Element<CRS_> operator()() const {
	return Element<CRS_>(0);
    }
};

template<typename id, typename Derived>
struct EqualsNeutralElement<Semigroup<TableChainRing_Add<id> >, Derived> : public EqualsNeutralElement<Root, Derived> {
    typedef typename Derived::type::BaseSet CRS_;
    bool operator()(Element<CRS_> const & a) const {
	return (a.value() == 0);
    }
};

template<typename id>
struct Add<TableChainRing_BaseSet<id> > : public TableChainRing_Add<id> {};

template<typename id>
struct Base<Semigroup<Add<TableChainRing_BaseSet<id> > > >{
    typedef Semigroup<TableChainRing_Add<id> > type;
};

//multiplicativeSemigroup
template<typename id>
struct TableChainRing_Mul : public Map<boost::mpl::vector<TableChainRing_BaseSet<id>, TableChainRing_BaseSet<id> >,TableChainRing_BaseSet<id> > {
    typedef TableChainRing_BaseSet<id> CRS_;
    typedef TableChainRing<id> CR_;
    Element<CRS_> operator()(Element<CRS_> const & a, Element<CRS_> const & b) const {
	return CR_::getMulTable(a.value(),b.value());
    }
};

template<typename id, typename Derived>
struct HasNeutralElement<Semigroup<TableChainRing_Mul<id> >, Derived> : public boost::mpl::bool_<true> {};

template<typename id, typename Derived>
struct NeutralElement<Semigroup<TableChainRing_Mul<id> >, Derived> : public NeutralElement<Root, Derived> {
    typedef typename Derived::type::BaseSet CRS_;
    Element<CRS_> operator()() const {
	return Element<CRS_>(1);
    }
};

template<typename id, typename Derived>
struct EqualsNeutralElement<Semigroup<TableChainRing_Mul<id> >, Derived> : public EqualsNeutralElement<Root, Derived> {
    typedef typename Derived::type::BaseSet CRS_;
    bool operator()(Element<CRS_> const &a) const {
	return (a.value() == 1);
    }
};

template<typename id, typename Derived>
struct InverseElement<Semigroup<TableChainRing_Mul<id> >, Derived> : public InverseElement<Root, Derived> {
    typedef typename Derived::type::BaseSet CRS_;
    typedef typename CRS_::BaseTableChainRing CR_;
    Element<CRS_> operator()(Element<CRS_> const & a) const {
	return GeneralizedRightInverse<CR_>()(a);
    }
};

template<typename id>
struct Mul<TableChainRing_BaseSet<id> > : public TableChainRing_Mul<id> {};

template<typename id>
struct Base<Semigroup<Mul<TableChainRing_BaseSet<id> > > >{
    typedef Semigroup<TableChainRing_Mul<id> > type;
};

//chainRing.hpp
template<typename id, typename Derived>
struct IsChainRing<TableChainRing<id>, Derived> : public boost::mpl::bool_<true> {};

template<typename id, typename Derived>
struct BasePrime<TableChainRing<id>, Derived> : public BasePrime<Root, Derived> {
    unsigned int operator()() const {
	return TableChainRing<id>::getBasePrime();
    }
};

template<typename id, typename Derived>
struct Rank<TableChainRing<id>, Derived> : public Rank<Root, Derived> {
    unsigned int operator()() const {
	return TableChainRing<id>::getRank();
    }
};

template<typename id, typename Derived>
struct ResidueFieldOrder<TableChainRing<id>, Derived> : public ResidueFieldOrder<Root, Derived> {
    unsigned int operator()() const {
	return TableChainRing<id>::getResidueFieldOrder();
    }
};

template<typename id, typename Derived>
struct ChainLength<TableChainRing<id>, Derived> : public ChainLength<Root, Derived> {
    unsigned int operator()() const {
	return TableChainRing<id>::getChainLength();
    }
};

template<typename id, typename Derived>
struct Theta<TableChainRing<id>, Derived> : public Theta<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> operator()() const {
	if (TableChainRing<id>::getChainLength() == 1) { //then R is a finite field
	    return Element<CRS_>(0);
	}
	return Element<CRS_>(TableChainRing<id>::getResidueFieldOrder());
    }
};

template<typename id, typename Derived>
struct Alpha<TableChainRing<id>, Derived> : public Alpha<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> operator()() const {
	if (TableChainRing<id>::getResidueFieldOrder() > 2){
	    return Element<CRS_>(2);
	} else {
	    return Element<CRS_>(1);
	}
    }
};

template<typename id, typename Derived>
struct Period<TableChainRing<id>, Derived> : public Period<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    unsigned int operator()(Element<CRS_> const & a) const {
	unsigned int res = ChainLength<CR_>()();
	unsigned int b = a.value();
	unsigned int q = ResidueFieldOrder<CR_>()();
	do {
	    if(b % q != 0) break;
	    b /= q;
	    --res;
	} while (res != 0);
	return res;
    }
};

template<typename id, typename Derived>
struct RightMultiplicationByThetaPower<TableChainRing<id>, Derived> : public RightMultiplicationByThetaPower<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> operator()(Element<CRS_> const & a, unsigned int const p) const {
	return Element<CRS_>((a.value() * mathHelper::pow(ResidueFieldOrder<CR_>()(),p)) % Cardinality<CRS_>()());
    }
};

template<typename id, typename Derived>
struct RightDivisionByThetaPower<TableChainRing<id>, Derived> : public RightDivisionByThetaPower<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> operator()(Element<CRS_> const & a, unsigned int const p) const {
	return Element<CRS_>(a.value()/mathHelper::pow(ResidueFieldOrder<CR_>()(),p));
    }
};

template<typename id, typename Derived>
struct GeneralizedRightInverse<TableChainRing<id>, Derived> : public GeneralizedRightInverse<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> operator()(Element<CRS_> const & a) const {
	return CR_::getInvTable(a.value());
    }
};

template<typename id, typename Derived>
struct RepresentativeModThetaPower<TableChainRing<id>, Derived> : public RepresentativeModThetaPower<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> operator()(Element<CRS_> const & a, unsigned int const p) const {
	return Element<CRS_>(a.value() % mathHelper::pow(ResidueFieldOrder<CR_>()(),p));
    }
};

template<typename id, typename Derived>
struct InPlaceRightMultiplicationByThetaPower<TableChainRing<id>, Derived> : public InPlaceRightMultiplicationByThetaPower<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> & operator()(Element<CRS_> & a, unsigned int const p) const {
	a.setValue((a.value()*mathHelper::pow(ResidueFieldOrder<CR_>()(),p)) % Cardinality<CRS_>()());
	return a;
    }
};

template<typename id, typename Derived>
struct InPlaceRightDivisionByThetaPower<TableChainRing<id>, Derived> : public InPlaceRightDivisionByThetaPower<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> & operator()(Element<CRS_> & a, unsigned int const p) const {
	a.setValue(a.value()/mathHelper::pow(ResidueFieldOrder<CR_>()(),p));
	return a;
    }
};

template<typename id, typename Derived>
struct InPlaceRepresentativeModThetaPower<TableChainRing<id>, Derived> : public InPlaceRepresentativeModThetaPower<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> & operator()(Element<CRS_> & a, unsigned int const p) const {
	a.setValue(a.value()%mathHelper::pow(ResidueFieldOrder<CR_>()(),p));
	return a;
    }
};

template<typename id, typename Derived>
struct RightThetaAdicExpansion<TableChainRing<id>, Derived> : public RightThetaAdicExpansion<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    std::vector<unsigned int> operator()(Element<CRS_> const & a) const {
	unsigned int val=a.value();
	unsigned int counter=ChainLength<CR_>()();
	std::vector<unsigned int> res;
	unsigned int q=ResidueFieldOrder<CR_>()();
	while(val!=0){
	    unsigned int nextEntry=(val%q);
	    res.push_back(nextEntry);
	    val /= q;
	    counter--;
	}
	while(counter > 0){
	    res.push_back(0);
	    counter--;
	}
	return res;
    }
};

template<typename id, typename Derived>
struct CompressedRightThetaAdicExpansion<TableChainRing<id>, Derived> : public CompressedRightThetaAdicExpansion<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    unsigned int operator()(Element<CRS_> const & a) const {
	return RankElement<CRS_>()(a);
    }
};

template<typename id, typename Derived>
struct SetRightThetaAdicExpansion<TableChainRing<id>, Derived> : public SetRightThetaAdicExpansion<Root, Derived> {
    typedef Derived CR_;
    typedef typename CR_::type::BaseSet CRS_;
    Element<CRS_> & operator()(Element<CRS_> & a, std::vector<unsigned int> const & v) const {
	unsigned int val=a.value();
	unsigned int q=ResidueFieldOrder<CR_>()();
	unsigned int pow=mathHelper::pow(q,ChainLength<CR_>()()-v.size());
	val%=pow;
	std::vector<unsigned int>::const_iterator iter=v.begin();
	const std::vector<unsigned int>::const_iterator endIter=v.end();
	while(iter!=endIter){
	    val+=(*iter)*pow;
	    pow*=q;
	    iter++;
	}
	a=Element<CRS_>(val);
	return a;
    }
};

template<typename id> 
template<typename CR>
void TableChainRing<id>::initFromChainRing(){
    typedef typename CR::type::BaseSet CRBS;
    //set values
    basePrime=BasePrime<CR>()();
    rank=Rank<CR>()();
    chainLength=ChainLength<CR>()();
    residueFieldOrder=ResidueFieldOrder<CR>()();
    name=Name<CRBS>()();

    //run through CR in tae-order and set element names etc.
    std::vector<Element<CRBS> > elementWithNr;
    elementWithNr.resize(Cardinality<CRBS>()());
    std::map<Element<CRBS>, unsigned int, Compare<CRBS> > elementToNr;
    elementName.resize(Cardinality<CRBS>()());
    std::vector<unsigned int> tae(chainLength,0);
    Element<CRBS> el = Zero<CRBS>()();
    unsigned int counter=0;
    while(true) {
	SetRightThetaAdicExpansion<CR>()(el,tae);
	elementWithNr[counter]=el;
	std::stringstream ss;
	ss<<el;
	elementName[counter]=ss.str();
	elementToNr.insert(std::pair<Element<CRBS>,unsigned int>(el,counter));
	counter++;
	if (counter>=Cardinality<CRBS>()()){
	    break;
	}
	unsigned int qExponent=0;
	unsigned int counterCp=counter;
	while ((counterCp%residueFieldOrder)==0) {
	    ++qExponent;
	    counterCp/=residueFieldOrder;
	}
	++tae[qExponent];
	for (unsigned int j=0; j<qExponent; ++j) {
	    tae[j]=0;
	}
    }

    //set add- and multable
    addTable.resize(Cardinality<CRBS>()()*Cardinality<CRBS>()());
    mulTable.resize(Cardinality<CRBS>()()*Cardinality<CRBS>()());
    ElVecIter aIter = addTable.begin();
    ElVecIter mIter = mulTable.begin();
    for (unsigned int i=0; i<Cardinality<CRBS>()(); i++){
	for (unsigned int j=0; j<Cardinality<CRBS>()(); j++){
	    (*aIter).setValue(elementToNr.find(elementWithNr[i]+elementWithNr[j])->second);
	    (*mIter).setValue(elementToNr.find(elementWithNr[i]*elementWithNr[j])->second);
	    aIter++;
	    mIter++;
	}
    }
    TableChainRing_BaseSet<id>::initUpperBound(Cardinality<CRBS>()());

    R_PG_BS::init();
    typedef std::vector<Element<typename AutomorphismGroup<CR>::type::BaseSet> > OldAGVecT;
    
    /*
    OldAGVecT oldGenerators = Generators<AutomorphismGroup<CR> >()();
    AutVec autGenerators(oldGenerators.size());

    //initialize autGenerators
    unsigned int i = 0;
    for (typename OldAGVecT::const_iterator it = oldGenerators.begin(); it != oldGenerators.end(); ++it) {
	Element<R_BS> a(elementToNr.find(Application<AutomorphismGroup<CR> >()(*it,Alpha<CR>()()))->second);
	Element<R_BS> t(elementToNr.find(Application<AutomorphismGroup<CR> >()(*it,Theta<CR>()()))->second);
	for (unsigned int rk = 0; rk < Cardinality<CRBS>()(); ++rk) {
	    Element<R_BS> x = UnrankElement<R_BS>()(rk);
	    Element<R_BS> erg = ApplyAlphaThetaImagesToElement<TableChainRing<id>, TableChainRing<id> >()(x,a,t);
	    unsigned int ergrk = RankElement<R_BS>()(erg);
	    Coordinate<R_PG_BS>()(autGenerators[i],rk) = ergrk;
	}
	++i;
    }
    SemigroupByGenerators_BaseSet<R_PG, boost::mpl::pair<id, boost::mpl::int_<0> > >::init(autGenerators);
    */
    init_right_inv_table();
}

#ifdef DB_SUPPORT
#include "Datenbank/db_tableChainRing.hpp"

template<typename id> bool TableChainRing<id>::initFromDB(unsigned int dbid){
    DB_TableChainRing::ChainRingInfo cri;
    std::vector<std::string> dbelemname;
    std::string dbaddtable;
    std::string dbmultable;
    try{
	cri = DB_TableChainRing::getCRInfo(dbid);
	dbelemname=DB_TableChainRing::getElementNames(dbid);
	dbaddtable=DB_TableChainRing::getBlobAddTable(dbid);
	dbmultable=DB_TableChainRing::getBlobMulTable(dbid);
    }
    catch(std::string error){
	std::cout<<"TableChainRing::initFromDB: "<<error;
	return false;
    }

    //set values
    basePrime=cri.p;
    rank=cri.r;
    chainLength=cri.m;
    residueFieldOrder=mathHelper::pow(basePrime,rank);
    name=cri.name;

    elementName.resize(dbelemname.size());
    for (unsigned int i=0; i<dbelemname.size(); i++){
	elementName[i]=dbelemname[i];
    }

    //initialize addtable
    char* buffer=new char[sizeof(unsigned int)*cri.size*cri.size+1];
    unsigned int* p;
    for (unsigned int i=0; i<dbaddtable.length(); i++){
	buffer[i]=dbaddtable[i];
    }
    p=(unsigned int*)(buffer);
    addTable.resize(cri.size * cri.size);
    for(typename ElVec::iterator i = addTable.begin(); i != addTable.end(); ++i){
	(*i).setValue(*p);
	p++;
    }

    //initialize multable
    for (unsigned int i=0; i<dbmultable.length(); i++){
	buffer[i]=dbmultable[i];
    }
    p=(unsigned int*)(buffer);
    mulTable.resize(cri.size * cri.size);
    for(typename ElVec::iterator i = mulTable.begin(); i != mulTable.end(); ++i){
	(*i).setValue(*p);
	p++;
    }

    TableChainRing_BaseSet<id>::initUpperBound(cri.size);

    R_PG_BS::init();
    AutVec autGenerators(cri.nrautgenerators);

    //initialize autGenerators
    std::vector<std::pair<unsigned int, unsigned int> > dbautgens = DB_TableChainRing::getAutGenerators(dbid);
    for (unsigned int i = 0; i < autGenerators.size(); ++i) {
	Element<R_BS> a = UnrankElement<R_BS>()(dbautgens[i].first);
	Element<R_BS> t = UnrankElement<R_BS>()(dbautgens[i].second);
	for (unsigned int rk = 0; rk < cri.size; ++rk) {
	    Element<R_BS> x = UnrankElement<R_BS>()(rk);
	    Element<R_BS> erg = ApplyAlphaThetaImagesToElement<TableChainRing<id>, TableChainRing<id> >()(x,a,t);
	    unsigned int ergrk = RankElement<R_BS>()(erg);
	    Coordinate<R_PG_BS>()(autGenerators[i],rk) = ergrk;
	}
    }
    SemigroupByGenerators_BaseSet<R_PG, boost::mpl::pair<id, boost::mpl::int_<0> > >::init(autGenerators);

    delete[] buffer;
    init_right_inv_table();
    return true;
}
#endif

template<typename id> void TableChainRing<id>::init(std::istream & is) {
    std::string token;
    unsigned int order=0;
    basePrime = 0;
    rank = 0;
    residueFieldOrder = 0;
    chainLength = 0;
    for (is >> token; !is.eof(); is >> token) {
	if (token == std::string("baseprime")) {
	    is >> basePrime;
	    assert (basePrime != 0);
	    if (rank != 0 && chainLength != 0) {
		residueFieldOrder = mathHelper::pow(basePrime,rank);
		order = mathHelper::pow(residueFieldOrder,chainLength);
	    }
	} else if (token == std::string("rank")) {
	    is >> rank;
	    assert (rank != 0);
	    if (basePrime != 0 && chainLength != 0) {
		residueFieldOrder = mathHelper::pow(basePrime,rank);
		order = mathHelper::pow(residueFieldOrder,chainLength);
	    }
	} else if (token == std::string("chainlength")) {
	    is >> chainLength;
	    assert (chainLength != 0);
	    if (basePrime != 0 && rank != 0) {
		residueFieldOrder = mathHelper::pow(basePrime,rank);
		order = mathHelper::pow(residueFieldOrder,chainLength);
	    }
	} else if (token == std::string("name")) {
	    is >> name;
	} else if (token == std::string("elementnames")) {
	    assert(order != 0);
	    elementName.resize(order);
	    for(StrVec::iterator i = elementName.begin(); i != elementName.end(); ++i) {
		is >> (*i);
	    }
	} else if (token == std::string("add")) {
	    assert(order != 0);
	    addTable.resize(order * order);
	    for(typename ElVec::iterator i = addTable.begin(); i != addTable.end(); ++i) {
		unsigned int buf;
		is >> buf;
		(*i).setValue(buf);
	    }
	} else if (token == std::string("mul")) {
	    assert(order != 0);
	    mulTable.resize(order * order);
	    for(typename ElVec::iterator i = mulTable.begin(); i != mulTable.end(); ++i) {
		unsigned int buf;
		is >> buf;
		(*i).setValue(buf);
	    }
	}
    }
    TableChainRing_BaseSet<id>::initUpperBound(order);
    init_right_inv_table();
}

template<typename id> void TableChainRing<id>::init(std::string const filename) {
    std::ifstream is(filename.c_str());
    //std::cout << is.is_open() << std::endl;
    init(is);
}

template<typename id> void TableChainRing<id>::reduceToFather(){
    std::stringstream newName;
    newName<<"("<<name<<")/("<<Theta<TableChainRing<id> >()()<<"^"<<chainLength-1<<")";
    name=newName.str();
    chainLength--;
    unsigned int oldOrder=Cardinality<TableChainRing_BaseSet<id> >()();
    unsigned int order=oldOrder/residueFieldOrder;
    TableChainRing_BaseSet<id>::initUpperBound(order);
    typename TableChainRing<id>::ElVec newAddTable(order*order);
    typename TableChainRing<id>::ElVec newMulTable(order*order);
    elementName.resize(order);
    for (unsigned int i=0; i<order; i++){
	for (unsigned int j=0; j<order; j++){
	    newAddTable[i*order+j].setValue(addTable[i*oldOrder+j].value()%order);
	    newMulTable[i*order+j].setValue(mulTable[i*oldOrder+j].value()%order);
	}
	elementName[i]="res("+elementName[i]+")";
    }
    addTable=newAddTable;
    mulTable=newMulTable;
    init_right_inv_table();
}

template<typename id> void TableChainRing<id>::init_right_inv_table(){
    typedef TableChainRing<id> CR_;
    typedef typename CR_::type::BaseSet CR_BS_;
    rightInvTable.resize(Cardinality<CR_BS_>()());
    for(unsigned int i = 0; i < Cardinality<CR_BS_>()(); ++i) {
	unsigned int thetaPower = (mathHelper::pow(ResidueFieldOrder<CR_>()(),Height<CR_>()(Element<CR_BS_>(i)))) % Cardinality<CR_BS_>()();
	for(unsigned int j = 0; j < Cardinality<CR_BS_>()(); ++j) {
	    if (CR_::getMulTable(i,j).value() == thetaPower) {
		rightInvTable[i] = j;
		break;
	    }
	}
    }
}

template<typename id>
struct AutomorphismGroup<TableChainRing<id> > : public SemigroupByGenerators<PermutationGroup<TableChainRing_BaseSet<id> >, boost::mpl::pair<id, boost::mpl::int_<0> > > {};

template<typename id>
struct Base<AutomorphismGroup<TableChainRing<id> > > {
    typedef SemigroupByGenerators<PermutationGroup<TableChainRing_BaseSet<id> >, boost::mpl::pair<id, boost::mpl::int_<0> > > type;
};

//better: instead of writing the following code, write a general application of a permutation group to elements of the base set
template<typename id, typename Derived>
struct Application<SemigroupByGenerators<PermutationGroup<TableChainRing_BaseSet<id> >, boost::mpl::pair<id, boost::mpl::int_<0> > >, Derived> : public Application<Root, Derived> {
    typedef Derived AG_;
    typedef typename AG_::type::BaseSet AG_BS;
    typedef typename AG_BS::PG_BaseSet R_BS;
    typedef typename R_BS::BaseTableChainRing R_;
    Element<R_BS> operator()(Element<AG_BS> const & aut, Element<R_BS> const & x) const {
	return UnrankElement<R_BS>()(Coordinate<AG_BS>()(aut,RankElement<R_BS>()(x)).value());
    }
};

#endif
