/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef ABSTRACT_CODEWORD_MANAGER_HPP
#define ABSTRACT_CODEWORD_MANAGER_HPP

#include "./Ringe/tableChainRing.hpp"
#include "./Ringe/mathHelper.hpp"
#include "defs.h"
#include "heurico2.h"
#include "log.h"
#include "twoDimArray.hpp"
#include "messageType.hpp"
#include "typeSection.hpp"
#include "abstractWeightTable.hpp"
#include "homogenousWeightTable.hpp"
#include "hammingWeightTable.hpp"
#include "grayLikeAlgorithm.hpp"
#include "neighbourhoodClass.h"
#include "helpFunctions.h"

#include <string>
#include <algorithm>
#include <sstream>
#include <vector>
#include <iomanip>
#include <iostream>

template <class CHAINRING>
class AbstractCodewordManager{
    public:
        AbstractCodewordManager();
        ~AbstractCodewordManager();
        virtual void initializeEval()=0;
	virtual EVAL_TYPE getEval(COLUMN_NO)=0;
	virtual std::pair<EVAL_TYPE,HASH_TYPE> getEvalAndHash(COLUMN_NO)=0;
	void initializeSearch(unsigned int, WEIGHT_SUM);
	void initializeGeometry(unsigned int mLength, std::vector<PERIOD> lambdas, unsigned int whichWeight, PERIOD minPeriodCols, PERIOD minPeriodMess, bool _additiveOnly, std::string iMatFileName);
        void push_column(COLUMN_NO);
        void pop_column();
	bool plotkinCut(unsigned char* colIgnoreList);
	COLUMN_NO convertColumnToNumber(Element<typename CHAINRING::type::BaseSet>*);
	COLUMN_NO getNumberOfColumns();
	unsigned int getColumnCounter(COLUMN_NO c);
	bool fitsSystematicEnoding(COLUMN_NO,unsigned int);
	bool getAdditiveOnly();
	std::string getCurrentGMAsString();
	std::string getNHCDistributionAsString();
	std::string getChosenColumnsAsString();
	std::string getIndividualWeightsAsString();
	std::string getWeightDistributionAsString();
	std::string getSymmetrizedWeightDistributionAsString();
        std::string getColumnsAsString();
        std::string getMessagesAsString();
        std::string getIncidenceMatrixAsString();
	unsigned int getMessageLength();
	void setOutputLevel(unsigned int l);
	unsigned int getOutputLevel();
        std::vector<std::pair<COLUMN_NO, unsigned int> > getNonzeroBounds();
    protected:
	MessageType<CHAINRING> computeMessageType(Element<typename CHAINRING::type::BaseSet>*);
	void generateMessagesInTypeOrder();
	void generateColumnRepresentatives();
	void generateIncidenceMatrix();
	void computeSystematicCodingSteps();
	void freeMemory();
	void freeTypeSectionMemory();
        unsigned int messageLength; //number of lambdas
	std::vector<PERIOD> lambda;
	PERIOD minPeriod_cols;
	PERIOD minPeriod_mess;
	bool additiveOnly;
        COLUMN_NO c_tot; //total number of (projective) columns with the chosen minPeriod
        MESSAGE_NO m_tot; //total number of (projective) messages corresponding to the chosen minPeriod
	TwoDimArray<Element<typename CHAINRING::type::BaseSet> > c_rep; //(projective) representatives of the columns
        TwoDimArray<Element<typename CHAINRING::type::BaseSet> > r_rep; //(projective) representatives of the messages
        TwoDimArray<WEIGHT_TYPE> incidence_matrix;
        std::vector<TypeSection<CHAINRING>* > type_sections;
	std::vector<NumberListEntry*> removedWords[MAX_LENGTH+1];
	std::vector<unsigned int> additiveMultiplicity; //only used if additiveOnly==true
	unsigned int outputLevel;
        unsigned int n;
        WEIGHT_SUM d;
	std::vector<COLUMN_NO> nonzeroColumns;
        std::vector<COLUMN_NO> columnHistory; //numbers of the columns in the GM
	
	//using dynamic memory
	AbstractWeightTable<CHAINRING>* weightTable;
        unsigned int* systematicCodingStep;
	unsigned int* columnCounter;
        WEIGHT_SUM* codewordWeight;
	std::vector<MESSAGE_NO>* wordsOfWeight;
	NeighbourhoodClass* topNHC;
	NeighbourhoodClass** nbcByColumnNumber;
	NeighbourhoodClass** nbcByMessageNumber;
};

//******************** implementation ********************

template <class CHAINRING>
AbstractCodewordManager<CHAINRING>::AbstractCodewordManager() : outputLevel(OUTPUT_NORMAL){
    this->weightTable=NULL;
    this->systematicCodingStep=NULL;
    this->columnCounter=NULL;
    this->codewordWeight=NULL;
    this->wordsOfWeight=NULL;
    this->topNHC=NULL;
    this->nbcByColumnNumber=NULL;
    this->nbcByMessageNumber=NULL;
}

template <class CHAINRING>
AbstractCodewordManager<CHAINRING>::~AbstractCodewordManager(){
    this->freeMemory();
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::freeMemory(){
    if (this->weightTable!=NULL){
	delete this->weightTable;
    }
    if (this->systematicCodingStep!=NULL){
        delete[] this->systematicCodingStep;
    }
    if (this->columnCounter!=NULL){
	delete[] this->columnCounter;
    }	    
    if (this->codewordWeight!=NULL){
        delete[] this->codewordWeight;
    }
    if (this->topNHC!=NULL){
	delete this->topNHC; //this frees the memory for all the subNHCs!
    }
    if (this->nbcByColumnNumber!=NULL){
        delete[] this->nbcByColumnNumber;
    }
    if (this->nbcByMessageNumber!=NULL){
	delete[] this->nbcByMessageNumber;
    }
    this->freeTypeSectionMemory();
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::freeTypeSectionMemory(){
    //clean up type sections
    for (unsigned int i=0; i<this->type_sections.size(); i++){
	delete this->type_sections[i];
    }
    type_sections.clear();
}

template <class CHAINRING>
bool AbstractCodewordManager<CHAINRING>::getAdditiveOnly(){
    return this->additiveOnly;
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::initializeGeometry(unsigned int mLength, std::vector<PERIOD> lambdas, unsigned int whichWeight, PERIOD minPeriodCols, PERIOD minPeriodMess, bool _additiveOnly, std::string iMatFileName){
    this->freeMemory();
    PERIOD m=CHAINRING::getChainLength();

    this->messageLength=mLength;
    this->lambda=lambdas;
    this->additiveOnly=_additiveOnly;
    switch(whichWeight){
	case WEIGHT_HAMWEIGHT:
	    this->weightTable=(AbstractWeightTable<CHAINRING>*)(new HammingWeightTable<CHAINRING>());
	    break;
	case WEIGHT_HOMWEIGHT:
	    this->weightTable=(AbstractWeightTable<CHAINRING>*)(new HomogenousWeightTable<CHAINRING>());
	    break;
	default:
	    writeToConsole("Error: AbstractCodewordManager<CHAINRING>::initializeGeometry(): unkwown weight ID!. Exiting...\n");
	    quitHeurico(-1);
	    break;
    }
    this->weightTable->init();
    this->minPeriod_cols=minPeriodCols;
    this->minPeriod_mess=minPeriodMess;

    //do basic checks
    for (unsigned int i=0; i<this->messageLength; i++){
        if (m<this->lambda[i]){
	    if (this->outputLevel>=OUTPUT_ALWAYS){
                writeToConsole(std::string("Error in AbstractCodewordManager(): lambda > chainLength!\n"));
	    }
            quitHeurico(-1);
        }
        if (i>0 && lambda[i-1]<lambda[i]){
	    if (this->outputLevel>=OUTPUT_ALWAYS){
                writeToConsole(std::string("Error in AbstractCodewordManager(): lambdas are not monoton decreasing!\n"));
	    }
            quitHeurico(-1);
        }    
    }

    //generate representatives for columns and messages
    generateColumnRepresentatives();
    generateMessagesInTypeOrder();

    //generate table containing the systematic coding steps
    computeSystematicCodingSteps();

    //generate (weighted) incidence matrix
    generateIncidenceMatrix();
    if (iMatFileName!=""){
	std::ofstream ofs(iMatFileName.c_str());
	if (ofs.is_open()){
	    ofs<<this->getIncidenceMatrixAsString();
	    ofs.close();
	}
	else{
	    std::stringstream ss;
	    ss<<"Error in AbstractCodewordManager(): could not open "<<iMatFileName<<" to store the incidence matrix."<<std::endl;
	    writeToConsole(ss.str());
	    quitHeurico(-1);
	}
    }

    //allocate memory for codeword weights
    this->codewordWeight=new(std::nothrow) WEIGHT_SUM[this->m_tot];
    if (this->codewordWeight==NULL){
	if (this->outputLevel>=OUTPUT_ALWAYS){
	    writeToConsole(std::string("Error in AbstractCodewordManager(): could not reserve memory for codeword weights!\n"));
	}
        quitHeurico(-1);
    }
    if (this->outputLevel>=OUTPUT_NORMAL){
        writeToConsole(std::string("memory for codeword weights allocated!\n"));
    }
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
        this->codewordWeight[i]=0;
    }
    this->columnCounter=new unsigned int[this->c_tot];
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::initializeSearch(unsigned int _n, WEIGHT_SUM _d){
    this->n=_n;
    this->d=_d;
    this->columnHistory.clear();
    for (unsigned int i=0; i<this->c_tot; i++){
	this->columnCounter[i]=0;
    }
    this->nonzeroColumns.clear();
    if (this->wordsOfWeight!=NULL){
	delete[] this->wordsOfWeight;
    }
    this->wordsOfWeight=new std::vector<MESSAGE_NO>[this->d+1];
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
	this->wordsOfWeight[0].push_back(i);
    }
}

//*** generate (projective) column representatives ***  
template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::generateColumnRepresentatives(){
    PERIOD m=CHAINRING::getChainLength();
    unsigned int q=mathHelper::pow(CHAINRING::getBasePrime(),CHAINRING::getRank());

    //first: compute total number of column representatives
    this->c_tot=0;
    TwoDimArray<COLUMN_NO> n_c;
    if (n_c.init(m+1,this->messageLength)!=0){
	if (this->outputLevel>=OUTPUT_ALWAYS){
	    writeToConsole(std::string("Error in AbstractCodewordManager(): could not reserve memory for n_c!\n"));
	}
	quitHeurico(-1);
    }
    for (PERIOD i=m; i>=minPeriod_cols; i--){
        for (unsigned int j=0; j<this->messageLength && lambda[j]>=i; j++){
            n_c[i][j]=1;
            n_c[i][j]*=mathHelper::pow(q,(i-1)*j);
            for (unsigned int k=j+1; k<this->messageLength; k++){
                n_c[i][j]*=mathHelper::pow(q,(std::min)(lambda[k],i));
            }
            this->c_tot+=n_c[i][j];
        }
    }

    //now: generate them
    if (this->c_rep.init(this->c_tot,this->messageLength)!=0){
	if (this->outputLevel>=OUTPUT_ALWAYS){
	    writeToConsole(std::string("Error in AbstractCodewordManager(): could not reserve memory for c_rep!\n"));
	}
        quitHeurico(-1);
    }
    COLUMN_NO col_counter=0;
    unsigned int* pot=new unsigned int[this->messageLength]; //possibilities for the different positions
    std::vector<unsigned int> tmSet,nullSet;
    for (unsigned int i=0; i<q; i++){
	tmSet.push_back(i);
    }
    nullSet.push_back(0);
    Element<typename CHAINRING::type::BaseSet> theta=Theta<CHAINRING>()();
    Element<typename CHAINRING::type::BaseSet> thetaPow=NeutralElement<Semigroup<typename CHAINRING::type::Mul> >()();
    std::vector<NeighbourhoodClass*>* nhcStack=new std::vector<NeighbourhoodClass*>[m];
    this->nbcByColumnNumber=new NeighbourhoodClass*[this->c_tot];
    NeighbourhoodClass* currentNHC=new NeighbourhoodClass();
    for (PERIOD i=m; i>=this->minPeriod_cols; i--){
        for (unsigned int j=0; j<this->messageLength && this->lambda[j]>=i; j++){
            if (this->messageLength>0){
                for (unsigned int k=this->messageLength-1; k>j; k--){
                    pot[k]=(std::min)(lambda[k],i);
                }
            }        
            pot[j]=0;
            unsigned int k=j;
            while(k!=0){
                k--;
                pot[k]=i-1;
            }
	    std::vector<std::vector<unsigned int>*> container;
	    for (PERIOD l=1; l<=m; l++){
		for (int p=this->messageLength-1; p>=0; p--){
		    if (i>=l && pot[p]>=l){
			container.push_back(&tmSet);
		    }
		    else{
			container.push_back(&nullSet);
		    }
		}
	    }
	    GrayLikeAlgorithm<std::vector<unsigned int> > gla;
	    gla.initialize(container);
	    std::vector<unsigned int> v=gla.resetToBegin();
	    int endIndex=this->messageLength*m;
	    while(true){
		for (unsigned int p=0; p<this->messageLength; p++){
		    std::vector<unsigned int> tmExp;
		    for (int index=endIndex-1-p; index>=0; index-=this->messageLength){
			tmExp.push_back(v[index]);
		    }
		    SetRightThetaAdicExpansion<CHAINRING>()(c_rep[col_counter][p],tmExp);
		    c_rep[col_counter][j]=thetaPow;
		}
		this->nbcByColumnNumber[col_counter]=currentNHC;
		currentNHC->classSize++;
		col_counter++;
		int changedPos=gla.next(v);
		if (!changedPos){
		    break;
		}
		PERIOD pChange=m-(changedPos-1)/this->messageLength;
		for (PERIOD pdist=1; pdist<pChange; pdist++){
    		    currentNHC->distance=pdist;
		    nhcStack[pdist].push_back(currentNHC);
		    if (pdist>1){
			currentNHC->setChildren(nhcStack[pdist-1]);
			nhcStack[pdist-1].clear();
		    }
		    currentNHC=new NeighbourhoodClass();
		}
	    }
    	    for (PERIOD pdist=1; pdist<i; pdist++){
    		 currentNHC->distance=pdist;
		 nhcStack[pdist].push_back(currentNHC);
		 if (pdist>1){
		     currentNHC->setChildren(nhcStack[pdist-1]);
		    nhcStack[pdist-1].clear();
		 }
		 currentNHC=new NeighbourhoodClass();
	    }
        }
	thetaPow*=theta;
    }
    for (PERIOD pdist=minPeriod_cols; pdist<this->lambda[0]; pdist++){
	currentNHC->distance=pdist;
	nhcStack[pdist].push_back(currentNHC);
	if (pdist>1){
	    currentNHC->setChildren(nhcStack[pdist-1]);
	    nhcStack[pdist-1].clear();
	}
	currentNHC=new NeighbourhoodClass();
    }
    this->topNHC=currentNHC;
    this->topNHC->distance=this->lambda[0];
    if (this->lambda[0]>1){
	this->topNHC->setChildren(nhcStack[this->lambda[0]-1]);
    }
    delete[] nhcStack;
    delete[] pot;
    if (this->outputLevel>=OUTPUT_NORMAL){
	std::stringstream ss;
	ss<<col_counter<<" columns were generated!\n";
	writeToConsole(ss.str());
	if (this->outputLevel>=OUTPUT_DEBUG){
	    writeToConsole(this->getColumnsAsString());
	}
    }
}   

//attention: columns have to be generated before this function is called!
template <class CHAINRING>
MessageType<CHAINRING> AbstractCodewordManager<CHAINRING>::computeMessageType(Element<typename CHAINRING::type::BaseSet>* temp_row){
    MessageType<CHAINRING> temp_type;
    /*
    for (COLUMN_NO j=0; j<this->c_tot; j++){
    Element<typename CHAINRING::type::BaseSet> res=zero<CHAINRING>();
    for (unsigned int k=0; k<this->messageLength; k++){
    res=res+temp_row[k]*this->c_rep[j][k];
    }
    temp_type.occurencesOfPeriod[period(res)]++;
    }*/

    //Go trough all column representatives in the following manner:
    //p is the period to which the element on position pos (see below) is normalized; all preceding elements must
    //have period<=p-1, all following ones period<=p
    PERIOD m=CHAINRING::getChainLength();
    unsigned int q=mathHelper::pow(CHAINRING::getBasePrime(),CHAINRING::getRank());
    for (PERIOD p=m; p>=this->minPeriod_cols; p--){ 
	//totalPowCounter is an intermediate value from which the q-log of the number all those
	//representatives (powCounter) can be computed.
	unsigned int totalPowCounter=0;
	for (unsigned int pos=0; pos<this->messageLength; pos++){
	    totalPowCounter+=(std::min)(lambda[pos],p);
	}
	for (unsigned int pos=0; pos<this->messageLength; pos++){
	    if (lambda[pos]<p){ //then its impossible to normalize this position to p
		break;
	    }
	    //to compute powCounter, consider that all preceding elements have period <=p-1 (this explains pos*1);
	    //further, p has to be subtracted as the element at position pos is normalized.
	    unsigned int powCounter=totalPowCounter-pos*1-p;

	    //maxPeriod indicates the max period occuring by multiplying a message entry (with position != pos) with 
	    //the corresponding entry of one of the considered column representatives
	    int maxPeriod=0;
	    for (unsigned int j=0; j<pos; j++){
		maxPeriod=(std::max)(maxPeriod,(int)(p-1-(m-period(temp_row[j]))));
	    }
	    for (unsigned int j=pos+1; j<this->messageLength; j++){
		maxPeriod=(std::max)(maxPeriod,(int)((std::min)(p,lambda[j])-(m-period(temp_row[j]))));
	    }
	    int periodOnPos=p-(m-period(temp_row[pos]));
	    if (maxPeriod>=periodOnPos){
		//the elements with period <=maxPeriod will occur with the same frequency when multiplying the message
		//with all different column representatives; u_pos*v_pos is only a translation.
		unsigned int factor=mathHelper::pow(q,powCounter-maxPeriod);
		unsigned int factor2=0;
		for (int j=0; j<=maxPeriod; j++){
		    temp_type.occurencesOfPeriod[j]+=factor-factor2;
		    factor2=factor;
		    factor*=q;
		}
	    }
	    else{
		//in this case, the multiplication of the message with all column representatives will give an
		//element of period periodOnPos
		temp_type.occurencesOfPeriod[periodOnPos]+=mathHelper::pow(q,powCounter);
	    }
	}
    }
    return temp_type;
}

//attention: columns have to be generated before this function is called!
template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::generateMessagesInTypeOrder(){
    //first: compute message types by which the messages are ordered
    //TO DO: adapt it for general chain rings, not only tableChainRings
    this->freeTypeSectionMemory();
    PERIOD m=CHAINRING::getChainLength();

    //only used if this->additiveOnly==true
    std::vector<std::set<Element<typename CHAINRING::type::BaseSet>,Compare<typename CHAINRING::type::BaseSet> > > additiveSubRingThetaMultiples(m+1);
    std::vector<unsigned int> tempAddMultiplicity(this->c_tot);
    if (this->additiveOnly){
	this->additiveMultiplicity.clear();
	Element<typename CHAINRING::type::BaseSet> one=NeutralElement<Semigroup<typename CHAINRING::type::Mul> >()();
	Element<typename CHAINRING::type::BaseSet> u=one;    
	while(true){
	    for (unsigned int mul=0; mul<=m; mul++){ 
		additiveSubRingThetaMultiples[mul].insert(RightMultiplicationByThetaPower<CHAINRING>()(u,mul));
	    }
	    if (EqualsNeutralElement<Semigroup<typename CHAINRING::type::Add> >()(u)){
		break;
	    }
	    u+=one;
	}
    }
    std::vector<std::pair<MessageType<CHAINRING>,std::vector<int> > > temp_messageTypes;
    Element<typename CHAINRING::type::BaseSet>* temp_row=new Element<typename CHAINRING::type::BaseSet>[this->messageLength];
    this->m_tot=0;
    for (COLUMN_NO i=0; i<this->c_tot; i++){
	if (this->additiveOnly){
	    tempAddMultiplicity[i]=0;
	}
	PERIOD minPeriod=0;
        for (unsigned int j=0; j<this->messageLength; j++){ //temp_row is the message currently investigated
	    PERIOD currPeriod=Period<CHAINRING>()(c_rep[i][j]);
	    minPeriod=(std::max)(minPeriod,currPeriod);
            temp_row[j]=RightDivisionByThetaPower<CHAINRING>()(this->c_rep[i][j],m-this->lambda[j]);
        }
	if (minPeriod<this->minPeriod_mess){
	    continue;
	}
	if (this->additiveOnly){
	    //run through the units modulo theta^minperiod and test whether this gives a vector which
	    //is equivalent to a vector over additiveSubRing (if yes -> add multiplicity by one, independent
	    //how many such equivalent vectors exist, as they all lead to the same codeword)
	    for (Iterator<typename CHAINRING::type::BaseSet> it = Begin<typename CHAINRING::type::BaseSet>()(); !IsAtEnd<typename CHAINRING::type::BaseSet>()(it); it++){
		if (Height<CHAINRING>()(*it)>0){
		    continue;
		}
		if (RepresentativeModThetaPower<CHAINRING>()(*it,minPeriod)!=(*it)){
		    continue;
		}
		bool fits=true;
		for (unsigned int j=0; j<this->messageLength; j++){
		    if (additiveSubRingThetaMultiples[m-this->lambda[j]].find(RightMultiplicationByThetaPower<CHAINRING>()(temp_row[j]*(*it),m-this->lambda[j]))==additiveSubRingThetaMultiples[m-this->lambda[j]].end()){
			fits=false;
			break;
		    }
		}
		if (fits){
		    tempAddMultiplicity[i]++;
		}
	    }
	    if (tempAddMultiplicity[i]==0){
		continue;
	    }
	    /*
	    std::cout<<"multiplicity "<<tempAddMultiplicity[i]<<": ";
	    for (unsigned int j=0; j<this->messageLength; j++){
		std::cout<<temp_row[j]<<" ";
	    }
	    std::cout<<std::endl;*/
	}
	this->m_tot++;

	MessageType<CHAINRING> temp_type=computeMessageType(temp_row);

        bool newType=true;
        for (unsigned int j=0; j<temp_messageTypes.size(); j++){
            if (temp_messageTypes[j].first==temp_type){
                temp_messageTypes[j].second.push_back(i);
                newType=false;
                break;
            }
        }
        if (newType){
            std::vector<int> t;
            t.push_back(i);
            temp_messageTypes.push_back(std::pair<MessageType<CHAINRING>,std::vector<int> >(temp_type,t));
	    if (this->outputLevel>=OUTPUT_NORMAL){
		std::stringstream output;
		output<<"New message type found. Message ";
		for (unsigned int i=0; i<this->messageLength; i++){
		    output<<temp_row[i]<<" ";
		}
		output<<std::endl;
		output<<"has type ";
		for (unsigned int i=0; i<=CHAINRING::getChainLength(); i++){
		    output<<temp_type.occurencesOfPeriod[i]<<"x"<<i<<" ";
		}
		output<<std::endl;
		writeToConsole(output.str());
	    }
        }
    }
    delete[] temp_row;
    if (this->outputLevel>=OUTPUT_NORMAL){
	std::stringstream output;
        output<<temp_messageTypes.size()<<" type orders generated!\n";
	writeToConsole(output.str());
    }	    

    //now generate messages in type order
    if (this->r_rep.init(this->m_tot,this->messageLength)!=0){
	if (this->outputLevel>=OUTPUT_ALWAYS){
	    writeToConsole(std::string("Error in AbstractCodewordManager(): could not reserve memory for r_rep!\n"));
	}
        quitHeurico(-1);
    }
    MESSAGE_NO row_counter=0;
    this->nbcByMessageNumber=new NeighbourhoodClass*[this->m_tot];
    for (unsigned int i=0; i<temp_messageTypes.size();i++){
	std::vector<MESSAGE_NO> messageNumbers;
	for (unsigned int j=0; j<temp_messageTypes[i].second.size(); j++){
            COLUMN_NO no=temp_messageTypes[i].second[j];
            for (unsigned int k=0; k<this->messageLength; k++){
                this->r_rep[row_counter][k]=RightDivisionByThetaPower<CHAINRING>()(this->c_rep[no][k],m-this->lambda[k]);
            }
	    messageNumbers.push_back(row_counter);
	    this->nbcByMessageNumber[row_counter]=this->nbcByColumnNumber[no];
	    if (this->additiveOnly){
		this->additiveMultiplicity.push_back(tempAddMultiplicity[no]);
	    }
            row_counter++;
        }
        this->type_sections.push_back(new TypeSection<CHAINRING>(temp_messageTypes[i].first,messageNumbers));
    }
    if (this->outputLevel>=OUTPUT_NORMAL){
	std::stringstream output;
        output<<row_counter<<" messages generated!\n";
	writeToConsole(output.str());
	if (this->outputLevel>=OUTPUT_DEBUG){
	    writeToConsole(this->getMessagesAsString());
	}
    }
}

//TO DO: adapt this for general chain rings, not only tableChainRings
template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::computeSystematicCodingSteps(){
    this->systematicCodingStep=new unsigned int[this->c_tot];
    unsigned int size=Cardinality<typename CHAINRING::type::BaseSet>()();
    unsigned int q=ResidueFieldOrder<CHAINRING>()();
    for (COLUMN_NO c=0; c<this->c_tot; c++){
	this->systematicCodingStep[c]=this->messageLength;
	unsigned int pos=this->messageLength-1;
	while(true){
	    if (!EqualsNeutralElement<Semigroup<typename CHAINRING::type::Add> >()(this->c_rep[c][pos])){
		break;
	    }
	    pos--;
	}

	unsigned int theta_exp=ChainLength<CHAINRING>()()-lambda[pos];
	if (Height<CHAINRING>()(this->c_rep[c][pos])>theta_exp){
	    continue;
	}   
	Element<typename CHAINRING::type::BaseSet> theta_pot=RightMultiplicationByThetaPower<CHAINRING>()(NeutralElement<Semigroup<typename CHAINRING::type::Mul> >()(),theta_exp);
	Element<typename CHAINRING::type::BaseSet> r;
	for (unsigned int i=0; i<size; i++){
	    if ((i%q)==0){
		continue;
	    }
	    r.setValue(i);
	    if (!(r*c_rep[c][pos]==theta_pot)){
		continue;
	    }
	    bool fits=true;
	    for (unsigned int j=0; j<pos; j++){
		if (!EqualsNeutralElement<Semigroup<typename CHAINRING::type::Add> >()(RightDivisionByThetaPower<CHAINRING>()(r*c_rep[c][j],theta_exp))){
		    fits=false;
		    break;
		}
	    }
	    if (!fits){
		continue;
	    }
	    this->systematicCodingStep[c]=pos;
	    break;
	}
    }
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::generateIncidenceMatrix(){
    // generate (weighted) incidence matrix  
    if (this->incidence_matrix.init(this->c_tot,this->m_tot)!=0){
	if (this->outputLevel>=OUTPUT_ALWAYS){
            writeToConsole(std::string("Error in AbstractCodewordManager(): could not reserve memory for incidence_matrix!\n"));
	}
        quitHeurico(-1);
    }

    this->weightTable->init();
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
        for (COLUMN_NO j=0; j<this->c_tot; j++){
            Element<typename CHAINRING::type::BaseSet> res=NeutralElement<Semigroup<typename CHAINRING::type::Add> >()();
            for (unsigned int k=0; k<this->messageLength; k++){
                res=res+this->r_rep[i][k]*this->c_rep[j][k];
            }
            this->incidence_matrix[j][i]=this->weightTable->get_weight(period(res));
        }
    }
    if (this->outputLevel>=OUTPUT_NORMAL){
	writeToConsole(std::string("incidence matrix generated!\n"));
    }
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::push_column(COLUMN_NO c){
    unsigned int current_n=columnHistory.size();
    this->removedWords[current_n].clear();
    columnHistory.push_back(c);
    if (this->columnCounter[c]==0){
	this->nonzeroColumns.push_back(c);
    }
    this->columnCounter[c]++;
    for (unsigned int i=0; i<=this->d; i++){
	this->wordsOfWeight[i].clear();
    }
    for (unsigned int i=0; i<this->type_sections.size(); i++){
	NumberListEntry* p_end=this->type_sections[i]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
            WEIGHT_SUM w=this->codewordWeight[p->nr]+this->incidence_matrix[c][p->nr];
	    if (w<this->d){
		this->codewordWeight[p->nr]=w;
		this->wordsOfWeight[w].push_back(p->nr);
	    }
	    else{
		this->removedWords[current_n].push_back(p);
		p->removeFromList();
		this->wordsOfWeight[this->d].push_back(p->nr);
	    }
        }
    }
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::pop_column(){
    unsigned int current_n=this->columnHistory.size()-1;
    COLUMN_NO c=this->columnHistory[current_n];
    this->columnHistory.pop_back();
    this->columnCounter[c]--;
    if (this->columnCounter[c]==0){
	this->nonzeroColumns.pop_back();
    }
    for (unsigned int i=0; i<=this->d; i++){
	this->wordsOfWeight[i].clear();
    }
    for (unsigned int i=0; i<this->type_sections.size(); i++){
	NumberListEntry* p_end=this->type_sections[i]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
            this->codewordWeight[p->nr]-=this->incidence_matrix[c][p->nr];
	    this->wordsOfWeight[this->codewordWeight[p->nr]].push_back(p->nr);
        }
    }
    std::vector<NumberListEntry*>::reverse_iterator iter=this->removedWords[current_n].rbegin();
    std::vector<NumberListEntry*>::reverse_iterator endIter=this->removedWords[current_n].rend();
    while(iter!=endIter){
	(*iter)->undoRemove();
	this->wordsOfWeight[this->codewordWeight[(*iter)->nr]].push_back((*iter)->nr);
	iter++;
    }
}

template <class CHAINRING>
bool AbstractCodewordManager<CHAINRING>::plotkinCut(unsigned char* colIgnoreList){
    unsigned int remaining_n=this->n-this->columnHistory.size();
    WEIGHT_SUM* colWeightSum=new WEIGHT_SUM[this->c_tot];
    for (unsigned int j=0; j<this->c_tot; j++){
	colWeightSum[j]=0;
    }
    WEIGHT_SUM missingWeightSum=0;
    for (unsigned int w=0; w<this->d; w++){
  	unsigned int size=this->wordsOfWeight[w].size();
	if (size==0){
	    continue;
	}
	missingWeightSum+=(this->d-w)*size;
	WEIGHT_SUM maxWeightSum=0;
	for (unsigned int j=0; j<this->c_tot; j++){
	    if (colIgnoreList[j]>0){
		continue;
	    }
    	    for (unsigned int i=0; i<size; i++){
		colWeightSum[j]+=this->incidence_matrix[j][this->wordsOfWeight[w][i]];
	    }
	    maxWeightSum=(std::max)(maxWeightSum,colWeightSum[j]);
	}
	if (missingWeightSum>maxWeightSum*remaining_n){
	    delete[] colWeightSum;
	    return true;
	}
    }
    delete[] colWeightSum;
    return false;
}

template <class CHAINRING>
COLUMN_NO AbstractCodewordManager<CHAINRING>::convertColumnToNumber(Element<typename CHAINRING::type::BaseSet>* v){
    PERIOD maxPeriod=0;
    unsigned int pos=0;
    for (unsigned int i=0; i<this->messageLength; i++){
	if (Period<CHAINRING>()(v[i])>maxPeriod){
	    maxPeriod=Period<CHAINRING>()(v[i]);
	    pos=i;
	}
    }
    if (maxPeriod==0){
	if (this->outputLevel>=OUTPUT_ALWAYS){
	    writeToConsole(std::string("Error in AbstractCodewordManager::convertColumnToNumber: zero column not allowed!\n"));
	}
	throw int(ERROR_CONVCOL2NUMBER);
    }
    Element<typename CHAINRING::type::BaseSet> el=RightDivisionByThetaPower<CHAINRING>()(v[pos],ChainLength<CHAINRING>()()-maxPeriod);
    Element<typename CHAINRING::type::BaseSet> inv=InverseElement<Semigroup<typename CHAINRING::type::Mul> >()(el);
    Element<typename CHAINRING::type::BaseSet>* w=new Element<typename CHAINRING::type::BaseSet>[this->messageLength];
    for (unsigned int i=0; i<this->messageLength; i++){
	w[i]=inv*v[i];
    }
    for (COLUMN_NO c=0; c<this->c_tot; c++){
	bool isIdentical=true;
	for (unsigned int i=0; i<this->messageLength; i++){
	    if (this->c_rep[c][i]!=w[i]){
		isIdentical=false;
		break;
	    }
	}
	if (isIdentical){
	    delete[] w;
	    return c;
	}
    }
    delete[] w;
    if (this->outputLevel>=OUTPUT_NORMAL){
        writeToConsole(std::string("Error in AbstractCodewordManager::convertColumnToNumber: could not convert column!\n"));
	throw int(ERROR_CONVCOL2NUMBER);
    }
    return 0xFFFFFFFF;
}

template <class CHAINRING>
COLUMN_NO AbstractCodewordManager<CHAINRING>::getNumberOfColumns(){
    return this->c_tot;
}

template <class CHAINRING>
unsigned int AbstractCodewordManager<CHAINRING>::getColumnCounter(COLUMN_NO c){
    return this->columnCounter[c];
}

template <class CHAINRING>
bool AbstractCodewordManager<CHAINRING>::fitsSystematicEnoding(COLUMN_NO c, unsigned int curr_n){
    if (curr_n>=this->messageLength){
	return true;
    }
    return (this->systematicCodingStep[c]==curr_n);
}

template <class CHAINRING>
std::vector<std::pair<COLUMN_NO, unsigned int> > AbstractCodewordManager<CHAINRING>::getNonzeroBounds(){
    std::vector<std::pair<COLUMN_NO, unsigned int> > res;
    unsigned int size=this->nonzeroColumns.size();
    for (unsigned int i=0; i<size; i++){
	res.push_back(std::pair<COLUMN_NO, unsigned int>(this->nonzeroColumns[i],this->columnCounter[this->nonzeroColumns[i]]));
    }
    return res;
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getCurrentGMAsString(){
    if (this->columnHistory.size()==0){
	return std::string("");
    }
    std::stringstream s;
    for (unsigned int i=0; i<this->messageLength; i++){
        for (unsigned int j=0; j<this->columnHistory.size()-1; j++){
	    s<<std::setw(2)<<this->c_rep[this->columnHistory[j]][i]<<" ";
        }
        s<<std::setw(2)<<this->c_rep[this->columnHistory[this->columnHistory.size()-1]][i]<<std::endl;
    }
    return s.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getNHCDistributionAsString(){
    if (this->columnHistory.size()==0){
	return std::string("");
    }
    std::stringstream s;
    std::map<NeighbourhoodClass*, unsigned int> nhcDist;
    for (unsigned int j=0; j<this->columnHistory.size(); j++){
	NeighbourhoodClass* nhc=this->nbcByColumnNumber[this->columnHistory[j]];
	std::map<NeighbourhoodClass*,unsigned int>::iterator iter=nhcDist.find(nhc);
	if (iter!=nhcDist.end()){
	    iter->second+=1;
	}
	else{
	    nhcDist.insert(std::pair<NeighbourhoodClass*,unsigned int>(nhc,1));
	}
    }
    std::map<NeighbourhoodClass*,unsigned int>::iterator iter=nhcDist.begin();
    std::map<NeighbourhoodClass*,unsigned int>::iterator endIter=nhcDist.end();
    std::stringstream res;
    while(iter!=endIter){
	res<<iter->second<<" ";
	iter++;
    }
    return res.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getChosenColumnsAsString(){
    std::stringstream s;
    for (unsigned int j=0; j<this->columnHistory.size()-1; j++){
	s<<this->columnHistory[j]<<" ";
    }
    s<<this->columnHistory[this->columnHistory.size()-1]<<std::endl;
    return s.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getIndividualWeightsAsString(){
    std::stringstream s;
    for (unsigned int i=0; i<this->type_sections.size(); i++){
	NumberListEntry* p_end=this->type_sections[i]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
            for (unsigned int j=0; j<this->messageLength; j++){
	        s<<this->r_rep[p->nr][j]<<" ";
	    }
	    s<<": w="<<this->codewordWeight[p->nr]<<std::endl;
	}
    }
    s<<std::endl;
    return s.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getWeightDistributionAsString(){
    std::stringstream s;
    unsigned int q=ResidueFieldOrder<CHAINRING>()();
    unsigned int m=ChainLength<CHAINRING>()();
    const unsigned int size=this->columnHistory.size()*this->weightTable->max_weight();
    WEIGHT_SUM* total_w=new WEIGHT_SUM[size+1];
    for (WEIGHT_SUM i=0; i<=size; i++){
        total_w[i]=0;
    }        
    if (this->minPeriod_mess<=1){
        total_w[0]=1;
    }
    WEIGHT_SUM maxWeight=0;
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
	WEIGHT_SUM w=0;
	std::vector<COLUMN_NO>::iterator iter=this->columnHistory.begin();
	std::vector<COLUMN_NO>::iterator endIter=this->columnHistory.end();
	while(iter!=endIter){
	    w+=this->incidence_matrix[*iter][i];
	    iter++;
	}
	PERIOD minPeriod=0;
	for (unsigned int j=0; j<this->messageLength; j++){
	    minPeriod=(std::max)((int)minPeriod,(int)Period<CHAINRING>()(this->r_rep[i][j])-(int)(m-this->lambda[j]));
	}
	if (this->minPeriod_mess<=1){
	    unsigned int multiplicity;
	    if (this->additiveOnly){
		multiplicity=this->additiveMultiplicity[i];
	    }
	    else{
		multiplicity=(q-1)*mathHelper::pow(q,minPeriod-1);
	    }
	    total_w[w]+=multiplicity;
	}
	else{
	    total_w[w]++;
	}
        maxWeight=(std::max)(maxWeight,w);
    }
    for (WEIGHT_SUM i=0; i<maxWeight; i++){
	if (total_w[i]!=0){
            s<<total_w[i]<<"x^"<<i<<"+";
	}
    }
    if (total_w[maxWeight]!=0){
        s<<total_w[maxWeight]<<"x^"<<maxWeight;
    }
    delete[] total_w;
    return s.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getSymmetrizedWeightDistributionAsString(){
    unsigned int m=ChainLength<CHAINRING>()();
    unsigned int q=ResidueFieldOrder<CHAINRING>()();
    std::map<std::string,unsigned int> weightCounter;
    unsigned int* tempWeight=new unsigned int[m+1];
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
	unsigned int multiplicity=1;
	if (this->minPeriod_mess<=1){
	    PERIOD minPeriod=0;
	    for (unsigned int j=0; j<this->messageLength; j++){
		minPeriod=(std::max)((int)minPeriod,(int)Period<CHAINRING>()(this->r_rep[i][j])-(int)(m-this->lambda[j]));
	    }
	    multiplicity=(q-1)*mathHelper::pow(q,minPeriod-1);
	}
	for (unsigned int j=0; j<m+1; j++){
	    tempWeight[j]=0;
	}
	std::vector<COLUMN_NO>::iterator iter=this->columnHistory.begin();
	std::vector<COLUMN_NO>::iterator endIter=this->columnHistory.end();
	while(iter!=endIter){
            Element<typename CHAINRING::type::BaseSet> res=NeutralElement<Semigroup<typename CHAINRING::type::Add> >()();
            for (unsigned int j=0; j<this->messageLength; j++){
                res=res+this->r_rep[i][j]*this->c_rep[*iter][j];
            }
	    tempWeight[Height<CHAINRING>()(res)]++;
	    iter++;
	}
	std::stringstream ss;
	for (unsigned int j=0; j<m+1; j++){
	    char letter='a'+j;
	    ss<<letter<<"^"<<tempWeight[j];
	}
	/*
	if (ss.str()=="a^8b^6c^2"){
	    for (unsigned int j=0; j<this->messageLength; j++){
		std::cout<<this->r_rep[i][j]<<",";
		if (j==this->messageLength-1){
		    std::cout<<std::endl;
		}
	    }
	}*/
	std::map<std::string,unsigned int>::iterator iterWC=weightCounter.find(ss.str());
	if (iterWC==weightCounter.end()){
	    weightCounter.insert(std::pair<std::string,unsigned int>(ss.str(),multiplicity));
	}
	else{
	    iterWC->second+=multiplicity;
	}
    }
    delete[] tempWeight;
    std::map<std::string,unsigned int>::iterator iterWC=weightCounter.begin();
    std::map<std::string,unsigned int>::iterator endIterWC=weightCounter.end();
    std::stringstream ss_res;
    while(true){
	ss_res<<iterWC->second<<iterWC->first;
	iterWC++;
	if (iterWC==endIterWC){
	    break;
	}
	else{
	    ss_res<<"+";
	}
    }
    return ss_res.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getColumnsAsString(){
    std::stringstream s;
    s<<"column representatives:"<<std::endl;
    for (COLUMN_NO i=0; i<this->c_tot; i++){
        s<<i<<": ";
        for (unsigned j=0; j<this->messageLength; j++){
            s<<c_rep[i][j]<<" ";
        }
	s<<std::endl;
    }
    return s.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getMessagesAsString(){
    std::stringstream s;
    s<<"message representatives:"<<std::endl;
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
        s<<i<<": ";
        for (unsigned int j=0; j<this->messageLength; j++){
            s<<this->r_rep[i][j]<<" ";
        }
        s<<std::endl;
    }
    return s.str();
}

template <class CHAINRING>
std::string AbstractCodewordManager<CHAINRING>::getIncidenceMatrixAsString(){
    std::stringstream s;
    s<<"Incidence matrix:"<<std::endl;
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
        int type[]={0,0,0,0};
        for (COLUMN_NO j=0; j<this->c_tot; j++){
            s<<(unsigned int)this->incidence_matrix[j][i]<<" ";
            type[this->incidence_matrix[j][i]]++;
        }/*
	s<<" (";
        for (unsigned int j=0; j<this->messageLength; j++){
            s<<this->r_rep[i][j]<<" ";
        }
	s<<") ;";
        s<<"("<<type[0]<<","<<type[1]<<","<<type[2]<<","<<type[3]<<")"<<std::endl;*/
	s<<std::endl;
    }
    return s.str();
}

template <class CHAINRING>
unsigned int AbstractCodewordManager<CHAINRING>::getMessageLength(){
    return this->messageLength;
}

template <class CHAINRING>
void AbstractCodewordManager<CHAINRING>::setOutputLevel(unsigned int l){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"AbstractCodewordManager<CHAINRING>::setOutputLevel()"<<std::endl;
	ss<<"l = "<<l<<std::endl;
	writeToConsole(ss.str());
    }
    this->outputLevel=l;
}

template <class CHAINRING>
unsigned int AbstractCodewordManager<CHAINRING>::getOutputLevel(){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"AbstractCodewordManager<CHAINRING>::getOutputLevel()"<<std::endl;
	writeToConsole(ss.str());
    }
    return this->outputLevel;
}

    
#endif

