/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef ARCSEARCH_MANAGER
#define ARCSEARCH_MANAGER

#include "codewordManager.hpp"
#include "arcEvaluationTable.h"

template <class CHAINRING>
class ArcSearchManager: public AbstractCodewordManager<CHAINRING>{
public:
    ArcSearchManager();
    ~ArcSearchManager();
    void initializeEval();
    EVAL_TYPE getEval(COLUMN_NO);
    std::pair<EVAL_TYPE,HASH_TYPE> getEvalAndHash(COLUMN_NO);
    void initializeGeometry(unsigned int mLength, unsigned int whichWeight,std::string iMatFileName);
    private:
	TwoDimArray<ArcEvaluationTable> arcEvaluationTables; 	
};

//******************** implementation ********************
template <class CHAINRING>
ArcSearchManager<CHAINRING>::ArcSearchManager(){
}

template <class CHAINRING>
ArcSearchManager<CHAINRING>::~ArcSearchManager(){
}

template <class CHAINRING>
void ArcSearchManager<CHAINRING>::initializeGeometry(unsigned int mLength, unsigned int whichWeight, std::string iMatFileName){
    unsigned int m=ChainLength<CHAINRING>()();
    std::vector<PERIOD> lambdas;
    for (unsigned int i=0; i<mLength; i++){
	lambdas.push_back(m);
    }
    AbstractCodewordManager<CHAINRING>::initializeGeometry(mLength, lambdas, whichWeight, m, m, false, iMatFileName);
    
    //separate hyperplanes due to their distance to (1 0 0 ... 0)^orth
    MessageType<CHAINRING> mt=this->type_sections[0]->messageType;
    delete this->type_sections[0]; //not needed
    this->type_sections.pop_back();
    std::vector<MESSAGE_NO>* hyperPlanesOfDistanceToE1=new std::vector<MESSAGE_NO>[m+1];
    Element<typename CHAINRING::type::BaseSet> unit=NeutralElement<Semigroup<typename CHAINRING::type::Mul> >()();
    for (MESSAGE_NO i=0; i<this->m_tot; i++){
	if (this->r_rep[i][0]!=unit){
	    hyperPlanesOfDistanceToE1[m].push_back(i);
	}
	else{
  	    unsigned int t=0;
	    for (unsigned int j=1; j<this->messageLength; j++){
		t=(std::max)(Period<CHAINRING>()(this->r_rep[i][j]),t);
	    }
	    hyperPlanesOfDistanceToE1[t].push_back(i);
	}
    }
    for (unsigned int t=0; t<=m; t++){
	this->type_sections.push_back(new TypeSection<CHAINRING>(mt,hyperPlanesOfDistanceToE1[t]));
    }
    delete[] hyperPlanesOfDistanceToE1;
}

template <class CHAINRING>
EVAL_TYPE ArcSearchManager<CHAINRING>::getEval(COLUMN_NO c){
    MESSAGE_NO EE=(this->type_sections[0]->messages[0].nextEntry->nr);
    int d1=((int)this->d)-(this->codewordWeight[EE]+this->incidence_matrix[c][EE]);
    if (d1<0){
	return MINUS_INFTY_LL;
    }
    unsigned int m=ChainLength<CHAINRING>()();
    unsigned int remaining_n=this->n-(this->columnHistory.size()+1);
    EVAL_TYPE eval=0;    
    for (unsigned int t=1; t<=m; t++){
	ArcEvaluationTable* evalTable=&(this->arcEvaluationTables[t][d1]);
	NumberListEntry* p_end=this->type_sections[t]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
	    int temp=this->codewordWeight[p->nr]+this->incidence_matrix[c][p->nr];
	    int w_diff=(std::max)(0,(int)this->d-temp);
	    EVAL_TYPE ev=evalTable->get_eval(remaining_n,w_diff);
	    if (ev==MINUS_INFTY_LL){
		//p->sortToFront(); //->AbstractCWM::pop_column must call undoRemoveToEnd();
		return MINUS_INFTY_LL;
	    } 
	    eval+=ev;
	}
    }
    return eval;
}

template <class CHAINRING>
std::pair<EVAL_TYPE,HASH_TYPE> ArcSearchManager<CHAINRING>::getEvalAndHash(COLUMN_NO c){
    MESSAGE_NO EE=(this->type_sections[0]->messages[0].nextEntry->nr);
    int d1=((int)this->d)-(this->codewordWeight[EE]+this->incidence_matrix[c][EE]);
    if (d1<0){
	return std::pair<EVAL_TYPE,HASH_TYPE>(MINUS_INFTY_LL,0);
    }
    unsigned int m=ChainLength<CHAINRING>()();
    unsigned int remaining_n=this->n-(this->columnHistory.size()+1);
    EVAL_TYPE eval=0;    
    this->topNHC->resetHash();
    for (unsigned int t=1; t<=m; t++){
	ArcEvaluationTable* evalTable=&(this->arcEvaluationTables[t][d1]);
	NumberListEntry* p_end=this->type_sections[t]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
	    int temp=this->codewordWeight[p->nr]+this->incidence_matrix[c][p->nr];
	    int w_diff=(std::max)(0,(int)this->d-temp);
	    EVAL_TYPE ev=evalTable->get_eval(remaining_n,w_diff);
	    this->nbcByMessageNumber[p->nr]->hash+=(HASH_TYPE)(ev);
	    if (ev==MINUS_INFTY_LL){
		//p->sortToFront(); //->AbstractCWM::pop_column must call undoRemoveToEnd();
		return std::pair<EVAL_TYPE,HASH_TYPE>(MINUS_INFTY_LL,0);
	    } 
	    eval+=ev;
	}
    }
    return std::pair<EVAL_TYPE,HASH_TYPE>(eval,this->topNHC->getHash());
}

template <class CHAINRING>
void ArcSearchManager<CHAINRING>::initializeEval(){
    unsigned int m=ChainLength<CHAINRING>()();
    unsigned int q=ResidueFieldOrder<CHAINRING>()();
    unsigned int k=this->messageLength;

    MESSAGE_NO EE=(this->type_sections[0]->messages[0].nextEntry->nr);
    int d_max=((int)this->d)-this->codewordWeight[EE];
    d_max=(std::max)(0,d_max);
    int n_max=this->n-this->columnHistory.size();

    this->arcEvaluationTables.init(m+1,d_max+1);
    for (unsigned int t=0; t<=m; t++){
	//consider the weight distribution of the already chosen columns
	WEIGHT_SUM minWeight=this->d;
	NumberListEntry* p_end=this->type_sections[t]->messages;
	NumberListEntry* p=p_end->nextEntry;
	if (p!=p_end){
	    minWeight=this->codewordWeight[p->nr];
	    for (p=p->nextEntry; p!=p_end; p=p->nextEntry){
    		minWeight=(std::min)(minWeight,this->codewordWeight[p->nr]);
	    }
	}

	unsigned int V=mathHelper::pow(q,(m-1)*(k-1))*(mathHelper::pow(q,k)-1)/(q-1);
	unsigned int N=mathHelper::pow(q,(m-1)*(k-2))*(mathHelper::pow(q,k-1)-1)/(q-1);
	unsigned int Dt;
	if (t>=1){
	    Dt=mathHelper::pow(q,(m-1)*(k-2)-t+1)*(mathHelper::pow(q,k-2)-1)/(q-1);
	}
	else{
	    Dt=mathHelper::pow(q,(m-1)*(k-2))*(mathHelper::pow(q,k-1)-1)/(q-1);
	}
	unsigned int EE=V-2*N+Dt;
	for (int d1=0; d1<=d_max; d1++){
	    if (this->d>minWeight){
    		this->arcEvaluationTables[t][d1].initialize(n_max, V, V-N, V-N, EE, d1, this->d-minWeight);
	    }
	    else{
    		this->arcEvaluationTables[t][d1].initialize(n_max, V, V-N, V-N, EE, d1, 0);
	    }
	}
    }
}

#endif

