/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef SEARCH_ALGORITHM_HPP
#define SEARCH_ALGORITHM_HPP

#include <string>
#include <vector>
#include <algorithm>
#include <set>
#include <fstream>
#include <iomanip>
#include <map>
#include "codewordManager.hpp"
#include "numberListEntry.h"
#include "defs.h"
#include "heurico2.h"
#include "time.h"
#include "random.h"
#include "htmlTreeCreator.h"
#include "helpFunctions.h"
#include "evalHashTable.h"
#include "columnHashTable.h"
#include "abstractPreselectionModule.h"

#ifdef STORE_ISOMORPHIC_GMS
    typedef std::map<HASH_TYPE, std::vector<std::string> > GM_STORAGE;
#endif

template<class CHAINRING>
class SearchAlgorithm{
    public:
	SearchAlgorithm();
        ~SearchAlgorithm();
	void setOutputLevel(unsigned int l);
	unsigned int getOutputLevel();
	void initializeGeometry(AbstractCodewordManager<CHAINRING>*);
        void initializeSearch(unsigned int _n, WEIGHT_SUM _d, unsigned int maxHD, unsigned int max_branch, double _a, unsigned int _evalHashSize, unsigned int _columnHashSize, std::string _solutionFile, double _timeout, bool _htmlTree);
	std::pair<int,bool> startSearch(const std::vector<COLUMN_NO>& prescribed_columns, const std::vector<COLUMN_NO>& learnVector, AbstractPreselectionModule* psm);
	void setPath(unsigned int current_n, unsigned int i);
	std::vector<int> getPath();
	void resetPath();
    private:
	bool recursiveSearch(unsigned int current_n, std::set<COLUMN_NO>& learnSet, AbstractPreselectionModule* psm, unsigned int maxBranchNumber, EVAL_TYPE maxBranchDiff, double branch, bool PV);
	unsigned int outputLevel;
	EVAL_TYPE alterBound;
        unsigned int n;
	WEIGHT_SUM d;
	unsigned int maxHashDepth;
	unsigned int evalHashSize,columnHashSize;
        long long nodeCounter;
        long long evaluationCalls;
	long long hashCuts;
	NumberListEntry* colChoice;
	std::vector<COLUMN_NO> nonChoiceableColumns;
	unsigned char* colIgnoreList;
	bool searchSystematicGM;
	bool htmlTree;
        bool foundSolution;
        bool stopSearch;
        std::string solution;
        std::string solutionFile;
        double timeout;
	unsigned int maxBranch;
        timeval startTime;
        timeval lastUpdate;
        AbstractCodewordManager<CHAINRING>* codewordManager;
	unsigned int noOfColumns;
	EvalHashTable* evalHashTable;
	ColumnHashTable* columnHashTable;
	HASH_TYPE* columnHashRandoms;
	HASH_TYPE columnHashValue;
	HTMLTreeCreator htmlTreeCreator;
	std::vector<int> path;
	#ifdef STORE_ISOMORPHIC_GMS
	    GM_STORAGE isomorphicGMs;	
	    std::string generateMagmaScript(GM_STORAGE::iterator& iter, unsigned int _n);
	#endif
};    

template<class CHAINRING>
SearchAlgorithm<CHAINRING>::SearchAlgorithm() : outputLevel(OUTPUT_NORMAL){
    if (this->outputLevel>=OUTPUT_DEBUG){
        std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::SearchAlgorithm()"<<std::endl;
	writeToConsole(ss.str());
    }
    this->colChoice=NULL;
    this->colIgnoreList=NULL;
    this->evalHashTable=NULL;
    this->columnHashTable=NULL;
    this->columnHashRandoms=NULL;
}

template<class CHAINRING>
SearchAlgorithm<CHAINRING>::~SearchAlgorithm(){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::~SearchAlgorithm()"<<std::endl;
	writeToConsole(ss.str());
    }
    if (this->colChoice!=NULL){
	delete[] this->colChoice;
    }
    if (this->colIgnoreList!=NULL){
	delete[] this->colIgnoreList;
    }
    if (this->evalHashTable!=NULL){
        delete this->evalHashTable;
    }
    if (this->columnHashTable!=NULL){
	delete this->columnHashTable;
    }
    if (this->columnHashRandoms!=NULL){
	delete[] this->columnHashRandoms;
    }
}

template<class CHAINRING>
void SearchAlgorithm<CHAINRING>::setOutputLevel(unsigned int l){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::setOutputLevel()"<<std::endl;
	ss<<"l = "<<l<<std::endl;
	writeToConsole(ss.str());
    }
    this->outputLevel=l;
}

template<class CHAINRING>
unsigned int SearchAlgorithm<CHAINRING>::getOutputLevel(){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::getOutputLevel()"<<std::endl;
	writeToConsole(ss.str());
    }
    return this->outputLevel;
}

template<class CHAINRING>
void SearchAlgorithm<CHAINRING>::initializeGeometry(AbstractCodewordManager<CHAINRING>* cwm){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::initializeGeometry()"<<std::endl;
	ss<<"cwm = "<<cwm<<std::endl;
	writeToConsole(ss.str());
    }
    this->codewordManager=cwm;
    this->noOfColumns=this->codewordManager->getNumberOfColumns();
    if (this->colIgnoreList!=NULL){
	delete[] this->colIgnoreList;
    }
    this->colIgnoreList=new unsigned char[this->noOfColumns];
    if (this->colChoice!=NULL){
	delete[] this->colChoice;
    }
    this->colChoice=new NumberListEntry[this->noOfColumns+1];
    this->colChoice[0].nr=0xFFFFFFFF;
    this->colChoice[0].nextEntry=&this->colChoice[1];
    this->colChoice[0].previousEntry=&this->colChoice[this->noOfColumns];
    this->colChoice[0].header=this->colChoice;
    for (unsigned int i=1; i<this->noOfColumns; i++){
	this->colChoice[i].nr=i-1;
	this->colChoice[i].nextEntry=&this->colChoice[i+1];
	this->colChoice[i].previousEntry=&this->colChoice[i-1];
	this->colChoice[i].header=this->colChoice;
    }	
    this->colChoice[this->noOfColumns].nr=this->noOfColumns-1;
    this->colChoice[this->noOfColumns].previousEntry=&this->colChoice[this->noOfColumns-1];
    this->colChoice[this->noOfColumns].nextEntry=&this->colChoice[0];
    this->colChoice[this->noOfColumns].header=this->colChoice;
    if (this->columnHashRandoms!=NULL){
	delete[] this->columnHashRandoms;
    }
    this->columnHashRandoms=new HASH_TYPE[this->noOfColumns];
    for (COLUMN_NO i=0; i<this->noOfColumns; i++){
	this->columnHashRandoms[i]=get64BitRandom();
    }
}

template<class CHAINRING>
void SearchAlgorithm<CHAINRING>::initializeSearch(unsigned int _n, WEIGHT_SUM _d, unsigned int maxHD, unsigned int max_branch, double _a, unsigned int _evalHashSize, unsigned int _columnHashSize, std::string _solutionFile, double _timeout, bool _htmlTree){
    this->evalHashSize=_evalHashSize;
    this->columnHashSize=_columnHashSize;

    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::initializeSearch()"<<std::endl;
	ss<<"_n = "<<_n<<std::endl;
	ss<<"_d = "<<_d<<std::endl;
	ss<<"maxHD = "<<maxHD<<std::endl;
	ss<<"max_branch = "<<max_branch<<std::endl;
	ss<<"a = "<<_a<<std::endl;
	ss<<"_evalHashSize = "<<_evalHashSize<<std::endl;
	ss<<"_columnHashSize = "<<_columnHashSize<<std::endl;
	ss<<"solutionFile = "<<_solutionFile<<std::endl;
	ss<<"timeout = "<<_timeout<<std::endl;
	ss<<"htmlTree = "<<_htmlTree<<std::endl;
	writeToConsole(ss.str());
    }
    if (this->evalHashTable!=NULL){
        delete this->evalHashTable;
    }
    this->evalHashTable=new EvalHashTable(_n,_evalHashSize);
    if (this->columnHashTable!=NULL){
	delete this->columnHashTable;
    }
    this->columnHashTable=new ColumnHashTable(_n,_columnHashSize);
    this->n=_n;
    this->d=_d;
    this->maxHashDepth=maxHD;
    this->maxBranch=max_branch;
    this->alterBound=(EVAL_TYPE)(log(_a)*EVAL_SHIFT);
    this->solutionFile=_solutionFile;
    this->timeout=_timeout;
    this->htmlTree=_htmlTree;
    if (this->htmlTree){
        this->htmlTreeCreator.newRootNode(std::string("htmltree/"),10000000,0);
    }
    for (unsigned int i=0; i<this->noOfColumns; i++){
	this->colIgnoreList[i]=0;
    }
    this->nonChoiceableColumns.clear();
    this->codewordManager->initializeSearch(this->n,this->d);
    this->nodeCounter=0;
    this->evaluationCalls=0;
    this->hashCuts=0;
    this->resetPath();
}

//things done here should depend on prescribed_columns, otherwise put it in initializeSearch!
template<class CHAINRING>
std::pair<int,bool> SearchAlgorithm<CHAINRING>::startSearch(const std::vector<COLUMN_NO>& prescribed_columns, const std::vector<COLUMN_NO>& learnVector, AbstractPreselectionModule* psm){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::startSearch()"<<std::endl;
	ss<<"pc: "<<std::endl;
	for (unsigned int i=0; i<prescribed_columns.size(); i++){
	    ss<<prescribed_columns[i]<<std::endl;
	}
	writeToConsole(ss.str());
    }
    std::set<COLUMN_NO> learnSet;
    for (unsigned int i=0; i<learnVector.size(); i++){
	learnSet.insert(learnVector[i]);
    }
    this->startTime=getCurrentTime();
    this->lastUpdate=this->startTime;
    this->stopSearch=false;
    this->foundSolution=false;
    this->searchSystematicGM=true;
    if (prescribed_columns.size()>0){
	this->searchSystematicGM=false;
        if (prescribed_columns.size()>=this->n){
	    writeToConsole(std::string("Error in SearchAlgorithm::startSearch(): number of prescribed columns >= n!"));
	    quitHeurico(-1);
	}
    }
    if (!learnVector.empty()){
	this->searchSystematicGM=false;
    }
    if (!psm->supportsSystematicGM()){
	this->searchSystematicGM=false;
    }
    if (this->codewordManager->getAdditiveOnly()){
	this->searchSystematicGM=false;
    }
    this->columnHashValue=0;
    for (unsigned int i=0; i<prescribed_columns.size(); i++){
        this->codewordManager->push_column(prescribed_columns[i]);
	this->columnHashValue+=this->columnHashRandoms[prescribed_columns[i]];
    }
    this->codewordManager->initializeEval();
    bool pruned=this->recursiveSearch(prescribed_columns.size(),learnSet,psm,0,0,1,true);
    for (unsigned int i=0; i<prescribed_columns.size(); i++){
        this->codewordManager->pop_column();
    }
    #ifdef STORE_ISOMORPHIC_GMS
    {
	std::ofstream magmaFile("./magma.txt");
	if (magmaFile.is_open()){
	    unsigned int counter=0;
	    for (unsigned int _n=1; _n<=this->n; _n++){
		GM_STORAGE::iterator iterGMs=this->isomorphicGMs.begin();
		GM_STORAGE::iterator endIterGMs=this->isomorphicGMs.end();
		while(iterGMs!=endIterGMs){
		    if (iterGMs->second.size()>1){
			std::string s=this->generateMagmaScript(iterGMs,_n);
			if (s!=""){
			    magmaFile<<"print \"isomorphism class "<<counter<<" (n = "<<_n<<", hash = "<<iterGMs->first<<", size = "<<iterGMs->second.size()<<"):\";"<<std::endl;
			    magmaFile<<s;
			    counter++;
			}
		    }
		    iterGMs++;
		}
	    }
	    magmaFile.close();
	}
    }
    #endif
    timeval currTime=getCurrentTime();
    double time=getTimeDiff(currTime,this->startTime);
    if (this->foundSolution){
	writeToConsole("generator matrix:\n"+this->solution);
        std::stringstream s;
	s<<"nodes = "<<this->nodeCounter<<std::endl;
        s<<"evaluations = "<<this->evaluationCalls<<std::endl;
	s<<"hash cuts = "<<this->hashCuts<<std::endl;
        s<<"time = "<<time<<"s"<<std::endl;
        writeToConsole(s.str());
	return std::pair<int,bool>(FOUNDSOLUTION,pruned);
    }
    else{
        std::stringstream s;
        s<<"Search failed!";
        if (this->stopSearch){
            s<<" (timeout)";
        }
        s<<std::endl;
	s<<"nodes = "<<this->nodeCounter<<std::endl;
        s<<"evaluations = "<<this->evaluationCalls<<std::endl;
	s<<"hash cuts = "<<this->hashCuts<<std::endl;
        s<<"time = "<<time<<"s"<<std::endl;
	if (pruned){
	    s<<"There was some pruning."<<std::endl;
	}
	else{
	    s<<"No pruning was done."<<std::endl;
	}
        writeToConsole(s.str());
        if (this->stopSearch){
	    return std::pair<int,bool>(NOSOLUTION_TIMEOUT,pruned);
        }
        else{
	    return std::pair<int,bool>(NOSOLUTION_COMPLETE,pruned);
        }
    }
}

template<class CHAINRING>
bool SearchAlgorithm<CHAINRING>::recursiveSearch(unsigned int current_n, std::set<COLUMN_NO>& learnSet, AbstractPreselectionModule* psm, unsigned int maxBranchNumber, EVAL_TYPE maxBranchDiff, double branch, bool PV){
    if (this->outputLevel>=OUTPUT_ALL){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::recursiveSearch()"<<std::endl;
	ss<<"current_n = "<<current_n<<std::endl;
	ss<<"maxBranchNumber = "<<maxBranchNumber<<std::endl;
	ss<<"maxBranchDiff = "<<maxBranchDiff<<std::endl;
	ss<<"branch = "<<branch<<std::endl;
	ss<<"PV = "<<PV<<std::endl;
	writeToConsole(ss.str());
    }
    this->nodeCounter++;
    NODE_NUMBER nodeNumber=this->nodeCounter;
    timeval currTime=getCurrentTime();
    if (getTimeDiff(currTime,this->startTime)>=this->timeout){
        this->stopSearch=true;
    }
    if (this->foundSolution || this->stopSearch){
	if (this->outputLevel>=OUTPUT_ALL){
	    writeToConsole("Returning because search is stopping!\n");
	}
	if (this->htmlTree){
	    this->htmlTreeCreator.addComment(nodeNumber,"Stopping!\n",this->evaluationCalls);
	}
        return true;
    }
    this->columnHashTable->insert(this->columnHashValue,current_n);
    if (this->codewordManager->plotkinCut(this->colIgnoreList)){
	if (this->outputLevel>=OUTPUT_ALL){
	    writeToConsole("Plotkin-Cut\n");
	}
	if (this->htmlTree){
	    this->htmlTreeCreator.addComment(nodeNumber,"Plotkin-Cut!\n",this->evaluationCalls);
	}
	return false;
    }
    if (this->htmlTree){
	std::stringstream commentGM;
	commentGM<<"GM:<br> "<<this->codewordManager->getCurrentGMAsString();
	std::string asString=commentGM.str();
	asString=ReplaceString(std::string("\n"),std::string("<br>"),asString);
	this->htmlTreeCreator.addComment(nodeNumber,asString,this->evaluationCalls);
    }
    if (current_n<this->n-1){
	bool pruned=false;
        std::vector<NUMBER_EVAL_PAIR> alternatives;
        NUMBER_EVAL_PAIR bestPair;
        bestPair.first=0;
        bestPair.second.first=MINUS_INFTY_LL;
        EVAL_TYPE evalHashValue1=0,evalHashValue2=0;
	std::vector<NumberListEntry*> removedChoices;
	bool tryIsomorphyCut=(current_n>this->codewordManager->getMessageLength() && current_n<=this->maxHashDepth);
	for (NumberListEntry* p=colChoice[0].nextEntry; p!=colChoice; p=p->nextEntry){
	    if (this->outputLevel>=OUTPUT_ALL){
		std::stringstream ss;
		ss<<"Evaluating column "<<p->nr<<"..."<<std::endl;
		writeToConsole(ss.str());
	    }
	    if (!tryIsomorphyCut && this->colIgnoreList[p->nr]>0){
		if (this->outputLevel>=OUTPUT_ALL){
		    std::stringstream ss;
		    ss<<"Removing column as it is ignored and tryIsomorphyCut=false."<<std::endl;
		    writeToConsole(ss.str());
		}
		p->removeFromList();
    		removedChoices.push_back(p);
		this->nonChoiceableColumns.push_back(p->nr);
		this->colIgnoreList[p->nr]+=2;
		continue;
	    }
	    if (this->searchSystematicGM){
    		//check whether this column fits for construction of a systematic GM
		if (!this->codewordManager->fitsSystematicEnoding(p->nr,current_n)){
		    if (this->outputLevel>=OUTPUT_ALL){
			writeToConsole("Skipped as column is not compatible with systematic encoding.\n");
		    }
		    continue;
		}
	    }
	    if (!psm->isFeasible(p->nr,current_n)){
		continue;
	    }
            NUMBER_EVAL_PAIR tempPair;
            tempPair.first=p->nr;
            tempPair.second.first=this->codewordManager->getEval(p->nr);
            evaluationCalls++;
	    if (this->outputLevel>=OUTPUT_ALL){
		std::stringstream ss;
		ss<<"evaluation = "<<tempPair.second.first/EVAL_SHIFT<<std::endl;
		writeToConsole(ss.str());
	    }
    	    if (tempPair.second.first==MINUS_INFTY_LL){
		p->removeFromList();
    		removedChoices.push_back(p);
		this->nonChoiceableColumns.push_back(p->nr);
		this->colIgnoreList[p->nr]+=2;
		if (this->outputLevel>=OUTPUT_ALL){
		    writeToConsole("Column removed from list because eval=0.\n");
		}
		continue;
	    }
	    else{
		evalHashValue1+=hashTwist(tempPair.second.first);
		evalHashValue2^=hashTwist(tempPair.second.first);
		tempPair.second.first-=alternatives.size(); //to make ordering of columns with same eval unique
		if (!this->columnHashTable->find(this->columnHashValue+this->columnHashRandoms[p->nr],current_n+1)){
		    alternatives.push_back(tempPair);
		}
		else{
    		    if (this->outputLevel>=OUTPUT_ALL){
			writeToConsole("Column not added to alternatives because it would lead to a transposition.\n");
		    }
		}
		if (tempPair.second>bestPair.second){
		    bestPair=tempPair;
		}
	    }
        }
	EVAL_TYPE evalHashValue=((evalHashValue1&0xAAAAAAAAAAAAAAAA)^(evalHashValue2&0x5555555555555555));
        if (this->htmlTree){
    	    std::stringstream hashComment;
	    hashComment<<"hash = "<<evalHashValue<<"."<<std::endl;
	    this->htmlTreeCreator.addComment(nodeNumber,hashComment.str(),this->nodeCounter);
	}
        if (bestPair.second.first!=MINUS_INFTY_LL){
	    EVALHASHTYPE::iterator iter;
	    bool hashCut=false;
	    if (tryIsomorphyCut && bestPair.second.first<0.0){ //otherwise solutions could be ignored, as they dont change the hashkey
		bool magmaVerified=true;
    		#ifdef STORE_ISOMORPHIC_GMS
		{
		    GM_STORAGE::iterator iterGMs=this->isomorphicGMs.find((HASH_TYPE)evalHashValue);
		    if (iterGMs==this->isomorphicGMs.end()){
			std::vector<std::string> v;
			v.push_back(this->codewordManager->getCurrentGMAsString());
			this->isomorphicGMs.insert(std::pair<HASH_TYPE,std::vector<std::string> >((HASH_TYPE)evalHashValue,v));
		    }
		    else{
			iterGMs->second.push_back(this->codewordManager->getCurrentGMAsString());
		    }
		}
    		#endif
		if (this->evalHashTable->find((HASH_TYPE)evalHashValue,current_n,iter) && magmaVerified){
		    hashCut=true;
    		    if (this->htmlTree){
			this->htmlTreeCreator.addComment(nodeNumber,std::string("Hashcut!<br>"),this->evaluationCalls);
			#ifdef HTMLTREE_HASHLINKS
    			    this->htmlTreeCreator.addEdge(nodeNumber,iter->second.second,std::string("first occurrence of this hashvalue"),this->evaluationCalls);			
			#endif
		    }
		    this->hashCuts++;
    		    #ifdef HTMLTREE_HASHLINKS
			pruned|=iter->second.first;
		    #else
			pruned|=iter->second;
		    #endif
		}
	    }
	    if (hashCut){
		if (this->outputLevel>=OUTPUT_DEBUG){
		    writeToConsole("Hash cut!\n");
		}
	    }
	    else{
		if (this->outputLevel>=OUTPUT_DEBUG){
		    std::stringstream ss;
		    ss<<"Start branching for n = "<<current_n<<std::endl;
		    writeToConsole(ss.str());
		}
		std::sort(alternatives.begin(),alternatives.end(),comp_NUMBER_EVAL_PAIR());
		unsigned int lastAlternative;
		unsigned int stop=(std::min)((unsigned int)alternatives.size(),this->maxBranch);
		pruned|=(stop<alternatives.size());
		if (this->htmlTree){
		    std::stringstream ss;
		    ss<<"best pair: "<<bestPair.first<<"("<<exp(bestPair.second.first/EVAL_SHIFT)<<")";
    		    this->htmlTreeCreator.addComment(nodeNumber,ss.str(),this->evaluationCalls);
    		    this->htmlTreeCreator.addComment(nodeNumber,std::string("alternatives:"),this->evaluationCalls);
		}
		for (lastAlternative=0; lastAlternative<stop; lastAlternative++){
		    EVAL_TYPE branchDiff=bestPair.second.first-alternatives[lastAlternative].second.first;
		    if (branchDiff>this->alterBound){
			pruned=true;
			break;
		    }
		    if (this->htmlTree){
			std::stringstream ss;
			ss<<alternatives[lastAlternative].first<<"("<<exp(alternatives[lastAlternative].second.first/EVAL_SHIFT)<<")"<<std::endl;
    			this->htmlTreeCreator.addComment(nodeNumber,ss.str(),this->evaluationCalls);
		    }
		}
		if (this->outputLevel>=OUTPUT_DEBUG){
		    std::stringstream ss;
		    ss<<"best pair: "<<bestPair.first<<" ("<<bestPair.second.first/EVAL_SHIFT<<")"<<std::endl;
		    ss<<"lastAlternative = "<<lastAlternative<<std::endl;
		    writeToConsole(ss.str());
		}
		int ignoredColumns=0;
    		for (unsigned int i=0; i<lastAlternative; i++){
		    if (this->outputLevel>=OUTPUT_DEBUG){
			std::stringstream ss;
			ss<<"testing column "<<alternatives[i].first<<" (eval = "<<alternatives[i].second.first/EVAL_SHIFT<<")"<<std::endl;
			writeToConsole(ss.str());
		    }
		    if (!this->stopSearch){
			setPath(current_n,i);
		    }
		    if (this->foundSolution){
			if (this->outputLevel>=OUTPUT_DEBUG){
			    writeToConsole("Breaking because solution was found!\n");
			}
			break;
		    }
		    if (this->colIgnoreList[alternatives[i].first]>0){
			if (this->outputLevel>=OUTPUT_DEBUG){
			    writeToConsole("Skipping because column is in the ignore list!\n");
			}
			continue;
		    }
		    this->codewordManager->push_column(alternatives[i].first);
		    this->columnHashValue+=this->columnHashRandoms[alternatives[i].first];
		    if (PV && this->outputLevel>=OUTPUT_FEW){
			timeval time=getCurrentTime();
			double timeDiff=getTimeDiff(time,this->startTime);
			std::stringstream s;
			s<<"Examining "<<i+1<<"th branch of PV-node at n = "<<current_n<<": col = "<<alternatives[i].first<<" ("<<i+1<<"/"<<lastAlternative<<"), eval = "<<((double)alternatives[i].second.first)/EVAL_SHIFT<<", branch = "<<branch<<"."<<std::endl;
			s<<this->codewordManager->getCurrentGMAsString()<<std::endl;
			s<<"time: "<<timeDiff<<"s"<<std::endl;
			writeToConsole(s.str());
		    }
	            if (this->htmlTree){
    			std::stringstream comment;
    			comment<<"col = "<<alternatives[i].first<<" ["<<i+1<<"/"<<lastAlternative<<": "<<std::setprecision(3)<<exp((bestPair.second.first-alternatives[i].second.first)/EVAL_SHIFT)<<"]";
    			this->htmlTreeCreator.addEdge(nodeNumber, this->nodeCounter+1, comment.str(), this->evaluationCalls);
		    }
		    bool wasPruned=true;
		    if (learnSet.size()!=0){
			if (learnSet.find(alternatives[i].first)!=learnSet.end()){
			    learnSet.erase(alternatives[i].first);
			    wasPruned=this->recursiveSearch(current_n+1,learnSet,psm,(std::max)(maxBranchNumber,i+1),(std::max)(maxBranchDiff,bestPair.second.first-alternatives[i].second.first),branch*lastAlternative,((i==0)&PV));
			}
			else{
			    wasPruned=true;
			}
		    }
		    else{
			wasPruned=this->recursiveSearch(current_n+1,learnSet,psm,(std::max)(maxBranchNumber,i+1),(std::max)(maxBranchDiff,bestPair.second.first-alternatives[i].second.first),branch*lastAlternative,((i==0)&PV));
		    }
		    pruned|=wasPruned;
		    this->columnHashValue-=this->columnHashRandoms[alternatives[i].first];
		    this->codewordManager->pop_column();
		    if (!wasPruned && !this->foundSolution){
			if (this->outputLevel>=OUTPUT_DEBUG){
			    writeToConsole("No pruning was done, therefore the column was put on the ignore list!\n");
			}
			this->colIgnoreList[alternatives[i].first]++;
			nonChoiceableColumns.push_back(alternatives[i].first);
			ignoredColumns++;
		    }
		}
		while(ignoredColumns>0){
		    COLUMN_NO temp=this->nonChoiceableColumns.back();	
		    if (this->outputLevel>=OUTPUT_DEBUG){
			std::stringstream ss;
			ss<<"Removing column "<<temp<<" from the ignore list."<<std::endl;
			writeToConsole(ss.str());
		    }
		    this->nonChoiceableColumns.pop_back();
		    this->colIgnoreList[temp]--;
		    ignoredColumns--;
		}
	    }
	}
        else{
    	    if (this->htmlTree){
    	        std::stringstream comment;
    	        comment<<"Keine Lsungen mglich."<<std::endl;
	        this->htmlTreeCreator.addComment(nodeNumber,comment.str(),this->evaluationCalls);
	    }
	}
	//clean up
	{
	    unsigned int i=removedChoices.size();
	    while(i>0){
		if (this->outputLevel>=OUTPUT_DEBUG){
		    std::stringstream ss;
		    ss<<"Undoing the removal of column "<<removedChoices[i-1]->nr<<std::endl;
		    writeToConsole(ss.str());
		}
		removedChoices[i-1]->undoRemove();
		this->colIgnoreList[removedChoices[i-1]->nr]-=2;
		this->nonChoiceableColumns.pop_back();
  		i--;
	    }
	}
	if (current_n<=this->maxHashDepth){
            #ifdef HTMLTREE_HASHLINKS
    		this->evalHashTable->insert(std::pair<HASH_TYPE,HASH_MAP_SECOND_VAL>(HASH_TYPE(evalHashValue),HASH_MAP_SECOND_VAL(pruned,nodeNumber)),current_n);
	    #else
		this->evalHashTable->insert(std::pair<HASH_TYPE,HASH_MAP_SECOND_VAL>(HASH_TYPE(evalHashValue),pruned),current_n);
	    #endif
	}
	return pruned;
    }
    else{
	for (NumberListEntry* p=colChoice[0].nextEntry; p!=colChoice; p=p->nextEntry){
	    if (this->colIgnoreList[p->nr]>0){
		continue;
	    }
            EVAL_TYPE tempEval=this->codewordManager->getEval(p->nr);
            evaluationCalls++;
            if (tempEval>=0.0){
		if (this->htmlTree){
    		    std::stringstream solComment;
		    solComment<<"This is a solution!"<<std::endl;
		    this->htmlTreeCreator.addComment(nodeNumber,solComment.str(),this->evaluationCalls);
		}
	        timeval currTime=getCurrentTime();
		double time=getTimeDiff(currTime,this->startTime);
		std::stringstream ss;
		ss<<std::setprecision(3)<<std::showpoint;
		this->foundSolution=true;
                this->codewordManager->push_column(p->nr);
		ss<<"Found a solution after "<<time<< " seconds:\n";
		ss<<"maxBranchDiff = "<<exp(maxBranchDiff/EVAL_SHIFT)<<std::endl;
		ss<<"maxBranchNumber = "<<maxBranchNumber<<std::endl;
                writeToConsole(ss.str());
		std::string weightDistribution="weight enumerator: w(x) = "+this->codewordManager->getWeightDistributionAsString()+"\n";
		std::string symmWeightDistribution="symmetrized weight distribution = "+this->codewordManager->getSymmetrizedWeightDistributionAsString()+"\n";
                writeToConsole(weightDistribution);
                writeToConsole(symmWeightDistribution);
		writeToConsole("NHC distribution: "+this->codewordManager->getNHCDistributionAsString()+"\n");
                this->solution=this->codewordManager->getCurrentGMAsString();
		if (this->solutionFile!=""){
		    std::ofstream solfile;
		    solfile.open(this->solutionFile.c_str());
		    if (solfile.is_open()){
			solfile<<std::setprecision(3)<<std::showpoint;
			solfile<<"%solution found by HeuricoII."<<std::endl;
			solfile<<getHeuricoInformation();
			solfile<<"%nodes = "<<this->nodeCounter<<std::endl;
			solfile<<"%evaluations = "<<this->evaluationCalls<<std::endl;
			solfile<<"%time = "<<time<<"s"<<std::endl;
			solfile<<"%a = "<<exp((double)this->alterBound/EVAL_SHIFT)<<std::endl;
			solfile<<"%m = "<<this->maxBranch<<std::endl;
			solfile<<"%maxhd = "<<this->maxHashDepth<<std::endl;
			solfile<<"%ehs = "<<this->evalHashSize<<std::endl;
			solfile<<"%chs = "<<this->columnHashSize<<std::endl;
			solfile<<"%columns: "<<this->codewordManager->getChosenColumnsAsString();
			solfile<<"%"<<weightDistribution<<std::endl;
			solfile<<this->solution;
			solfile.close();
		    }
		    else{
			writeToConsole(std::string("Error: could not open solution file!\n"));
		    }
		}
                this->codewordManager->pop_column();
                return false;
            }
        }
	return false;
    }
}

template<class CHAINRING>
std::vector<int> SearchAlgorithm<CHAINRING>::getPath(){
    return this->path;
}

template<class CHAINRING>
void SearchAlgorithm<CHAINRING>::setPath(unsigned int current_n, unsigned int i){
    this->path[current_n]=(int)i;
    this->path[current_n+1]=-1;
}

template<class CHAINRING>
void SearchAlgorithm<CHAINRING>::resetPath(){
    path.resize(MAX_LENGTH+2);
    for (unsigned int i=0; i<MAX_LENGTH+2; i++){
	path[i]=-1;
    }
}

#ifdef STORE_ISOMORPHIC_GMS
template<class CHAINRING>
std::string SearchAlgorithm<CHAINRING>::generateMagmaScript(GM_STORAGE::iterator& iterClass, unsigned int _n){
    if (this->outputLevel>=OUTPUT_DEBUG){
	std::stringstream ss;
	ss<<"SearchAlgorithm<CHAINRING>::generateMagmaScript()"<<std::endl;
	ss<<"_n = "<<_n<<std::endl;
	writeToConsole(ss.str());
    }
    std::stringstream ss;
    std::string empty_str("");
    std::string comma_str(",");
    std::vector<std::string>::iterator iterStrings=iterClass->second.begin();
    std::vector<std::string>::iterator endIterStrings=iterClass->second.end();
    unsigned int counter=0;
    ss<<"k:="<<this->codewordManager->getMessageLength()<<";"<<std::endl;
    while(iterStrings!=endIterStrings){
	std::string temp=*iterStrings;
	temp=ReplaceString(std::string("    "),comma_str,temp);
	temp=ReplaceString(std::string("   "),comma_str,temp);
	temp=ReplaceString(std::string("  "),comma_str,temp);
	temp=ReplaceString(std::string(" "),comma_str,temp);
	temp=ReplaceString(std::string("\n\n"),empty_str,temp);
	temp=ReplaceString(std::string("\n"),empty_str,temp);
	temp=temp.substr(temp.find(",")+1);
	unsigned int commas=countOccurence(comma_str,temp);
	unsigned int length=(commas+1)/this->codewordManager->getMessageLength();
	if (length!=_n){
	    return "";
	}
	ss<<"tmp_G:=Matrix(R,"<<length<<",["<<temp<<"]);"<<std::endl;
	if (counter==0){
	    ss<<"G_first:=tmp_G;";
    	    ss<<"backtrack_start(tmp_G, ~tmp_best, ~best_vals, ~tmp_autom, ~tmp_transporter_el, inner_stab, \"\",~equivalent);"<<std::endl;
	}
	else{
	    ss<<"backtrack_start(tmp_G, ~tmp_best, ~best_vals, ~tmp_autom, ~tmp_transporter_el, inner_stab, \"\",~equivalent: checkEquivalence:=true);"<<std::endl;
	    ss<<"if equivalent ne 0 then print G_first,\"\\nis not equivalent to\\n\",tmp_G,\"\\n\"; end if;"<<std::endl;
	}
	iterStrings++;
	counter++;
    }
    return ss.str();
}
#endif

#endif

