/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/CommonRoutines.h"
#include "../h/Log.h"

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <map>
#include <limits>

#ifdef LINUX
    #include <sys/time.h>
    #include <stdio.h>
    #include <termios.h>
    #include <unistd.h>
#else 
    #define NOMINMAX //otherwise std::numeric_limits<...> does not work
    #include <winsock2.h>
    #include <sys/timeb.h>
    #include <conio.h>
#endif

using namespace std;

Problem* constructProblemFromSStream(stringstream& f){
    Problem* p=new Problem();
    string s;
    {
	getline(f,s);
	string::size_type t=0;
	while(s[t]==' '){t++;}
	p->matrix.rows=atoi(s.substr(t,s.length()-t).c_str());
	while(s[t]!=' '){t++;}
	while(s[t]==' '){t++;}
	p->matrix.cols=atoi(s.substr(t,s.length()-t).c_str());
	if (p->matrix.rows<=0){
	    stringstream s2;
	    s2<<"Error in constructProblemFromSStream(): rows<0!"<<std::endl;
	    writeToConsole(s2);
	    p->clear();
	    return NULL;
	}
	if (p->matrix.cols<=0){
	    stringstream s2;
	    s2<<"Error in constructProblemFromSStream(): cols<=0!"<<std::endl;
	    writeToConsole(s2);
	    p->clear();
	    return NULL;
	}
	p->matrix.data=new m_entry[p->matrix.rows*p->matrix.cols];
	p->matrix.lowerBound=new sum_of_entries[p->matrix.rows];
	p->matrix.upperBound=new sum_of_entries[p->matrix.rows];
	p->controlRow=new m_entry[p->matrix.cols];
    }
    {
	unsigned int row,col;
	std::map<m_entry,unsigned int> CREntries;
	for (row=0; row<=p->matrix.rows; row++){
	    getline(f,s);
	    if (s==""){
		stringstream s2;
		s2<<"Error in constructProblemFromSStream(): missing row!"<<std::endl;
		writeToConsole(s2);
		p->clear();
		return NULL;
	    }
	    string::size_type t,t2;
	    t=s.find_first_of("[");
	    t2=s.find_first_of("]");
	    if (t==string::npos || t2==string::npos){
		stringstream s2;
		s2<<"Error in constructProblemFromSStream(): no RHS for row no. "<<row<<std::endl;
		writeToConsole(s2);
		p->clear();
		return NULL;
	    }
	    string rhs=s.substr(t+1,t2-t-1);
	    s=s.substr(0,t)+s.substr(t2+1,s.length()-t2-1);
	    t2=rhs.find_first_of(",")+1;
	    if (row<p->matrix.rows){
		p->matrix.lowerBound[row]=atoi(rhs.c_str());
		std::string ubstr=rhs.substr(t2,rhs.length()-t2);
		if (ubstr=="INF"){
		    p->matrix.upperBound[row]=NO_UPPER_BOUND;
		}
		else{
		    p->matrix.upperBound[row]=atoi(ubstr.c_str());
		}
	    }
	    else{
		p->CRLowerBound=atoi(rhs.c_str());
		p->CRUpperBound=atoi(rhs.substr(t2,rhs.length()-t2).c_str());
	    }
	    t=0;
	    for (col=0; col<p->matrix.cols; col++){
		while(t<s.length() && s[t]==' '){
		    t++;
		}
		if (t>=s.length()){
		    stringstream s2;
		    s2<<"Error in constructProblemFromSStream(): incomplete row! ("<<row<<")"<<std::endl;
		    writeToConsole(s2);
		    p->clear();
		    return NULL;
		}
		int val=atoi(s.substr(t,s.length()-t).c_str());
		if (val>std::numeric_limits<m_entry>::max() || val<std::numeric_limits<m_entry>::min()){
		    stringstream s2;
		    s2<<"Error: at least one matrix error exceeds the interval ["<<(int)std::numeric_limits<m_entry>::min()<<","<<(int)std::numeric_limits<m_entry>::max()<<"]."<<std::endl;
		    #ifndef FOUR_BYTES_PER_ENTRY
			#ifdef TWO_BYTES_PER_ENTRY
			    s2<<"Please use the 4bpe version instead."<<std::endl;
			#else
			    s2<<"Please use the 2bpe or 4bpe version instead."<<std::endl;
			#endif
		    #endif
		    writeToConsole(s2);
		    return NULL;
		}
		if (row<p->matrix.rows){
		    p->matrix.data[col*p->matrix.rows+row]=val;
		}
		else{
		    p->controlRow[col]=val;
		    if (CREntries.find(p->controlRow[col])==CREntries.end()){
			CREntries.insert(std::pair<m_entry,unsigned int>(p->controlRow[col],col));
		    }
		}	
		t=s.find(' ',t);
	    }
	    p->differentCREntries=CREntries.size();
	    for (std::map<m_entry,unsigned int>::reverse_iterator iter=CREntries.rbegin(); iter!=CREntries.rend(); iter++){
		p->CRPhaseValue.push_back(p->controlRow[iter->second]);
	    }
	}
    }
    return p;
}

void appendVarBoundInequations(Problem* p, int u){
    int newRows=p->matrix.rows+p->matrix.cols;
    int newCols=p->matrix.cols;
    m_entry* newData=new m_entry[newRows*newCols];
    sum_of_entries* newLB=new sum_of_entries[newRows];
    sum_of_entries* newUB=new sum_of_entries[newRows];
    for (unsigned int i=0; i<p->matrix.rows; i++){
	for (unsigned int j=0; j<p->matrix.cols; j++){
	    newData[j*newRows+i]=p->matrix.data[j*p->matrix.rows+i];
	}
	newLB[i]=p->matrix.lowerBound[i];
	newUB[i]=p->matrix.upperBound[i];
    }
    delete[] p->matrix.data;
    delete[] p->matrix.lowerBound;
    delete[] p->matrix.upperBound;
    p->matrix.data=newData;
    p->matrix.lowerBound=newLB;
    p->matrix.upperBound=newUB;
    p->matrix.rows=newRows;

    for (unsigned int i=p->matrix.rows-p->matrix.cols; i<p->matrix.rows; i++){
	for (unsigned int j=0; j<p->matrix.cols; j++){
	    if (i-(p->matrix.rows-p->matrix.cols)==j){
		p->matrix.data[j*p->matrix.rows+i]=1;
	    }
	    else{
		p->matrix.data[j*p->matrix.rows+i]=0;
	    }
	}
	p->matrix.lowerBound[i]=0;
	p->matrix.upperBound[i]=u;
    }
}

string ReplaceString(const string &stringSearchString, const string &stringReplaceString, string stringStringToReplace)
{   
	//stringSearchString = string der mit stringReplaceString ersetzt wird
	//stringStringToReplace ist der string in dem gesucht und ersetzt wird

	string::size_type pos = stringStringToReplace.find(stringSearchString, 0);
	int intLengthSearch = stringSearchString.length();

	while(string::npos != pos )
	{

		stringStringToReplace.replace(pos, intLengthSearch, stringReplaceString);
		pos = stringStringToReplace.find(stringSearchString, 0);

	}

	return stringStringToReplace;
}

ChoiceEntry* computeChoices(unsigned int numberOfColumns, m_entry* lhs, sum_of_entries min_sum, sum_of_entries max_sum)
{
	ChoiceEntry* table=new ChoiceEntry[max_sum+1];
	computeChoices_recursive(numberOfColumns, lhs,min_sum,max_sum, table);
	return table;
}

unsigned int computeChoices_recursive(unsigned int numberOfColumns, m_entry* lhs, sum_of_entries min_sum, sum_of_entries max_sum, ChoiceEntry* table){
	if (table[max_sum].computed==true){
		return table[max_sum].choices.size()+(min_sum<=0);
	}
	for (unsigned int j=0; j<numberOfColumns; j++){
		if (lhs[j]>max_sum || lhs[j]==0){
			continue;
		}	
		unsigned n=computeChoices_recursive(numberOfColumns, lhs, min_sum-lhs[j],max_sum-lhs[j],table);
		if (n>0){
			table[max_sum].choices.push_back(j);
		}
	}
	if (table[max_sum].choices.size()>0 || min_sum<=0){
		for (unsigned int j=0; j<numberOfColumns; j++){
			if (lhs[j]==0){
				table[max_sum].choices.push_back(j);
			}
		}
	}
	table[max_sum].computed=true;
	return table[max_sum].choices.size()+(min_sum<=0);
}

unsigned int twistUint(unsigned int u){
    const unsigned long long mod=4294967291LL; //thats the biggest prime below 2^32
    const unsigned long long lin=2147483647LL; //2^31-1; thats a prime, too, but this should not matter
    if (u>=mod){
	return u;
    }
    unsigned long long u_as_ll=(unsigned long long)(u);
    return (unsigned int)(((u_as_ll*u_as_ll)+u_as_ll*lin)%mod);
}

unsigned long long hashTwist(unsigned long long i){
    unsigned int l=(unsigned int)(i&0x00000000FFFFFFFFLL);
    unsigned int h=(unsigned int)((i&0xFFFFFFFF00000000LL)>>32);
    return (((unsigned long long)(twistUint(h)))<<32)+twistUint(l);    
}

void waitSomeTime(unsigned int nanosecs){
    #ifdef LINUX
	timespec t;
	t.tv_sec=0; //seconds
	t.tv_nsec=nanosecs;
	nanosleep(&t,NULL);
    #else //windows version
	unsigned int ms=(std::max)(1,(int)(nanosecs/1000000));
	::Sleep(ms);
    #endif
}

MyFPType getBinomialCoeff(int n, int k){
    MyFPType res(1.0);
    if (k>n-k){
    	k=n-k;
    }
    for (int i=1; i<=k; i++){
	MyFPType x((double)(n-i+1));
	MyFPType y((double)(i));
	res*=x;
	res/=y;
    }
    return res;
}

NumberListEntry* getColumnList(Problem* p){
    NumberListEntry* res=new NumberListEntry[p->matrix.cols+1];
    res[0].nr=0xFFFFFFFF;
    res[0].nextEntry=&res[1];
    res[0].previousEntry=&res[p->matrix.cols];
    for (unsigned int j=1; j<p->matrix.cols; j++){
	res[j].nr=j-1;
	res[j].previousEntry=&res[j-1];
	res[j].nextEntry=&res[j+1];
    }
    res[p->matrix.cols].nr=p->matrix.cols-1;
    res[p->matrix.cols].previousEntry=&res[p->matrix.cols-1];
    res[p->matrix.cols].nextEntry=&res[0];
    return res;
}

NumberListEntry* getOrderedColumnList(Problem* p){
    std::vector<CRColumnNumberPair> pairs;
    pairs.resize(p->matrix.cols);
    for (unsigned int j=0; j<p->matrix.cols; j++){
	pairs[j]=std::pair<m_entry, unsigned int>(p->controlRow[j],j);
    }
    std::sort(pairs.begin(),pairs.end(),compareCRColumnNumberPairs);
    NumberListEntry* res=new NumberListEntry[p->matrix.cols+1];
    res[0].nr=0xFFFFFFFF;
    res[0].nextEntry=&res[1];
    res[0].previousEntry=&res[p->matrix.cols];
    for (unsigned int j=1; j<p->matrix.cols; j++){
	res[j].nr=pairs[j-1].second;
	res[j].previousEntry=&res[j-1];
	res[j].nextEntry=&res[j+1];
    }
    res[p->matrix.cols].nr=pairs[p->matrix.cols-1].second;
    res[p->matrix.cols].previousEntry=&res[p->matrix.cols-1];
    res[p->matrix.cols].nextEntry=&res[0];
    return res;
}

RowListEntry* getRowList(Problem* p){
    RowListEntry* res=new RowListEntry[p->matrix.rows+1];
    res[0].nr=0xFFFFFFFF;
    res[0].header=res;
    res[0].nextEntry=&res[1];
    res[0].previousEntry=&res[p->matrix.rows];
    for (unsigned int j=1; j<=p->matrix.rows; j++){
	res[j].nr=j-1;
	res[j].header=res;
	res[j].previousEntry=&res[j-1];
	if (j<p->matrix.rows){
	    res[j].nextEntry=&res[j+1];
	}
	else{
	    res[j].nextEntry=&res[0];
	}
	res[j].correspondingLowerBound=p->matrix.lowerBound[j-1];
	res[j].correspondingUpperBound=p->matrix.upperBound[j-1];
	res[j].correspondingRHS=0;
    }
    return res;
}

bool compareCRColumnNumberPairs(CRColumnNumberPair a, CRColumnNumberPair b){
    if (a.first>=b.first){
	if (a.first>b.first || a.second<=b.second){
	    return true;
	}
    }
    return false;
};

