/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/ExactProbEvaluator.h"
#include "../h/Probtable.h"
#include "../h/Log.h"
#include "../h/Solver.h"
#include "../h/TimeManager.h"
#include "../h/Random.h"
#include "../h/commonStructs.h"
#include "../h/Polynomial.h"
#include "../h/CommonRoutines.h"
#include <iostream>
#include <map>
#include <math.h>
#include <limits>
using namespace std;

ExactProbEvaluator::ExactProbEvaluator() : problem(NULL){
}

ExactProbEvaluator::~ExactProbEvaluator(){
    this->freeMemory();
}

void ExactProbEvaluator::freeMemory(){
    if (this->problem!=NULL){
	for (unsigned int r=0; r<this->problem->matrix.rows; r++){
	    if (!this->rowInfoAllocated[r]){
		continue;
	    }
	    delete this->rowInfo[r];
	}
	delete[] this->rowInfoAllocated;
	delete[] this->rowInfo;
	this->problem=NULL;
    }
}

bool ExactProbEvaluator::getRowIntervals(RowIntervalTable** rivt, bool* rivtAlloc){
    if (this->problem==NULL){
	return false;
    }
    for (unsigned int r=0; r<this->problem->matrix.rows; r++){
	if (this->rowInfoAllocated[r]){
	    rivt[r]=new RowIntervalTable();
	    rivtAlloc[r]=true;
	    rivt[r]->init(this->problem->differentCREntries,this->problem->CRUpperBound+1);
	    for (unsigned int i=0; i<this->problem->differentCREntries; i++){ 
		for (int j=0; j<=this->problem->CRUpperBound; j++){
		    rivt[r]->riv[i][j].minChange=this->rowInfo[r]->minChange[i][j];
		    rivt[r]->riv[i][j].maxChange=this->rowInfo[r]->maxChange[i][j];
		}
	    }
	}
	else{
	    rivtAlloc[r]=false;
	    for (unsigned int r2=0; r2<r; r2++){
		if (this->rowInfo[r2]==this->rowInfo[r]){
		    rivt[r]=rivt[r2];
		}
	    }
	}
    }
    return true;
}

Problem* ExactProbEvaluator::getProblem(){
    return this->problem;
}

void ExactProbEvaluator::setProblem(Problem* p){
    this->freeMemory();
    this->problem=p;
    this->rowInfo=new RowInformation*[p->matrix.rows];
    this->rowInfoAllocated=new bool[p->matrix.rows];
    std::stringstream s;
    s<<"ExactProbEvaluator: initializing..."<<std::endl;
    writeToConsole(s);
    std::map<unsigned long long, unsigned int> hash;
    std::map<std::pair<m_entry,m_entry>,unsigned long long> randoms;
    std::map<std::pair<m_entry,m_entry>,unsigned long long>::iterator randIter;
    NumberListEntry* colList=getOrderedColumnList(this->problem);
    for (unsigned int r=0; r<p->matrix.rows; r++){
	unsigned long long hashValue=0;
	std::map<std::pair<m_entry,m_entry>, unsigned int> combinations;
	bool rowEntriesNonNegative=true;
	for (NumberListEntry* c=colList[0].nextEntry; c!=colList; c=c->nextEntry){
	    m_entry rowEntry=this->problem->matrix.data[c->nr*this->problem->matrix.rows+r];
	    if (rowEntry<0){
		rowEntriesNonNegative=false;
	    }
	    std::pair<m_entry,m_entry> combi(this->problem->controlRow[c->nr],rowEntry);
	    std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator combIter=combinations.find(combi);
	    if (combIter==combinations.end()){
		combinations.insert(std::pair<std::pair<m_entry,m_entry>,unsigned int>(combi,1));
	    }
	    else{
		combIter->second++;
	    }
	}

	for (std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator combIter=combinations.begin(); combIter!=combinations.end(); combIter++){
	    unsigned long long tempRand=(unsigned long long)(get_rand()*0xFFFFFFFFFFFFFFFFULL);
	    randIter=randoms.find(std::pair<m_entry,m_entry>(combIter->first));
	    if (randIter==randoms.end()){
		randoms.insert(std::pair<std::pair<m_entry,m_entry>,unsigned long long>(combIter->first,tempRand));
	    }
	    else{
		tempRand=randIter->second;
	    }
	    hashValue+=combIter->second*tempRand;
	}
	hashValue+=(unsigned long long)this->problem->matrix.lowerBound[r]+(((unsigned long long)this->problem->matrix.upperBound[r])<<32);

	std::map<unsigned long long, unsigned int>::iterator hashIter=hash.find(hashValue);
	if (hashIter!=hash.end()){
	    this->rowInfo[r]=this->rowInfo[(*hashIter).second];
	    this->rowInfoAllocated[r]=false;
	    continue;
	}
	else{
	    hash.insert(pair<unsigned long long, unsigned int>(hashValue,r));
	    this->rowInfo[r]=new RowInformation();
	    this->rowInfo[r]->evalTable.init(p->differentCREntries,p->CRUpperBound+1);
	    this->rowInfo[r]->minChange.init(p->differentCREntries,p->CRUpperBound+1);
	    this->rowInfo[r]->maxChange.init(p->differentCREntries,p->CRUpperBound+1);
	    this->rowInfo[r]->allEntriesNonNegative=rowEntriesNonNegative;
	    this->rowInfoAllocated[r]=true;
	}
	std::map<intpair,fppair> terms;
	terms.insert(pair<intpair,fppair>(intpair(0,0),fppair(1.0,1.0)));
	//intpair (a,b): a=exponent for control row; b=exponent for row
	//fppair (c,d): d=coeff from summation before; c=new coeff to add
	MyFPType one(1.0);
	int CRPhase=this->problem->differentCREntries-1;
	int yModulus=std::numeric_limits<int>::max(); //to unify the y-component where the exact value does not matter
	if (this->rowInfo[r]->allEntriesNonNegative){
	    if (p->matrix.upperBound[r]==NO_UPPER_BOUND){
		yModulus=p->matrix.lowerBound[r];
	    }
	    else{
		yModulus=p->matrix.upperBound[r]+1;
	    }
	}

	for (std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator combIter=combinations.begin(); combIter!=combinations.end();){
	    //determine maximum power for this combination
	    int k=1;
	    int l=combIter->second;
	    while(true){
		if (k*combIter->first.first>p->CRUpperBound){
		    k--;
		    break;
		}
		k++;
	    }
	    UnivariatePolynomial uvp;
	    uvp.coeff.resize(k+1);
	    for (int i=0; i<=k; i++){
		uvp.coeff[i]=getBinomialCoeff(l+i-1,l-1);
	    }

	    BivariatePolynomial bvp=replaceByXiYj(uvp,combIter->first.first,combIter->first.second);
	    for (int i=0; i<bvp.size(); i++){ 
		if (bvp.monoms[i].second.first==0 && bvp.monoms[i].second.second==0){ //skip first entry, thats the constant 1
		    continue; 
		}
		if (bvp.monoms[i].second.first>p->CRUpperBound){
		    continue;
		}
		if (bvp.monoms[i].first.getDoubleValue()==0.0){
		    continue;
		}

		std::map<intpair,fppair>::iterator iter=terms.begin();
		std::map<intpair,fppair>::const_iterator endIter=terms.end();
		while(iter!=endIter){
		    if (iter->second.second.getDoubleValue()!=0.0){
			intpair newexp(iter->first.first+bvp.monoms[i].second.first,iter->first.second+bvp.monoms[i].second.second);
			if (newexp.first<=p->CRUpperBound){
			    newexp.second=std::min(newexp.second,yModulus);
			    std::map<intpair,fppair>::iterator temp=terms.find(newexp);
			    if (temp!=terms.end()){
				(*temp).second.first+=iter->second.second*bvp.monoms[i].first;
			    }
			    else{
				terms.insert(pair<intpair,fppair>(newexp,fppair(iter->second.second*bvp.monoms[i].first,0.0)));
			    }
			}
		    }
		    iter++;
		}
	    }

	    //set second component of the terms to first one
	    std::map<intpair,fppair>::iterator iter=terms.begin();
	    std::map<intpair,fppair>::const_iterator endIter=terms.end();
	    while(iter!=endIter){
		(*iter).second.second=(*iter).second.first;
		iter++;		
	    }
	    m_entry lastCR=combIter->first.first;
	    combIter++;
	    if (combIter==combinations.end()){
		this->setRowInformation(this->rowInfo[r],this->problem->matrix.lowerBound[r],this->problem->matrix.upperBound[r],CRPhase,terms);
		break;
	    }
	    else{
		if (combIter->first.first!=lastCR){
		    this->setRowInformation(this->rowInfo[r],this->problem->matrix.lowerBound[r],this->problem->matrix.upperBound[r],CRPhase,terms);
		    CRPhase--;
		}
	    }
	}

    }
    delete[] colList;
    s.clear();
    s.str("");
    s<<"ExactProbEvaluator: initialization done."<<std::endl;
    s<<"time: "<<TimeManager::timeSinceStart()<<"s"<<std::endl;
    writeToConsole(s);
}

void ExactProbEvaluator::setRowInformation(RowInformation* rowInfo, sum_of_entries rowLB, sum_of_entries rowUB, unsigned int CRPhase, std::map<intpair,fppair>& terms){
     Problem* p=this->problem;
    std::vector<int> llimit(p->CRUpperBound+1);
    std::vector<int> ulimit(p->CRUpperBound+1);
    for (int i=0; i<=p->CRUpperBound; i++){
	llimit[i]=0x7FFFFFFF;
	ulimit[i]=-0x7FFFFFFF;
    }
    std::map<intpair,fppair>::iterator iter=terms.begin();
    std::map<intpair,fppair>::const_iterator endIter=terms.end();
    while(iter!=endIter){
	for (int i=(std::max)(p->CRLowerBound-iter->first.first,0); i<=p->CRUpperBound-iter->first.first; i++){
	    if (iter->first.second<llimit[i]){
		llimit[i]=iter->first.second;
	    }
	    if (iter->first.second>ulimit[i]){
		ulimit[i]=iter->first.second;
	    }
	}
	iter++;
    }

    ProbTable* temp=new ProbTable[p->CRUpperBound+1];
    for (int i=0; i<=p->CRUpperBound; i++){
	if (llimit[i]==0x7FFFFFFF){
	    temp[i].init(PROBTABLE_NOT_RELEVANT,PROBTABLE_NOT_RELEVANT);
	}
	else{
	    temp[i].init(llimit[i],ulimit[i]);
	}
    }
    iter=terms.begin();
    endIter=terms.end();
    while(iter!=endIter){
	for (int i=(std::max)(p->CRLowerBound-iter->first.first,0); i<=p->CRUpperBound-iter->first.first; i++){
	    for (int m=iter->first.second; m<=ulimit[i]; m++){
		temp[i].add_val(m,iter->second.first);
	    }
	}
	iter++;
    }
    for (int i=0; i<=p->CRUpperBound; i++){
	if (!(temp[i].get_llimit()==PROBTABLE_NOT_RELEVANT)){
	    temp[i].normalize();
	    rowInfo->minChange[CRPhase][i]=temp[i].get_llimit();
	    rowInfo->maxChange[CRPhase][i]=temp[i].get_ulimit();
	    //now compute a new ProbTable which directly contains the success probabilities instead of a prob.-distribution
	    int lb=rowLB-temp[i].get_ulimit();
	    if (rowInfo->allEntriesNonNegative){
		lb=(std::max)(0,lb);
	    }
	    int ub;
	    if (rowUB!=NO_UPPER_BOUND){ //compute the limits lb,ub for the current row sum where the prob. might be >0
		ub=rowUB-temp[i].get_llimit();
		rowInfo->evalTable[CRPhase][i].init(lb,ub,MINUS_INFTY_LL,MINUS_INFTY_LL);
	    }
	    else{
		ub=rowLB-temp[i].get_llimit()-1;
		rowInfo->evalTable[CRPhase][i].init(lb,ub,MINUS_INFTY_LL,0);
	    }
	    for (int j=lb; j<=ub; j++){
		int l=rowLB-j;
		int u=rowUB-j;
		double val=temp[i].get_val(u)-temp[i].get_val(l-1);
		if (val!=0.0){
		    rowInfo->evalTable[CRPhase][i].set_val(j,(EVAL_TYPE)(log(val)*EVAL_SHIFT));
		}
		else{
		    rowInfo->evalTable[CRPhase][i].set_val(j,MINUS_INFTY_LL);
		}
	    }
	}
	else{
	    rowInfo->evalTable[CRPhase][i].init(0,-1,MINUS_INFTY_LL,MINUS_INFTY_LL);
	    rowInfo->minChange[CRPhase][i]=NO_UPPER_BOUND;
	    rowInfo->maxChange[CRPhase][i]=NO_LOWER_BOUND;
	}
    }
    delete[] temp;
}

EVAL_TYPE ExactProbEvaluator::getEvaluation(unsigned int c, RowListEntry* relevantRows, unsigned int CRPhase, sum_of_entries currentCR){
    RowListEntry* rle;
    if (currentCR>this->problem->CRUpperBound){
	return MINUS_INFTY_LL;
    }
    EVAL_TYPE eval=0;
    int index=c*this->problem->matrix.rows;
    for (rle=relevantRows->nextEntry; rle!=relevantRows; rle=rle->nextEntry){
	EvalTable& et=this->rowInfo[rle->nr]->evalTable[CRPhase][currentCR];
	EVAL_TYPE val=et.get_val(rle->correspondingRHS+this->problem->matrix.data[index+rle->nr]);
	if (val==MINUS_INFTY_LL){
	    return MINUS_INFTY_LL;
	}
	eval+=val;
    }
    return eval;
}

