/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/HeuristicSolveAlgorithm.h"
#include "../h/CommonRoutines.h"
#include "../h/Random.h"
#include "../h/Log.h"
#include "../h/Solver.h"
#include "../h/TimeManager.h"

#include <math.h>
#include <iostream>
#include <set>
#include <map>
#include <algorithm>
#include <limits>

using namespace std;
typedef pair<int,int> intpair;

HeuristicSolveAlgorithm::HeuristicSolveAlgorithm(){
    this->setForbidColumns(false);
    this->setUseEvalHash(true);
    this->setAlternativeBound(DEFAULT_ALTERNATIVE_BOUND);
    this->setBacktrackMargin(DEFAULT_BACKTRACK_MARGIN);
    this->setMaxBranch(DEFAULT_MAX_BRANCH);
    this->setUpdateRate(DEFAULT_UPDATE_RATE);
    this->setRedundancyCheck(true);
    this->setRandomNoise(false);
    this->setSplitLimit(1);
    this->problem=NULL;
    this->CRPhaseEntries=NULL;
    this->adaptToTime=false;
}

HeuristicSolveAlgorithm::~HeuristicSolveAlgorithm(){
    this->freeMemory();
}

void HeuristicSolveAlgorithm::freeMemory(){
    if (this->problem!=NULL){
	if (this->getRedundancyCheck()){
	    for (unsigned int r=0; r<this->problem->matrix.rows; r++){
		if (!this->rivtMemAllocated[r]){
		    continue;
		}
		delete this->rivt[r];
	    }
	    delete[] this->rivtMemAllocated;
	    delete[] this->rivt;
	    this->problem=NULL;
	}
    }    
}

void HeuristicSolveAlgorithm::setEvaluator(AbstractEvaluator* e){
    this->itsEvaluator=e;
}

AbstractEvaluator* HeuristicSolveAlgorithm::getEvaluator(){
    return this->itsEvaluator;
}

void HeuristicSolveAlgorithm::setAlternativeBound(double a){
    this->alternativeBound=a;
    this->maxAlternativeDiff=(EVAL_TYPE)(log(this->alternativeBound)*EVAL_SHIFT);
}

double HeuristicSolveAlgorithm::getAlternativeBound(){
    return this->alternativeBound;
}

void HeuristicSolveAlgorithm::setMaxBranch(unsigned int m){
    this->maxBranch=m;
}

unsigned int HeuristicSolveAlgorithm::getMaxBranch(){
    return this->maxBranch;
}

void HeuristicSolveAlgorithm::setLastUpdate(double t){
    this->lastUpdate=t;
}

double HeuristicSolveAlgorithm::getLastUpdate(){
    return this->lastUpdate;
}

void HeuristicSolveAlgorithm::setUpdateRate(double u){
    this->updateRate=u;
}

double HeuristicSolveAlgorithm::getUpdateRate(){
    return this->updateRate;
}

void HeuristicSolveAlgorithm::setForbidColumns(bool f){
    this->forbidColumns=f;
}

bool HeuristicSolveAlgorithm::getForbidColumns(){
    return this->forbidColumns;
}

void HeuristicSolveAlgorithm::setUseEvalHash(bool f){
    this->useEvalHash=f;
}

bool HeuristicSolveAlgorithm::getUseEvalHash(){
    return this->useEvalHash;
}

void HeuristicSolveAlgorithm::setRedundancyCheck(bool r){
    this->redundancyCheck=r;
}

bool HeuristicSolveAlgorithm::getRedundancyCheck(){
    return this->redundancyCheck;
}

void HeuristicSolveAlgorithm::setRandomNoise(bool r){
    this->randomNoise=r;
}

bool HeuristicSolveAlgorithm::getRandomNoise(){
    return this->randomNoise;
}
void HeuristicSolveAlgorithm::setAdaptToTime(bool a){
    this->adaptToTime=true;
}

bool HeuristicSolveAlgorithm::getAdaptToTime(){
    return this->adaptToTime;
}

EVAL_TYPE HeuristicSolveAlgorithm::getMaxAlternativeDiff(){
    return this->maxAlternativeDiff;
}

void HeuristicSolveAlgorithm::undeleteColumns(unsigned int sizeBefore){
    for (int j=(int)(this->deletedColumns.size())-1; j>=(int)sizeBefore; j--){
	deletedColumns[j]->undoRemove();
    }
    deletedColumns.resize(sizeBefore);
}

void HeuristicSolveAlgorithm::undeleteRows(vector<RowListEntry*>& deletedRows){
    for (int j=(int)deletedRows.size()-1; j>=0; j--){
	deletedRows[j]->undoRemove();
    }
}

void HeuristicSolveAlgorithm::computeRowIntervals(){
    std::stringstream s;
    this->rivt=new RowIntervalTable*[this->problem->matrix.rows];
    this->rivtMemAllocated=new bool[this->problem->matrix.rows];
    if (this->itsEvaluator->getRowIntervals(this->rivt,this->rivtMemAllocated)){
	s<<"Copied the row intervals from the evaluation engine."<<std::endl;
	writeToConsole(s);
	return;
    }
    s<<"HeuristicSolveAlgorithm: computing row intervals...";
    writeToConsole(s);
    std::map<unsigned int, unsigned int> hash;
    std::map<std::pair<m_entry,m_entry>,unsigned int> randoms;
    std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator randIter;
    Problem* p=this->problem;
    NumberListEntry* colList=getOrderedColumnList(p);
    for (unsigned int r=0; r<p->matrix.rows; r++){
	unsigned int hashValue=0;
	bool allEntriesNonNegative=true;
	std::map<std::pair<m_entry,m_entry>, unsigned int> combinations;
	for (NumberListEntry* c=colList[0].nextEntry; c!=colList; c=c->nextEntry){
	    m_entry rowEntry=p->matrix.data[c->nr*p->matrix.rows+r];
	    if (rowEntry<0){
		allEntriesNonNegative=false;
	    }
	    std::pair<m_entry,m_entry> combi(p->controlRow[c->nr],rowEntry);
	    std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator combIter=combinations.find(combi);
	    if (combIter==combinations.end()){
		combinations.insert(std::pair<std::pair<m_entry,m_entry>,unsigned int>(combi,1));
	    }
	    else{
		combIter->second++;
	    }
	}

	for (std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator combIter=combinations.begin(); combIter!=combinations.end(); combIter++){
	    unsigned int tempRand=(unsigned int)(get_rand()*0xFFFFFFFF);
	    randIter=randoms.find(std::pair<m_entry,m_entry>(combIter->first));
	    if (randIter==randoms.end()){
		randoms.insert(std::pair<std::pair<m_entry,m_entry>,unsigned int>(combIter->first,tempRand));
	    }
	    else{
		tempRand=randIter->second;
	    }
	    hashValue+=combIter->second*tempRand;
	}
	hashValue+=p->matrix.lowerBound[r]+p->matrix.upperBound[r]*(1<<16);

	std::map<unsigned int, unsigned int>::iterator hashIter=hash.find(hashValue);
	if (hashIter!=hash.end()){
	    this->rivt[r]=this->rivt[(*hashIter).second];
	    this->rivtMemAllocated[r]=false;
	    continue;
	}
	else{
	    hash.insert(pair<unsigned int, unsigned int>(hashValue,r));
	    this->rivt[r]=new RowIntervalTable();
	    this->rivt[r]->init(p->differentCREntries,this->problem->CRUpperBound+1);
	    this->rivtMemAllocated[r]=true;
	}
	std::map<intpair,fppair> terms;
	terms.insert(pair<intpair,fppair>(intpair(0,0),fppair(1.0,1.0)));
	//intpair (a,b): a=exponent for control row; b=exponent for row
	//fppair (c,d): d=coeff from summation before; c=new coeff to add
	MyFPType one(1.0);
	int CRPhase=p->differentCREntries-1;
	int yModulus=std::numeric_limits<int>::max(); //to unify the y-component where the exact value does not matter
	if (allEntriesNonNegative){
	    if (p->matrix.upperBound[r]==NO_UPPER_BOUND){
		yModulus=p->matrix.lowerBound[r];
	    }
	    else{
		yModulus=p->matrix.upperBound[r]+1;
	    }
	}
	for (std::map<std::pair<m_entry,m_entry>,unsigned int>::iterator combIter=combinations.begin(); combIter!=combinations.end();){
	    //determine maximum power for this combination
	    int k=1;
	    int l=combIter->second;
	    while(true){
		if (k*combIter->first.first>p->CRUpperBound){
		    k--;
		    break;
		}
		k++;
	    }
	    UnivariatePolynomial uvp;
	    uvp.coeff.resize(k+1);
	    for (int i=0; i<=k; i++){
		uvp.coeff[i]=one;
	    }

	    uvp=PowerWithLimitedDeg(uvp,l,k);
	    BivariatePolynomial bvp=replaceByXiYj(uvp,combIter->first.first,combIter->first.second);
	    for (int i=0; i<bvp.size(); i++){ 
		if (bvp.monoms[i].second.first==0 && bvp.monoms[i].second.second==0){ //skip first entry, thats the constant 1
		    continue; 
		}
		if (bvp.monoms[i].second.first>p->CRUpperBound){
		    continue;
		}
		if (bvp.monoms[i].first.getDoubleValue()==0.0){
		    continue;
		}

		std::map<intpair,fppair>::iterator iter=terms.begin();
		std::map<intpair,fppair>::const_iterator endIter=terms.end();
		while(iter!=endIter){
		    if (iter->second.second.getDoubleValue()!=0.0){
			intpair newexp(iter->first.first+bvp.monoms[i].second.first,iter->first.second+bvp.monoms[i].second.second);
			if (newexp.first<=p->CRUpperBound){
			    newexp.second=std::min(newexp.second,yModulus);
			    std::map<intpair,fppair>::iterator temp=terms.find(newexp);
			    if (temp!=terms.end()){
				(*temp).second.first+=iter->second.second*bvp.monoms[i].first;
			    }
			    else{
				terms.insert(pair<intpair,fppair>(newexp,fppair(iter->second.second*bvp.monoms[i].first,0.0)));
			    }
			}
		    }
		    iter++;
		}
	    }

	    //set second component of the terms to first one
	    std::map<intpair,fppair>::iterator iter=terms.begin();
	    std::map<intpair,fppair>::const_iterator endIter=terms.end();
	    while(iter!=endIter){
		(*iter).second.second=(*iter).second.first;
		iter++;		
	    }
	    m_entry lastCR=combIter->first.first;
	    combIter++;
	    if (combIter==combinations.end()){
		this->setRowInterval(this->rivt[r]->riv,CRPhase,terms);
	    }
	    else{
		if (combIter->first.first!=lastCR){
		    this->setRowInterval(this->rivt[r]->riv,CRPhase,terms);
		    CRPhase--;
		}
	    }
	}

    }
    delete colList;
    s.clear();
    s.str("");
    s<<"done."<<std::endl;
    s<<"time: "<<TimeManager::timeSinceStart()<<"s"<<std::endl;
    writeToConsole(s);
}

void HeuristicSolveAlgorithm::setRowInterval(TwoDimArray<RowInterval>& rivarr, unsigned int CRPhase, std::map<intpair,fppair>& terms){
    for (int i=0; i<=this->problem->CRUpperBound; i++){
	rivarr[CRPhase][i].minChange=NO_UPPER_BOUND;
	rivarr[CRPhase][i].maxChange=NO_LOWER_BOUND;
    }

    std::map<intpair,fppair>::iterator iter=terms.begin();
    std::map<intpair,fppair>::const_iterator endIter=terms.end();
    while(iter!=endIter){
	for (int i=(std::max)(this->problem->CRLowerBound-(*iter).first.first,0); i<=this->problem->CRUpperBound-(*iter).first.first; i++){
	    if ((*iter).first.second<rivarr[CRPhase][i].minChange){
		rivarr[CRPhase][i].minChange=(*iter).first.second;
	    }
	    if ((*iter).first.second>rivarr[CRPhase][i].maxChange){
		rivarr[CRPhase][i].maxChange=(*iter).first.second;
	    }
	}
	iter++;
    }
}

void HeuristicSolveAlgorithm::timeOutCallBack(std::vector<void*> v){
    //v must have one entry which is a pointer to the corresponding HeuristicSolveAlgorithm object
    //This callback function has to be static, therefore the inconvenience.
    HeuristicSolveAlgorithm* p=(HeuristicSolveAlgorithm*)v[0];
    p->activeTimeoutDepth=p->plannedTimeoutDepth;
}

int* HeuristicSolveAlgorithm::solveRecursive(int step, unsigned int CRPhase, bool PV, double maxTime, double branch, unsigned int max_m, EVAL_TYPE maxDiff, EVAL_TYPE predecessorValue, EVAL_TYPE maxBTM){  
    if (this->adaptToTime){
	if (step>this->activeTimeoutDepth){ //react if timer got active
	    return NULL;
	}
	if (step<=this->plannedTimeoutDepth){ //clear timer
	    this->activeTimeoutDepth=INFTY_L;
	    this->plannedTimeoutDepth=INFTY_L;
	}
	if (this->plannedTimeoutDepth==INFTY_L && maxTime<MINIMUM_TIME_UNIT){ //set timer
	    this->plannedTimeoutDepth=step;
	    std::vector<void*> v;
	    v.push_back((void*)(this));
	    this->timeOutThread.setParameter(HeuristicSolveAlgorithm::timeOutCallBack,v,TimeManager::timeSinceStart()+maxTime);
	}
    }
    this->nodes++;
    this->maxStep=(std::max)(step,this->maxStep);
    double cTime=TimeManager::timeSinceStart();
    if (cTime-this->getLastUpdate()>this->getUpdateRate()){
	this->updateOutput=true;
    }

    int * solVector;
    unsigned long long newPermHashKey=this->currentPermHashKey^this->CRPhaseHash[CRPhase];
    if (this->permutationHash->find(newPermHashKey)){
	this->permutationHashCuts++;
	return NULL;
    }
    if (step>=this->splitLimit){
	this->permutationHash->insert(newPermHashKey);
    }

    vector<RowListEntry*> deletedRows;
    if (this->getRedundancyCheck()){
	bool CRHasSolution=this->removeRedundantRows(this->rowList,CRPhase,this->currentCR,deletedRows);
	if (!CRHasSolution){
	    this->permutationHash->insert(newPermHashKey);
    	    return NULL;
	}
    }

    unsigned int DCSizeBefore=this->deletedColumns.size();
    std::vector<ColumnEval> evalTable=this->computeBestColumns(this->rowList, CRPhase, this->currentCR,this->columnCounter,deletedColumns);
    std::sort(evalTable.begin(),evalTable.end(),CompColumnEval());
    unsigned int evalTableSize=evalTable.size();
    EVAL_TYPE bestEval=MINUS_INFTY_LL;
    unsigned long long evalHashKey=0;
    if (evalTableSize>0){
	bestEval=evalTable[0].eval;
	if (this->getUseEvalHash() && step<maxStep-2 && bestEval!=0){
	    unsigned long long hashKey1=0,hashKey2=0;
	    for (unsigned int i=0; i<evalTableSize; i++){
		ASSERT(evalTable[i].eval!=MINUS_INFTY_LL);
		hashKey1+=hashTwist(-evalTable[i].eval);
		hashKey2^=hashTwist(-evalTable[i].eval);
	    }
	    evalHashKey=((hashKey1&0xAAAAAAAAAAAAAAAALL)^(hashKey2&0x5555555555555555LL));
	    evalHashKey^=this->CRPhaseHash[CRPhase];
	    if (this->evalHash->find(evalHashKey)){
		this->evalHashCuts++;
		undeleteRows(deletedRows);
		undeleteColumns(DCSizeBefore);
		this->permutationHash->insert(newPermHashKey);
		return NULL;
	    }
	}
    }
    if (this->getRandomNoise()){
	if (evalTableSize>0){
	    EVAL_TYPE maxChange=-evalTable[0].eval;
	    if (maxChange>0){
		maxChange/=2;
		for (unsigned int i=0; i<evalTable.size(); i++){
		    EVAL_TYPE delta=(EVAL_TYPE)((get_rand()-0.5)*get_rand()*EVAL_SHIFT);
		    delta=(std::min)((std::max)(delta,-maxChange),maxChange);
		    evalTable[i].eval+=delta;
		}
	    }
	    std::sort(evalTable.begin(),evalTable.end(),CompColumnEval());
	    bestEval=evalTable[0].eval;
	}
    }
    std::vector<int> localForbiddenColumns;
    this->branchCounter[step]=0;
    lastAlternative[step]=0;
    while(lastAlternative[step]<evalTableSize &&
	  evalTable[lastAlternative[step]].eval+this->getMaxAlternativeDiff()>=bestEval && 
	  evalTable[lastAlternative[step]].eval+this->getLogOfBacktrackMargin()>=predecessorValue){
	lastAlternative[step]++;
    }
    EVAL_TYPE currMaxBTM=maxBTM;
    for (this->childNumber[step]=0; this->childNumber[step]<lastAlternative[step]; this->childNumber[step]++){
	double remainingTime=(maxTime-(TimeManager::timeSinceStart()-cTime));
	if (remainingTime<MINIMUM_TIME_UNIT && step<this->plannedTimeoutDepth && this->adaptToTime){
	    break;
	}
	if (this->branchCounter[step]>=this->getMaxBranch()){
	    break;
	}
	this->branchCounter[step]++;
	//a forbidden column counts for branching
	if (this->forbidColumns){
	    if (this->forbiddenColumns.find(evalTable[this->childNumber[step]].column)!=this->forbiddenColumns.end()){
		continue;
	    }
	}
	currMaxBTM=(std::max)(currMaxBTM,predecessorValue-evalTable[this->childNumber[step]].eval);
	this->currentCol[step]=evalTable[this->childNumber[step]].column;
	this->pushColumn(evalTable[this->childNumber[step]].column,this->problem,this->rowList,this->currentCR,this->columnCounter);
	this->currentPermHashKey+=this->columnHash[evalTable[this->childNumber[step]].column];
	bool is_solution=(currentCR>=this->problem->CRLowerBound && currentCR<=this->problem->CRUpperBound && isSolution(this->rowList));
	if (is_solution || this->updateOutput){
	    this->setLastUpdate(TimeManager::timeSinceStart());
	    this->updateOutput=false;
	    if (this->pvInfo.str()!=""){
    		stringstream s;
		s<<"                                                                               "<<std::endl;
		s<<"Data for "<<this->getName()<<":"<<std::endl;
		s<<this->pvInfo.str();
		this->pvInfo.str("");
		this->pvInfo.clear();
		s<<"nodes = "<<this->pvNodes<<std::endl;
		s<<"permutationhash cuts = "<<this->pvPermHashCuts<<std::endl;
		s<<"permutationhash size = "<<this->pvPermHashSize<<" (shared)"<<std::endl;
		s<<"evalhash cuts = "<<this->pvEvalHashCuts<<std::endl;
		s<<"evalhash size = "<<this->pvEvalHashSize<<" (shared)"<<std::endl;
		s<<"hashKey = "<<this->pvHashKey<<std::endl;
		s<<"path: ";
		for (unsigned int j=0; j<=this->pvStep; j++){
		    if (this->lastAlternative[j]>0){
    			s<<this->currentCol[j]<<"("<<this->childNumber[j]+1<<"/"<<this->lastAlternative[j]<<") ";
		    }
		    else{ //prescribed columns
			s<<this->currentCol[j]<<" ";
		    }
		}
		s<<std::endl<<"time: "<<TimeManager::timeSinceStart()<<"s"<<std::endl;
		s<<std::endl;
		writeToConsole(s);
	    }
	    else if (this->nodes>1){
		std::stringstream pathData;
		for (int j=this->pvStep+1; j<=step; j++){
		    pathData<<this->currentCol[j]<<"("<<this->childNumber[j]+1<<"/"<<this->lastAlternative[j]<<") ";
		}
		pathData<<"s: "<<step<<" e: "<<((double)(evalTable[this->childNumber[step]].eval)/EVAL_SHIFT);
		unsigned int size=pathData.str().size();
		bool smallPath=(size<=67);
		if (smallPath){
		    while(size<67){
    			pathData<<" ";
			size++;
		    }
		}
		else{
		    while(size<69){
    			pathData<<" ";
			size++;
		    }
		}
		std::string pathAsString=pathData.str();
		if (smallPath){
		    std::cout<<"path: PV... "<<pathAsString.substr(pathAsString.size()-67);
		}
		else{
		    std::cout<<"path: ... "<<pathAsString.substr(pathAsString.size()-69);
		}
		std::cout<<"\r";
	    }
	}
	if (is_solution){
	    std::stringstream s;
	    s<<"                                                                               "<<std::endl;
	    s<<this->getName()<<": Search successful!"<<std::endl;
	    s<<"path: ";
	    for (int j=0; j<=step; j++){
		if (this->lastAlternative[j]>0){
		    s<<this->currentCol[j]<<"("<<this->childNumber[j]+1<<"/"<<this->lastAlternative[j]<<") ";
		}
		else{ //prescribed columns
		    s<<this->currentCol[j]<<" ";
		}
	    }
	    s<<std::endl<<"nodes = "<<this->nodes<<std::endl;
	    s<<"permutationhash cuts = "<<this->permutationHashCuts<<std::endl;
	    s<<"permutationhash size = "<<this->permutationHash->getCurrentSize()<<" (shared)"<<std::endl;
	    s<<"evalhash cuts = "<<this->evalHashCuts<<std::endl;
	    s<<"evalhash size = "<<this->evalHash->getCurrentSize()<<" (shared)"<<std::endl;
	    s<<"maxDiff = "<<exp(((double)maxDiff/EVAL_SHIFT))<<std::endl;
	    s<<"maxBranch = "<<max_m<<std::endl;
	    this->setTerminationInfo(s.str());
	    solVector=new int[this->problem->matrix.cols];
	    for (unsigned int j=0; j<this->problem->matrix.cols; j++){
		solVector[j]=columnCounter[j];
	    }
	    this->currentPermHashKey-=this->columnHash[evalTable[this->childNumber[step]].column];
	    this->popColumn(evalTable[this->childNumber[step]].column,this->problem,this->rowList,this->currentCR,this->columnCounter);
	    undeleteRows(deletedRows);
	    undeleteColumns(DCSizeBefore);
	    return solVector;
	}
	if (this->childNumber[step]>0){
	    this->runs++;
	}
	if (PV){
	    this->pvInfo.str("");
	    this->pvInfo.clear();
	    this->pvInfo<<"Examining "<<this->childNumber[step]+1<<"th branch of PV-node at step = "<<step<<": col = "<<this->currentCol[step]<<" ("<<this->childNumber[step]+1<<"/"<<this->lastAlternative[step]<<"), eval = exp("<<((double)(evalTable[this->childNumber[step]].eval)/EVAL_SHIFT)<<"), branch = "<<branch<<"."<<std::endl;
	    this->pvStep=step;
	    this->pvNodes=this->nodes;
	    this->pvPermHashCuts=this->permutationHashCuts;
	    this->pvPermHashSize=this->permutationHash->getCurrentSize();
	    this->pvEvalHashCuts=this->evalHashCuts;
	    this->pvEvalHashSize=this->evalHash->getCurrentSize();
	    this->pvHashKey=newPermHashKey;
	}
	EVAL_TYPE nextAlternativeValue=MINUS_INFTY_LL;
	if (this->childNumber[step]+1<this->lastAlternative[step]){
	    nextAlternativeValue=evalTable[this->childNumber[step]+1].eval;
	}
	if (this->learnSet.size()!=0){
	    if (this->learnSet.find(evalTable[this->childNumber[step]].column)!=this->learnSet.end()){
		this->learnSet.erase(evalTable[this->childNumber[step]].column);
		solVector=this->solveRecursive(step+1, CRPhase, PV & (this->branchCounter[step]==1), remainingTime/2, branch*this->lastAlternative[step],(std::max)(max_m, this->branchCounter[step]),(std::max)(maxDiff,bestEval-evalTable[this->childNumber[step]].eval),(std::max)(predecessorValue,nextAlternativeValue), currMaxBTM);
	    }
	    else{
		solVector=NULL;
	    }
	}
	else{
	    solVector=this->solveRecursive(step+1, CRPhase, PV & (this->branchCounter[step]==1), remainingTime/2, branch*this->lastAlternative[step],(std::max)(max_m, this->branchCounter[step]),(std::max)(maxDiff,bestEval-evalTable[this->childNumber[step]].eval),(std::max)(predecessorValue,nextAlternativeValue), currMaxBTM);
	}
	this->currentPermHashKey-=this->columnHash[evalTable[this->childNumber[step]].column];
	this->popColumn(evalTable[this->childNumber[step]].column,this->problem,this->rowList,this->currentCR,this->columnCounter);
	if (solVector!=NULL){
	    undeleteRows(deletedRows);
	    undeleteColumns(DCSizeBefore);
	    return solVector;
	}
	if (this->forbidColumns){
	    this->forbiddenColumns.insert(evalTable[this->childNumber[step]].column);
	    localForbiddenColumns.push_back(evalTable[this->childNumber[step]].column);
	}
    }
    if (this->forbidColumns){
	for (unsigned int j=0; j<localForbiddenColumns.size(); j++){
	    this->forbiddenColumns.erase(localForbiddenColumns[j]);
	}
    }
    undeleteRows(deletedRows);
    undeleteColumns(DCSizeBefore);
    this->permutationHash->insert(newPermHashKey);
    if (evalHashKey!=0){
	this->evalHash->insert(evalHashKey);
    }
    if (CRPhase+1<this->problem->differentCREntries){
	this->nodes--;
	double remainingTime=(maxTime-(TimeManager::timeSinceStart()-cTime));
	return this->solveRecursive(step,CRPhase+1,false,remainingTime, branch, max_m,maxDiff,predecessorValue,currMaxBTM);
    }
    return NULL;
}

int* HeuristicSolveAlgorithm::solveProblem(Problem* p, vector<PCStruct>& prescribedColumns, double timeout){
    int i;
    unsigned int j;
    this->freeMemory();
    this->setLastUpdate(TimeManager::timeSinceStart());
    //copy data
    this->problem=p;
    //this->permutationHash->clear();
    //this->evalHash->clear();
    this->currentPermHashKey=0;
    this->permutationHashCuts=0;
    this->evalHashCuts=0;
    this->runs=0;
    this->nodes=0;
    this->maxStep=0;
    this->activeTimeoutDepth=INFTY_L;
    this->plannedTimeoutDepth=INFTY_L;
    this->currentCR=0;
    this->columnCounter=new int[this->problem->matrix.cols];
    for (j=0; j<this->problem->matrix.cols; j++){this->columnCounter[j]=0;}

    this->columnList=getOrderedColumnList(this->problem);
    this->computeCRPhaseEntries();
    this->rowList=getRowList(this->problem);

    if (this->getRedundancyCheck()){
	computeRowIntervals();
    }

    int startStep=0;
    int counter=0;
    j=0;
    while(j<prescribedColumns.size()){
	if (prescribedColumns[j].number>=0){
	    startStep+=prescribedColumns[j].number;
	    for (i=0; i<prescribedColumns[j].number; i++){
		this->permutationHash->insert(this->currentPermHashKey);
		this->pushColumn(prescribedColumns[j].col,this->problem,this->rowList,this->currentCR,this->columnCounter);
		this->currentPermHashKey+=this->columnHash[prescribedColumns[j].col];
		this->childNumber[counter]=0;
		this->branchCounter[counter]=0;
		this->lastAlternative[counter]=0;
		this->currentCol[counter]=prescribedColumns[j].col;
		counter++;
	    }
	}
	else{
	    startStep+=(-prescribedColumns[j].number);
	    for (i=0; i<-prescribedColumns[j].number; i++){			
		this->permutationHash->insert(this->currentPermHashKey);
		this->popColumn(prescribedColumns[j].col,this->problem,this->rowList,this->currentCR,this->columnCounter);
		this->currentPermHashKey-=this->columnHash[prescribedColumns[j].col];
		this->childNumber[counter]=0;
		this->branchCounter[counter]=0;
		this->lastAlternative[counter]=0;
		this->currentCol[counter]=prescribedColumns[j].col;
		counter++;
	    }
	}
	j++;
    }

    //solve problem stepwise
    this->updateOutput=false;
    int* res=this->solveRecursive(startStep,0,true,timeout,1.0,1,0,MINUS_INFTY_LL,MINUS_INFTY_LL);
    if (res==NULL){ //no solution found, so print search data
	std::stringstream s;
	s<<this->getName()<<": Search finished without solution. Some data:"<<std::endl;
	s<<std::endl<<"nodes = "<<this->nodes<<std::endl;
	s<<"permutationhash cuts = "<<this->permutationHashCuts<<std::endl;
	s<<"permutationhash size = "<<this->permutationHash->getCurrentSize()<<" (shared)"<<std::endl;
	s<<"evalhash cuts = "<<this->evalHashCuts<<std::endl;
	s<<"evalhash size = "<<this->evalHash->getCurrentSize()<<" (shared)"<<std::endl;
	s<<"runs = "<<runs<<std::endl;
        s<<"time: "<<TimeManager::timeSinceStart()<<"s"<<std::endl;
	this->setTerminationInfo(s.str());
    }
    delete[] this->columnCounter;
    delete[] this->columnList;
    delete[] this->rowList;
    return res;
}

void HeuristicSolveAlgorithm::computeCRPhaseEntries(){
    if (this->CRPhaseEntries!=NULL){
	delete[] this->CRPhaseEntries;
    }
    this->CRPhaseEntries=new NumberListEntry*[this->problem->differentCREntries];
    unsigned int phase=0;
    for (NumberListEntry* c=this->columnList[0].nextEntry; c!=this->columnList; c=c->nextEntry){
	static m_entry lastCR=-1;
	if (this->problem->controlRow[c->nr]!=lastCR){
	    this->CRPhaseEntries[phase]=c;
	    phase++;
	}
	lastCR=this->problem->controlRow[c->nr];
    }
}

std::vector<ColumnEval> HeuristicSolveAlgorithm::computeBestColumns(RowListEntry* relevantRows, unsigned int CRPhase, sum_of_entries currentCR, int* columnCounter, vector<NumberListEntry*>& deletedColumns){
    std::vector<ColumnEval> result;
    NumberListEntry* p=this->CRPhaseEntries[CRPhase];
    while (p->nextEntry->previousEntry!=p){
	p=p->nextEntry;
	if (p==this->columnList){
	    return result;
	}
    }
    for (m_entry currCR=this->problem->controlRow[p->nr];p!=this->columnList && this->problem->controlRow[p->nr]==currCR; p=p->nextEntry){
	unsigned int c=p->nr;
	if (columnCounter[c]>=this->problem->getUpperVarBound()){
	    p->removeFromList();
	    deletedColumns.push_back(p);
	    continue;
	}

	EVAL_TYPE tempEval=this->getEvaluator()->getEvaluation(c,relevantRows,CRPhase,currentCR+this->problem->controlRow[c]);
	if (tempEval==MINUS_INFTY_LL){
	    p->removeFromList();
	    deletedColumns.push_back(p);
	    continue;
	}
	ColumnEval ce;
	ce.eval=tempEval;
	ce.column=c;
	result.push_back(ce);
    }
    return result;
}

void HeuristicSolveAlgorithm::pushColumn(int c, Problem* p, RowListEntry* relevantRows, sum_of_entries& currentCR, int* columnCounter){
    RowListEntry* rle;
    int index=c*p->matrix.rows;
    for (rle=relevantRows->nextEntry; rle!=relevantRows; rle=rle->nextEntry){
	rle->correspondingRHS+=p->matrix.data[index+rle->nr];
    }
    currentCR+=p->controlRow[c];
    columnCounter[c]++;
}

void HeuristicSolveAlgorithm::popColumn(int c, Problem* p, RowListEntry* relevantRows, sum_of_entries& currentCR, int* columnCounter){
    RowListEntry* rle;
    int index=c*p->matrix.rows;
    for (rle=relevantRows->nextEntry; rle!=relevantRows; rle=rle->nextEntry){
	rle->correspondingRHS-=p->matrix.data[index+rle->nr];
    }
    currentCR-=p->controlRow[c];
    columnCounter[c]--;
}

bool HeuristicSolveAlgorithm::isSolution(RowListEntry* relevantRows){
    RowListEntry* rle;
    for (rle=relevantRows->nextEntry; rle!=relevantRows; rle=rle->nextEntry){
	if (rle->correspondingRHS<rle->correspondingLowerBound || rle->correspondingRHS>rle->correspondingUpperBound){
	    return false;
	}
    }
    return true;
}

bool HeuristicSolveAlgorithm::removeRedundantRows(RowListEntry* relevantRows, unsigned int CRPhase, sum_of_entries currentCR, vector<RowListEntry*>& vec){
    RowListEntry* rle;
    for (rle=relevantRows->nextEntry; rle!=relevantRows; rle=rle->nextEntry){
	RowInterval& ri=this->rivt[rle->nr]->riv[CRPhase][currentCR];
	//if the following happens than already for the first row, so nothing has been removed yet!
	if (ri.minChange==NO_UPPER_BOUND && ri.maxChange==NO_LOWER_BOUND){ 		 
	    return false;
	}
	if (rle->correspondingRHS+ri.maxChange<=rle->correspondingUpperBound &&
	    rle->correspondingRHS+ri.minChange>=rle->correspondingLowerBound){
    	    vec.push_back(rle);
	    rle->removeFromList();
	}
    }
    return true;
}

void HeuristicSolveAlgorithm::setBacktrackMargin(double q){
    this->setLogOfBacktrackMargin((EVAL_TYPE)(log(q)*EVAL_SHIFT));
}

double HeuristicSolveAlgorithm::getBacktrackMargin(){
    return exp((double)(this->getLogOfBacktrackMargin())/EVAL_SHIFT);
}

void HeuristicSolveAlgorithm::setLogOfBacktrackMargin(EVAL_TYPE q){
    this->logOfBacktrackMargin=q;
}

EVAL_TYPE HeuristicSolveAlgorithm::getLogOfBacktrackMargin(){
    return this->logOfBacktrackMargin;
}

void HeuristicSolveAlgorithm::setPermutationHash(unsigned long long* ch, unsigned long long* crph, HashTable* ht){
    this->columnHash=ch;
    this->CRPhaseHash=crph;
    this->permutationHash=ht;
}

std::pair<unsigned long long*, HashTable*> HeuristicSolveAlgorithm::getPermutationHash(){
    return std::pair<unsigned long long*, HashTable*>(this->columnHash,this->permutationHash);
}

void HeuristicSolveAlgorithm::setEvalHash(HashTable* h){
    this->evalHash=h;
}

HashTable* HeuristicSolveAlgorithm::getEvalHash(){
    return this->evalHash;
}

void HeuristicSolveAlgorithm::setSplitLimit(int d){
    this->splitLimit=d;
}

int HeuristicSolveAlgorithm::getSplitLimit(){
    return this->splitLimit;
}

void HeuristicSolveAlgorithm::setName(std::string s){
    this->name=s;
}

std::string HeuristicSolveAlgorithm::getName(){
    return this->name;
}

void HeuristicSolveAlgorithm::setLearnSet(std::set<unsigned int>& set){
    this->learnSet=set;
}

std::set<unsigned int> HeuristicSolveAlgorithm::getLearnSet(){
    return this->learnSet;
}

