/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/Polynomial.h"
#include "../h/CommonRoutines.h"
#include <string>
#include <sstream>
#include <map>


int UnivariatePolynomial::degree(){
    return (int)(coeff.size())-1;
}

std::string UnivariatePolynomial::asString(){
    std::stringstream ss;
    for (int i=this->degree(); i>0; i--){
	if (this->coeff[i].getDoubleValue()!=0.0){
	    MyFPType coeff=this->coeff[i];
	    coeff.expandDouble();
	    ss<<coeff.getDoubleValue()<<"*X^"<<i<< "+ ";
	}
    }
    if (this->coeff[0].getDoubleValue()!=0.0){
	MyFPType coeff=this->coeff[0];
	coeff.expandDouble();
	ss<<coeff.getDoubleValue();
    }
    return ss.str();
}

int BivariatePolynomial::size(){
    return monoms.size();
}

std::string BivariatePolynomial::asString(){
    std::stringstream ss;
    for (int i=this->monoms.size()-1; i>0; i--){
	if (this->monoms[i].first.getDoubleValue()!=0.0){
	    MyFPType coeff=this->monoms[i].first;
	    coeff.expandDouble();
	    ss<<coeff.getDoubleValue()<<"*X^"<<this->monoms[i].second.first<<"Y^"<<this->monoms[i].second.second<< "+ ";
	}
    }
    if (this->monoms[0].first.getDoubleValue()!=0.0){
	MyFPType coeff=this->monoms[0].first;
	coeff.expandDouble();
	ss<<coeff.getDoubleValue()<<"*X^"<<this->monoms[0].second.first<<"Y^"<<this->monoms[0].second.second;
    }
    return ss.str();
}

UnivariatePolynomial getXTTiMinus1TTl(int i,int l){
    UnivariatePolynomial res;
    res.coeff.resize(i*l+1);
    MyFPType zero(0.0);
    MyFPType minusOne(-1.0);
    int j=0;
    while(true){
    	MyFPType coeff=getBinomialCoeff(l,j);
	if ((l-j)&1){
	    coeff*=minusOne;
	}
	res.coeff[j*i]=coeff;
	if (j==l){
	    break;
	}
	for (int k=j*i+1; k<j*i+i; k++){
	    res.coeff[k]=zero;
	}
	j++;
    }
    return res;
}

UnivariatePolynomial operator* (UnivariatePolynomial& a, UnivariatePolynomial& b){
    std::map<int,MyFPType> monoms;
    for (int i=0; i<=a.degree(); i++){
	for (int j=0; j<=b.degree(); j++){
	    std::map<int,MyFPType>::iterator iter=monoms.find(i+j);
	    if (iter==monoms.end()){
		monoms.insert(std::pair<int,MyFPType>(i+j,a.coeff[i]*b.coeff[j]));
	    }
	    else{
		iter->second+=a.coeff[i]*b.coeff[j];
	    }
	}
    }
    UnivariatePolynomial result;
    int currDeg=0;
    MyFPType zero(0.0);
    for (std::map<int,MyFPType>::iterator iter=monoms.begin(); iter!=monoms.end(); iter++){
	while(currDeg<iter->first){
	    result.coeff.push_back(MyFPType(0.0));
	    currDeg++;
	}
	result.coeff.push_back(iter->second);
	currDeg++;
    }
    return result;
}

UnivariatePolynomial multiplyWithLimitedDeg(UnivariatePolynomial& a, UnivariatePolynomial& b, int maxDeg){
    std::map<int,MyFPType> monoms;
    int mina=(std::min)(maxDeg,a.degree());
    for (int i=0; i<=mina; i++){
	for (int j=0; j<=maxDeg-i; j++){
	    int newDeg=i+j;
	    std::map<int,MyFPType>::iterator iter=monoms.find(newDeg);
	    if (iter==monoms.end()){
		monoms.insert(std::pair<int,MyFPType>(newDeg,a.coeff[i]*b.coeff[j]));
	    }
	    else{
		iter->second+=a.coeff[i]*b.coeff[j];
	    }
	}
    }
    UnivariatePolynomial result;
    int currDeg=0;
    MyFPType zero(0.0);
    for (std::map<int,MyFPType>::iterator iter=monoms.begin(); iter!=monoms.end(); iter++){
	while(currDeg<iter->first){
	    result.coeff.push_back(MyFPType(0.0));
	    currDeg++;
	}
	result.coeff.push_back(iter->second);
	currDeg++;
    }
    return result;
}

UnivariatePolynomial Power(UnivariatePolynomial p, unsigned int pow){
    UnivariatePolynomial result;
    if (pow==1){
	return p;
    }
    result.coeff.push_back(MyFPType(1.0));
    if (pow==0){
	return result;
    }
    Power_loop:
    if (pow&1){
	result=result*p;
    }
    pow>>=1;
    if (pow == 0) {
	return result;
    } 
    else{
	p=p*p;
	goto Power_loop;
    }
    return result;
}

UnivariatePolynomial PowerWithLimitedDeg(UnivariatePolynomial p, unsigned int pow, int maxDeg){
    UnivariatePolynomial result;
    if (pow==1){
	return p;
    }
    result.coeff.push_back(MyFPType(1.0));
    if (pow==0){
	return result;
    }
    Power_loop:
    if (pow&1){
	result=multiplyWithLimitedDeg(result,p,maxDeg);
    }
    pow>>=1;
    if (pow == 0) {
	return result;
    } 
    else{
	p=multiplyWithLimitedDeg(p,p,maxDeg);
	goto Power_loop;
    }
    return result;
}

UnivariatePolynomial PolynomDivision(UnivariatePolynomial p, UnivariatePolynomial q){
    UnivariatePolynomial res;
    const int deg=p.degree()-q.degree();
    const int degq=q.degree();
    res.coeff.resize(deg+1);
    for (int i=deg; i>=0; i--){
	res.coeff[i]=p.coeff[p.degree()];
	res.coeff[i]/=q.coeff[q.degree()];
	p.coeff.pop_back();
	for (int j=0; j<degq; j++){
	    p.coeff[i+j]-=res.coeff[i]*q.coeff[j];
	}
    }
    return res;
}
    
BivariatePolynomial replaceByXiYj(UnivariatePolynomial p, int i, int j){
    BivariatePolynomial res;
    for (int k=0; k<=p.degree(); k++){
	if (p.coeff[k].getDoubleValue()==0.0){
	    continue;
	}
	res.monoms.push_back(BivariateMonomial(p.coeff[k],intpair(k*i,k*j)));
    }
    return res;
}
