/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/Solver.h"
#include "../h/FileProblemConstructor.h"
#include "../h/HeuristicSolveAlgorithm.h"
#include "../h/SampledProbEvaluator.h"
#include "../h/ExactProbEvaluator.h"
#include "../h/CommonRoutines.h"
#include "../h/Log.h"
#include "../h/NetCommunicator.h"
#include "../h/PAGCodeConstructor.h"
#include "../h/KMMConstructor.h"
#include "../h/TimeManager.h"
#include "../h/TimeOutThread.h"
#include "../h/solveThread.h"
#include "../h/Random.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cstdlib>

using namespace std;

string solutionFile;
string benchFile;
string benchFileCallString;

int main(int argc, char* argv[])
{
    unsigned int numberOfThreads=1;
    int splitLimit=3;
    unsigned long long permutationHashSize=DEFAULT_PERMHASH_SIZE;
    unsigned long long evalHashSize=DEFAULT_EVALHASH_SIZE;
    double timeout=INFTY_DB;
    TimeOutThread timeOutThread;
    ExactProbEvaluator epe;
    SampledProbEvaluator spe;

    //quick check if we have use multiple threads
    for (int i=1; i<argc; i++){
	if (std::string(argv[i])=="-nthreads" || std::string(argv[i])=="nthreads"){
	    if (i+1<argc){
		numberOfThreads=atoi(argv[i+1]);
		if (numberOfThreads>16){
		    std::stringstream ss;
		    ss<<"Error: Number of threads should be lower than 16."<<std::endl;
		    writeToConsole(ss);
		    quitSolver(ERROR_OCCURED);
		}
		else if (numberOfThreads==0){
		    std::stringstream ss;
		    ss<<"Warning: Number of threads has to be >=1."<<std::endl;
		    writeToConsole(ss);
		    quitSolver(ERROR_OCCURED);
		}
	    }
	    else{
		std::stringstream ss;
		ss<<"Error: -nthreads need <numberOfThreads> as further argument!"<<std::endl;
		writeToConsole(ss);
		quitSolver(ERROR_OCCURED);
	    }
	}
    }

    HeuristicSolveAlgorithm* h=new HeuristicSolveAlgorithm[numberOfThreads];
    TimeManager::setStartTimer();
    Problem* p=NULL;
    vector<PCStruct> prescribedColumns;
    std::set<unsigned int> learnSet;
    int upperBound=NO_UPPER_BOUND;
    double pac=DEFAULT_PAC;
    bool useSPE=false;
    for (int i=1; i<argc; i++){
	string s(argv[i]);
	if (s=="file" || s=="-file"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -file needs <filename> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    if (p!=NULL){
		stringstream s2;
		s2<<"Error: -file: problem already specified!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    FileProblemConstructor fpc(string((const char*)argv[i+1]));
	    p=fpc.getProblem();
	    i+=1;
	    continue;
	}
	if (s=="pagc" || s=="-pagc"){
	    if (argc<i+1+5){
		stringstream s2;
		s2<<"Error: -pagc needs <n>, <k>, <d>, <q>, <groupid> as further arguments!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    if (p!=NULL){
		stringstream s2;
		s2<<"Error: -pagc: problem already specified!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    int n=atoi(argv[i+1]);
	    int k=atoi(argv[i+2]);
	    int d=atoi(argv[i+3]);
	    int q=atoi(argv[i+4]);
	    int groupid=atoi(argv[i+5]);
	    PAGCodeConstructor cc(n,k,d,q,groupid);
	    p=cc.getProblem();
	    i+=5;
	    continue;
	}
	if (s=="pagqa" || s=="-pagqa"){
	    if (argc<i+1+3){
		stringstream s2;
		s2<<"Error: -pagqa needs <filename>, <n>, <lambda> as further arguments!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    string name(argv[i+1]);
	    sum_of_entries n=atoi(argv[i+2]);
	    sum_of_entries lambda=atoi(argv[i+3]);
	    KMMConstructor kc(name,IEQ_TYPE_LE,n,lambda);
	    p=kc.getProblem();
	    i+=3;
	}
	if (s=="timeout" || s=="-timeout"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -timeout needs <t> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    timeout=atof(argv[i+1]);
	    std::vector<void*> v;
	    timeOutThread.setParameter(&timeOutCallBack,v,timeout);
	    i+=1;
	    continue;
	}
	if (s=="timeadaption" || s=="-timeadaption"){
	    for (unsigned int j=0; j<numberOfThreads; j++){
    		h[j].setAdaptToTime(true);
	    }
	    continue;
	}
	if (s=="randomnoise" || s=="-randomnoise"){
	    for (unsigned int j=0; j<numberOfThreads; j++){
    		h[j].setRandomNoise(true);
	    }
	    continue;
	}
	if (s=="w" || s=="-w"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -w needs <max-branch-factor> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    unsigned int w=atoi(argv[i+1]);
	    for (unsigned int j=0; j<numberOfThreads; j++){
		h[j].setMaxBranch(w);
	    }
	    i+=1;
	    continue;
	}
	if (s=="a" || s=="-a"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -a needs <alternative-bound> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    double a=atof(argv[i+1]);
	    for (unsigned int j=0; j<numberOfThreads; j++){
		h[j].setAlternativeBound(a);
	    }
	    i+=1;
	    continue;
	}
	if (s=="btmargin" || s=="-btmargin"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -btmargin needs <backtrack-margin> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    double btm=atof(argv[i+1]);
	    for (unsigned int j=0; j<numberOfThreads; j++){
		h[j].setBacktrackMargin(btm);
	    }
	    i+=1;
	    continue;
	}
	if (s=="s" || s=="-s"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -s needs <samples> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    unsigned int samples=(unsigned int)atoi(argv[i+1]);
	    spe.setSampleRuns(samples);
	    i+=1;
	    continue;
	}
	if (s=="b" || s=="-b"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -b needs <upperbound> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    upperBound=(unsigned int)atoi(argv[i+1]);
	    i+=1;
	    continue;
	}
	if (s=="sample" || s=="-sample"){
	    useSPE=true;
	    continue;
	}
	if (s=="fc" || s=="-fc"){
	    for (unsigned int j=0; j<numberOfThreads; j++){
		h[j].setForbidColumns(true);
	    }
	    continue;
	}
	if (s=="noevalhash" || s=="-noevalhash"){
	    for (unsigned int j=0; j<numberOfThreads; j++){
    		h[j].setUseEvalHash(false);
	    }
	    continue;
	}
	if (s=="evalhashsize" || s=="-evalhashsize"){
	    if (argc<i+1){
		stringstream s2;
		s2<<"Error: -evalhashsize needs <size in MB> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    unsigned int sizeInMB=atoi(argv[i+1]);
	    evalHashSize=(1024*1024ULL)*(unsigned long long)(sizeInMB);
	    i+=1;
	    continue;
	}
	if (s=="permhashsize" || s=="-permhashsize"){
	    if (argc<i+1){
		stringstream s2;
		s2<<"Error: -permhashsize needs <size in MB> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    unsigned int sizeInMB=atoi(argv[i+1]);
	    permutationHashSize=(1024*1024ULL)*(unsigned long long)(sizeInMB);
	    i+=1;
	    continue;
	}
	if (s=="splitlimit" || s=="-splitlimit"){
	    if (argc<i+1){
		stringstream s2;
		s2<<"Error: -splitlimit needs <depth> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    splitLimit=atoi(argv[i+1]);
	    i+=1;
	    continue;
	}
	if (s=="noredundancycheck" || s=="-noredundancycheck"){
	    for (unsigned int j=0; j<numberOfThreads; j++){
		h[j].setRedundancyCheck(false);
	    }
	}
	if (s=="u" || s=="-u"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -u needs <update_rate> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    for (unsigned int j=0; j<numberOfThreads; j++){
		h[j].setUpdateRate(atof(argv[i+1]));
	    }
	    i+=1;
	    continue;
	}
	if (s=="l" || s=="-l"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -l needs <log_filename> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    setLogging(std::string(argv[i+1]));
	    i+=1;
	    continue;
	}
	if (s=="o" || s=="-o"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -o needs <solution_filename> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    solutionFile=std::string(argv[i+1]);
	    i+=1;
	    continue;
	}
	if (s=="benchfile" || s=="-benchfile"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -benchfile needs <bench_filename> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    benchFile=std::string(argv[i+1]);
	    i+=1;
	    std::stringstream bf;
	    for (int i=1; i<argc-1; i++){
		bf<<argv[i]<<" ";
	    }
	    bf<<argv[argc-1]<<": ";
	    benchFileCallString=bf.str();
	    continue;
	}
	if (s=="pc" || s=="-pc"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -pc needs <filename> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    ifstream file;
	    file.open(argv[i+1]);
	    if (file.is_open()){
		PCStruct p;
		while(true){
		    file>>p.number;
		    file>>p.col;
		    if (file.eof() || p.number==0){
			break;
		    }
		    prescribedColumns.push_back(p);
		}				
		file.close();
	    }
	    else{
		stringstream s2;
		s2<<"Error: could not open "<<argv[i+1]<<"!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    i+=1;
	    continue;
	}
	if (s=="learnfrom" || s=="-learnfrom"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -learnfrom needs <filename> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    ifstream file;
	    file.open(argv[i+1]);
	    if (file.is_open()){
		while(true){
		    unsigned int cn;
		    file>>cn;
		    if (file.eof()){
			break;
		    }
		    learnSet.insert(cn);
		}				
		file.close();
	    }
	    else{
		stringstream s2;
		s2<<"Error: could not open "<<argv[i+1]<<"!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    i+=1;
	    continue;
	}
	if (s=="p" || s=="-p"){
	    if (argc<i+1+1){
		stringstream s2;
		s2<<"Error: -p needs <probability> as further argument!"<<std::endl;
		writeToConsole(s2);
		quitSolver(ERROR_OCCURED);
	    }
	    pac=atof(argv[i+1]);
	    i+=1;
	    continue;
	}
    }

    if (p==NULL){
	stringstream s;
	s<<"Error: no problem specified!"<<std::endl;
	writeToConsole(s);
	quitSolver(ERROR_OCCURED);
    }

    p->setUpperVarBound(upperBound);
    if (upperBound!=NO_UPPER_BOUND){
	appendVarBoundInequations(p, upperBound);
    }

    spe.setPAC(pac);
    if (useSPE){
	spe.setProblem(p);
	for (unsigned int j=0; j<numberOfThreads; j++){
	    h[j].setEvaluator((AbstractEvaluator*)&spe);
	}
    }
    else{
        epe.setProblem(p);
	for (unsigned int j=0; j<numberOfThreads; j++){
	    h[j].setEvaluator((AbstractEvaluator*)&epe);
	}
    }
    HashTable permutationHash(permutationHashSize);
    HashTable evalHash(evalHashSize);
    unsigned long long* columnHash=new unsigned long long[p->matrix.cols];
    for (unsigned int j=0; j<p->matrix.cols; j++){
	columnHash[j]=((unsigned long long)(get_rand()*0xFFFFFFFF)<<32)+(unsigned long long)(get_rand()*0xFFFFFFFF);
    }
    unsigned long long* CRPhaseHash=new unsigned long long[p->differentCREntries];
    for (unsigned int j=0; j<p->differentCREntries; j++){
	CRPhaseHash[j]=((unsigned long long)(get_rand()*0xFFFFFFFF)<<32)+(unsigned long long)(get_rand()*0xFFFFFFFF);
    }

    AbstractSolveAlgorithm** asa=new AbstractSolveAlgorithm*[numberOfThreads];

    for (unsigned int j=0; j<numberOfThreads; j++){
	std::stringstream ss;
	ss<<"thread "<<j;
	h[j].setName(ss.str());
	h[j].setPermutationHash(columnHash,CRPhaseHash,&permutationHash);
        h[j].setEvalHash(&evalHash);
	h[j].setSplitLimit(splitLimit);
	h[j].setLearnSet(learnSet);
	asa[j]=(AbstractSolveAlgorithm*) &h[j];
    }

    SolveThread* st=new SolveThread[numberOfThreads];
    for (unsigned int j=0; j<numberOfThreads; j++){
	st[j].solveProblem(p,prescribedColumns,asa[j],timeout);
    }

    int* solVector=NULL;
    while(true){
	bool runningThreads=false;
	for (unsigned int j=0; j<numberOfThreads; j++){
	    if (!st[j].getRunning()){
		if (st[j].getSolution()!=NULL){
		    solVector=st[j].getSolution();
		    std::cout<<st[j].getTerminationInfo();
		    break;
		}
	    }
	    else{
		runningThreads=true;
	    }
	}
	if (solVector!=NULL || !runningThreads){
	    break;
	}
	waitSomeTime(50000000);
    }
    delete[] columnHash;
    delete[] CRPhaseHash;
    /*
    std::cout<<"Starting search a second time..."<<std::endl;
    solVector=solveAlgorithm->solveProblem(p,prescribedColumns);
    std::cout<<"And now a third time with cleared hash..."<<std::endl;
    permutationHash.clear();
    evalHash.clear();
    solVector=solveAlgorithm->solveProblem(p,prescribedColumns);
    getchar();*/

    sum_of_entries CR=0;
    if (solVector!=NULL){
	stringstream solutionss;
	solutionss<<"solution vector:"<<std::endl;
	for (unsigned int j=0; j<p->matrix.cols-1; j++){
	    for (int k=0; k<solVector[j];k++){
		CR+=p->controlRow[j];
	    }
	    solutionss<<solVector[j]<<" ";
	}
	for (int k=0; k<solVector[p->matrix.cols-1];k++){
	    CR+=p->controlRow[p->matrix.cols-1];
	}
	solutionss<<solVector[p->matrix.cols-1]<<std::endl;	
	solutionss<<"Control row with this solution is "<<CR<<"."<<std::endl;
	writeToConsole(solutionss);
	saveSolution(solutionss);
	delete[] solVector;
    }
    else{
	std::cout<<st[0].getTerminationInfo();
	stringstream s;
	s<<"                                                                               "<<std::endl;
	s<<"Search failed."<<std::endl;
	writeToConsole(s);
    }
    if (solVector==NULL){ //no solution
	quitSolver(SEARCH_NOT_STOPPED);
    }
    else{
	quitSolver(SOLUTION_FOUND); //solution found
    }
}

void stopSearch(int reason){
    stringstream s2;
    s2<<"Stopping search due to ";
    switch (reason){
		case SEARCH_STOPPED_TIMEOUT:
		    s2<<"timeout";
		    break;
		default:
		    s2<<"an unknown reason";
		    break;
    }
    s2<<"..."<<std::endl;
    writeToConsole(s2);
    quitSolver(SEARCH_STOPPED_TIMEOUT);
}

void quitSolver(int r){
    std::stringstream s;
    if (r==ERROR_OCCURED){ //error
	s<<"Program stopped due to an error."<<std::endl;
    }
    else if (r==SOLUTION_FOUND){ //solution found
	s<<"Program terminated normally, a solution was found."<<std::endl;
    }
    else{ //no solution found, but different "reasons" possible 
	if (r==SEARCH_NOT_STOPPED){
	    s<<"Program terminated normally, no solution was found (search completed)."<<std::endl;
	}
	else if (r==SEARCH_STOPPED_TIMEOUT){
	    s<<"Program terminated normally, no solution was found (time limit reached)."<<std::endl;
	}
	else{
	    s<<"Program terminated normally, no solution was found (unknown return value)."<<std::endl;
	}
    }
    s<<"time: "<<TimeManager::timeSinceStart()<<"s"<<std::endl;
    writeToConsole(s);
    if (benchFile!=""){
	ofstream bf;
	bf.open(benchFile.c_str(),ios::out|ios::app);
	if (bf.is_open()){
	    bf<<benchFileCallString;
	    if (r==SOLUTION_FOUND){
		bf<<"solved after "<<TimeManager::timeSinceStart()<<" s"<<std::endl<<flush;
	    }
	    else{
		bf<<"no solution found after "<<TimeManager::timeSinceStart()<<" s"<<std::endl<<flush;
	    }
	    bf.close();
	}
	else{
	    stringstream s2;
	    s2<<"Error: could not open benchfile "<<benchFile<<"!"<<std::endl;
	    writeToConsole(s2);
	}
    }
    closeLog();
    exit(r);
}

void timeOutCallBack(std::vector<void*> v){
    quitSolver(SEARCH_STOPPED_TIMEOUT);
}

void saveSolution(stringstream& s){
    if (solutionFile==""){
	return;
    }
    ofstream f;
    f.open(solutionFile.c_str());
    if (f.is_open()){
	f<<s.str();
    }
    else{
	stringstream s2;
	s2<<"Error: could not open solution file "<<solutionFile.c_str()<<"!"<<std::endl;
	writeToConsole(s2);
    }
}


