/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef HEURISTIC_SOLVE_ALGORITHM_H
#define HEURISTIC_SOLVE_ALGORITHM_H

#include "AbstractSolveAlgorithm.h"
#include "AbstractEvaluator.h"
#include "ColumnEval.h"
#include "NumberListEntry.h"
#include "RowListEntry.h"
#include "twoDimArray.hpp"
#include "hashTable.h"
#include "time.h"
#include "Defs.h"
#include "commonStructs.h"
#include "myFPType.h"
#include "Polynomial.h"
#include "TimeOutThread.h"
#include <set>
#include <vector>
#include <sstream>
#include <map>
#include <list>

using namespace std;

#define DEFAULT_ALTERNATIVE_BOUND 1.5
#define DEFAULT_BACKTRACK_MARGIN INFTY_DB
#define DEFAULT_MAX_BRANCH 0xFFFFFFFF
#define DEFAULT_UPDATE_RATE 1.0
#define MAX_STEPS 1000

class HeuristicSolveAlgorithm : AbstractSolveAlgorithm{
public:
    HeuristicSolveAlgorithm();
    ~HeuristicSolveAlgorithm();
    void freeMemory();
    void undeleteRows(vector<RowListEntry*>&);
    void undeleteColumns(unsigned int sizeBefore);
    void computeRowIntervals();
    bool removeRedundantRows(RowListEntry* relevantRows, unsigned int CRPhase, sum_of_entries currentCR, vector<RowListEntry*>& vec);
    int* solveProblem(Problem* p, vector<PCStruct>& prescribedColumns, double timeout);
    int* solveRecursive(int step, unsigned int CRPhase, bool PV, double maxTime, double branch, unsigned int max_m, EVAL_TYPE maxDiff, EVAL_TYPE predecessorValue, EVAL_TYPE maxBTM);
    bool isSolution(RowListEntry*);
    void pushColumn(int, Problem*, RowListEntry*, sum_of_entries&, int*);
    void popColumn(int, Problem*, RowListEntry*, sum_of_entries&, int*);
    void setEvaluator(AbstractEvaluator*);
    AbstractEvaluator* getEvaluator();
    void setMaxBranch(unsigned int);
    unsigned int getMaxBranch();
    void setAlternativeBound(double);
    double getAlternativeBound();
    void setLastUpdate(double);
    double getLastUpdate();
    void setUpdateRate(double);
    double getUpdateRate();
    void setForbidColumns(bool);
    bool getForbidColumns();
    void setUseEvalHash(bool);
    bool getUseEvalHash();
    void setRedundancyCheck(bool);
    bool getRedundancyCheck();
    void setAdaptToTime(bool);
    bool getAdaptToTime();
    EVAL_TYPE getMaxAlternativeDiff();
    void setBacktrackMargin(double q);
    double getBacktrackMargin();
    void setLogOfBacktrackMargin(EVAL_TYPE q);
    EVAL_TYPE getLogOfBacktrackMargin();
    void setPermutationHash(unsigned long long* ch, unsigned long long* crph, HashTable* ht);
    std::pair<unsigned long long*, HashTable*> getPermutationHash();
    void setEvalHash(HashTable* h);
    HashTable* getEvalHash();
    void setSplitLimit(int d);
    int getSplitLimit();
    void setName(std::string s);
    std::string getName();
    void setLearnSet(std::set<unsigned int>& set);
    std::set<unsigned int> getLearnSet();
    void setRandomNoise(bool);
    bool getRandomNoise();
    static void timeOutCallBack(std::vector<void*> v);
private:
    bool adaptToTime;
    bool forbidColumns;
    bool useEvalHash;
    bool redundancyCheck;
    bool randomNoise;
    Problem* problem;
    std::string name;
    int splitLimit;
    unsigned int maxBranch;
    double alternativeBound;
    EVAL_TYPE maxAlternativeDiff;
    EVAL_TYPE logOfBacktrackMargin;
    AbstractEvaluator* itsEvaluator;
    HashTable* permutationHash;
    HashTable* evalHash;
    double lastUpdate;
    double updateRate;
    unsigned int permutationHashCuts;
    unsigned int evalHashCuts;
    unsigned long long nodes;
    unsigned long long runs;
    sum_of_entries currentCR;
    int* columnCounter;
    unsigned long long* columnHash;
    unsigned long long* CRPhaseHash;
    unsigned long long currentPermHashKey;
    int maxStep;
    unsigned int currentCol[MAX_STEPS]; //current column investigated in a node
    unsigned int childNumber[MAX_STEPS]; //number of the current column in the list ordered by evaluation
    unsigned int branchCounter[MAX_STEPS]; //how many columns (with the current one) have been searched
    unsigned int lastAlternative[MAX_STEPS]; //how many alternatives exist in this node
    RowListEntry* rowList;
    NumberListEntry* columnList;
    NumberListEntry** CRPhaseEntries;
    void computeCRPhaseEntries();
    void setRowInterval(TwoDimArray<RowInterval>& rivarr, unsigned int CRPhase, std::map<intpair,fppair>& terms);
    std::vector<ColumnEval> computeBestColumns(RowListEntry* relevantRows, unsigned int CRPhase, sum_of_entries currentCR, int* columnCounter, vector<NumberListEntry*>& deletedColumns);
    std::vector<NumberListEntry*> deletedColumns;
    std::set<unsigned int> forbiddenColumns;
    std::set<unsigned int> learnSet;
    RowIntervalTable** rivt;
    bool* rivtMemAllocated;
    std::stringstream pvInfo;
    unsigned int pvStep;
    unsigned long long pvNodes;
    unsigned long long pvPermHashCuts;
    unsigned long long pvPermHashSize;
    unsigned long long pvEvalHashCuts;
    unsigned long long pvEvalHashSize;
    unsigned long long pvHashKey;
    bool updateOutput;
    int plannedTimeoutDepth;
    int activeTimeoutDepth;
    TimeOutThread timeOutThread;
};

#endif
